/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package content.exercises;

import java.awt.Color;
import java.util.HashMap;

import matrix.decoration.*;
import matrix.structures.FDT.*;
import matrix.structures.memory.*;
import matrix.util.Note;
/**
 * New Implementation of Vertex with priority attached to it. Used for calculating
 * the spanning tree for a graph. See, e.g., Prim's algorithm and the corresponding exercise.
 *
 * @author Ari Korhonen
 */
public class PriorityExerVertex extends ExerVertex implements Source, LabelDecorator, SimulationTree2, UndirectedGraph, Visitable {
    // static values that wont change during the simulation
    int priority = Integer.MAX_VALUE;
    int time_stamp = 0;
    static int global_time = 0;
    PriorityExerVertex[] succ; // the structure of the graph wont change after it's established
    HashMap weights = new HashMap();
    // dynamic values (the spanning tree) should be stored in memory structures 
    VirtualArray spanningtree = new VirtualArray();
    VirtualInteger children = new VirtualInteger(0);
    VirtualBoolean visited = new VirtualBoolean(false);
    VirtualObject father = new VirtualObject(null);
    VirtualObject label = new VirtualObject("");
    private VirtualArray edgeColor = new VirtualArray();

    public PriorityExerVertex() { 
	init();
    }
    
    public PriorityExerVertex(Object o) {
	super(o);
	init();
    }

    void init() {
	getStyleSheet().setVisitedBackgroundColor(Color.red);
	getStyleSheet().setVisitedBorderColor(Color.red);
	getStyleSheet().setVisitedFocusedBorderColor(Color.red);
    }

    public void insertChild(PriorityExerVertex v) {
	//Note.out(this, "new child " +v+" for " + this);
	spanningtree.setObject(v, children.eval());
	children.inc();
    }

    public boolean hasSuccessor(PriorityExerVertex u) {
	Vertex[] v = super.getSuccessors();
	for (int i=0; i<v.length; i++)
	    if (v[i] == u)
		return true;
	return false;
    }

    public void select(PriorityExerVertex u) {
	markVisited();
	//graphs should color both symmetric edges
	Vertex[] v = super.getSuccessors();
	for (int i=0; i<v.length; i++) {
	    if (u == v[i]) {
		//Note.out(this, "2) edge " + this.toString() + "->" + v[i].toString());
		edgeColor.setObject(Color.red, i);
		//Note.out(this, "set also " + this.toString());
            }
	}
    }

    //--- Implementation for interface Visitable ---
    public boolean isVisited() {
	return visited();
    }

    //--- End of interface Visitable  ---

    //--- Implementation for interface Source ---

    public void referenceSelectedTo(FDT s) {
	if (s instanceof PriorityExerVertex)
	    referenceSelectedTo((PriorityExerVertex)s);
	else
	    Note.warning(this, "Parameter " + s + " not applicable");
    }
    
    public void referenceSelectedTo(PriorityExerVertex to) {
	if (to.visited() && (!(this.visited()))) {
	    to.referenceSelectedTo(this);
	    return;
	}
	markVisited();
	//Note.out(this, ""+this.toString() + " visited");
	Vertex[] v = super.getSuccessors();
	for (int i=0; i<v.length; i++)
	    if (v[i] == to) {
		//Note.out(this, "1) edge " + this.toString() + "->" + v[i].toString());
		((PriorityExerVertex)v[i]).select(this);
		edgeColor.setObject(Color.red, i);
	    }
	for (int i=0; i<spanningtree.size(); i++)
	    if (spanningtree.getObject(i) == to)
		return;
	insertChild(to);
    }
    
    public boolean isVisitedReferenceTo(FDT to) {
	Vertex[] v = super.getSuccessors();
	for (int i=0; i<v.length; i++)
	    if (v[i] == to) {
		Object c = edgeColor.getObject(i);
		return c != null ? c.equals(Color.red) : false;
	    }
	return false;
    }

    //--- End Source ---

    //--- Implementation for interface Tree and SimulationTree(2) ---

    public int getSubTreeCount() {
	return children.eval();
    }

    public void setSubTree(Tree t, int pos) {
	//Note.out(this, "setSubTree " + t + " at " + pos);
	for (int i=spanningtree.size(); i>pos; i--)
	    spanningtree.setObject(spanningtree.getObject(i-1), i);
	spanningtree.setObject(t, pos);
    }

    public Tree getNewNode(Object o) {
	return new PriorityExerVertex(o);
    }

    public Tree[] getSubTrees() {
	Tree[] st = new Tree[spanningtree.size()];
	for (int i=0; i<spanningtree.size(); i++)
	    st[i] = (Tree)spanningtree.getObject(i);
	return st;
    }
    /*
    public void removeVertex(Vertex v) {
	Note.err(this, "Method not implemented");
    }

    public void addVertex(Object v) {
	Note.err(this, "Method not implemented");
    }
    */
    //--- End Tree and SimulationTree(2) ---

    //--- Start LabelDecorator ---

    public String getLabel() { 
	return label.getObject().toString().trim(); 
    }
    public void setLabel(String s) {  label.setObject(s); }
    public boolean isLabelEnabled() { return true; }
    public void setReferenceLabel(int i, String s) { }
    public String getReferenceLabel(int i) {
	if (i<children.eval())
	    return ""+getWeight((PriorityExerVertex)spanningtree.getObject(i));
	else
	    return "?";
    }
    public boolean isReferenceLabelEnabled() {
        return false;
    }

    //--- End LabelDecorator ---

    //--- Start Public Interface ---

    public void setPriority(int p) {
	priority = p;
	time_stamp = global_time++;
    }

    public int getPriority() {
	return priority;
    }

    public int getTimeStamp() {
	return time_stamp;
    }

    public void setVisited(boolean f) {
	markVisited(f);
	visited.assign(f);
	if (!f) {
	    spanningtree = new VirtualArray();
	    children = new VirtualInteger(0);
	    father = new VirtualObject(null);
	    label = new VirtualObject("");
	    edgeColor = new VirtualArray();
	}
    }

    public boolean getVisited() {
	return visited.eval();
    }

    public void setFather(PriorityExerVertex v) {
	father.setObject(v);
    }
    
    public PriorityExerVertex getFather() {
	return (PriorityExerVertex)father.getObject();
    }
    
    public PriorityExerVertex[] getAdjacentVertices() {
	if (succ == null) {
	    Vertex[] s = super.getSuccessors();
	    succ = new PriorityExerVertex[s.length];
	    for (int i=0; i<s.length; i++)
		succ[i] = (PriorityExerVertex)s[i];
	}
	return succ;
    }
    
    public void setWeight(PriorityExerVertex v, int w) {
	weights.put(v, new Integer(w));
    }

    public int getWeight(PriorityExerVertex v) {
	Integer I = (Integer)weights.get(v);
	if (I == null) {
	    Note.warning(this, "Weight for edge " + this + " to " + v + " not found");
	    return Integer.MAX_VALUE;
	}
	return I.intValue();
    }
    
    //--- End Public Interface ---

    //--- Methods overriden by this class ----

    private void newSuccessor(PriorityExerVertex v, int pos) {
	//Note.out(this, "new successor for " + v);
	addSuccessor(v); // as the last one
	Vertex[] successors = getSuccessors();
	int p = successors.length-1; // thus, it should be here
	if (successors[p] != v) {
	    Note.warning(this, "newSuccessor failed: Incorrect location");
	    return;
	}
	for ( ; p>pos ; p--) // shift to make room for v
	    successors[p] = successors[p-1];
	successors[p] = v;
	//Note.out(this, "set successors");
	setSuccessors(successors);
	//Note.out(this, "set done.");
    }

    public void insertSuccessor(PriorityExerVertex v, int weight) {
	Vertex[] u = getSuccessors();
	int i=0;
	for (; i<u.length; i++) 
	    if (v.getElement().toString().trim().compareTo(u[i].getElement().toString().trim()) < 0)
		break;
	//Note.out(this, "element " + this + " insert " + v + " at " + i);
	newSuccessor(v, i);
	setWeight(v, weight);
    }

    public String toString() {
	return "<"+getElement()+">";
    }

    private boolean equals(Tree[] t1, Tree[] t2) {
	if (t1.length != t2.length)
	    return false;
	for (int i=0; i<t1.length; i++) {
	    if ((t1[i] == null) && t2[i] == null) continue;
	    if (t1[i] == null) return false;
	    //Note.out(this, "i="+i+"; t1[i]="+t1[i]+"; t2[i]="+t2[i]);
	    if (!t1[i].equals(t2[i]))
		return false;
	}
	return true;
    }

    // for debug
    private String print(Tree[] t) {
	String s = "";
	for (int i=0; i<t.length; i++)
	    s+=t[i]!=null?t[i].toString():"-";
	return s;
    }

    public boolean equals(PriorityExerVertex v) {
	boolean b = getElement().equals(v.getElement());
	Tree[] myChild = this.getSubTrees();
	Tree[] hisChild = v.getSubTrees();
	//Note.out(this, "b="+b+"; compare ("+this.toString()+"): " + print(myChild) + " to " + print(hisChild));
	return b && equals(myChild, hisChild);
    }

    public boolean equals(Object v) {
	if (v instanceof PriorityExerVertex)
	    return equals((PriorityExerVertex)v);
	else
	    return super.equals(v);
    }

}
