/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualAnimator.java,v 1.26 2004/05/26 08:05:48 vkaravir Exp $ */

package matrix.animation;

import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.InvocationTargetException;

/** Implementation for Animator Window. */
public class VisualAnimator extends Panel {
    private Button undo, redo, begin, end, play, resetBeg;
    protected Scrollbar sb;
    protected Animator animator;
    private boolean buttonsLocked = false;

    public VisualAnimator(Animator a) {
        this.animator = a;
	layoutComponents();
    }

    protected void layoutComponents() {
        setLayout(new BorderLayout(100, 10));
        Panel p = new Panel();

        undo = new Button("Backward");
        redo = new Button("Forward");
        begin = new Button("Begin");
        end = new Button("End");
        play = new Button("Play");
        resetBeg = new Button("Reset beginning here");

        p.add(undo);
        p.add(redo);
        p.add(begin);
        p.add(end);
        p.add(play);

        this.add(p, "North");
        this.add(sb = new Scrollbar(Scrollbar.HORIZONTAL), "Center");
        this.add(resetBeg, "South");

        sb.addAdjustmentListener(new AdjustmentListener() {
                public void adjustmentValueChanged(AdjustmentEvent e) {
                    GoTo(e.getValue(), sb.getMaximum());
                    fireActionEvent(new ActionEvent(this, 0, "adjust"));
		    //focus must be transfered to get the shortcut keys for
		    //animator to work
		    sb.transferFocus();
                }
            });

        undo.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    undo();
                    fireActionEvent(e);
                }
            });

        redo.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    redo();
                    fireActionEvent(e);
                }
            });

        begin.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    rewind();
                    fireActionEvent(e);
                }
            });

        end.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                end();
                fireActionEvent(e);
            }
            });

	/* After "Play" is pressed, repeat:
	   - Wait 500 ms.
	   - Step forward in animation.
	   until the end of the animation is reached.

	   The step events are put on the AWT event queue to avoid
	   synchronisation problems between successive steps.
	*/

	play.addActionListener(new ActionListener() {
		public void actionPerformed(final ActionEvent e) {
		    new Thread() {
			public void run() {
			    while (animator.hasNextOperation()) {
				try {
				    Thread.currentThread().sleep(500);
				    try {
					EventQueue.invokeAndWait(
					    new Runnable() {
						public void run() {
						    animator.redo();
						    fireActionEvent(
							(ActionEvent)e);
						}
					    });
				    } catch(InvocationTargetException ite) {
					ite.getTargetException().
					    printStackTrace();
				    }
				} catch(InterruptedException ie) {}
			    }
			}
		    }.start();
		}
	    });

        resetBeg.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    animator.disposeUndo();
                    fireActionEvent(e);
                }
	    });
	setResetButtonVisible(false);
        validate();
    }

    public void actionPerformed(ActionEvent event) {
        getParent().validate();
    }


    public void setResetButtonVisible(boolean value) {
        resetBeg.setVisible(value);
    }


    public void setPlayButtonVisible(boolean value) {
        play.setVisible(value);
    }


    protected void undo() {
        if (animator.hasPreviousOperation()) {
            animator.undo();
        }
    }


    protected void redo() {
        if (animator.hasNextOperation()) {
            animator.redo();
        }
    }


    protected void rewind() {
        animator.rewind();
    }


    protected void end() {
        while (animator.hasNextOperation())
            animator.redo();
    }

    public void setButtonsLocked(boolean t) {
	buttonsLocked=t;
    }

    public void validate() {
        super.validate();
	if (buttonsLocked || play == null || redo == null || end == null || undo == null || begin == null)
	    return;
        if (animator.hasNextOperation()) {
            play.setEnabled(true);
            redo.setEnabled(true);
            end.setEnabled(true);
        } else {
            play.setEnabled(false);
            redo.setEnabled(false);
            end.setEnabled(false);
        }

        if (animator.hasPreviousOperation()) {
            undo.setEnabled(true);
            begin.setEnabled(true);
        } else {
            undo.setEnabled(false);
            begin.setEnabled(false);
        }
        validateScrollbar();
    }


    protected void validateScrollbar() {
        sb.setValues(animator.getState(), 1, 0, animator.getMaximum());
    }


    protected void GoTo(int index, int maximum) {
        animator.GoTo(index, maximum);
    }

    // AWT event handling
    protected ActionListener listener = null;

    /** Add new listener. */
    public void addActionListener(ActionListener a) {
        listener = AWTEventMulticaster.add(listener, a);
        //animator.addActionListener(a);
    }

    /** Remove listener. */
    public void removeActionListener(ActionListener a) {
        listener = AWTEventMulticaster.remove(listener, a);
        //animator.removeActionListener(a);
    }

    /** This method is invoked when user action takes place
     ** in this VisualAnimator.
     * All registered listeners are notified that they might be changed.
     */
    protected void fireActionEvent(ActionEvent e) {
      if (listener != null) {
        //Note.out(this, "firing action to listener " + listener);
          listener.actionPerformed(e);
      }
    }
    // -------------------

    static final long serialVersionUID = -11601511645348702L;
}

