/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: ColorDecorator.java,v 1.7 2004/05/26 08:05:48 vkaravir Exp $ */

package matrix.decoration;

import java.awt.Color;

import matrix.structures.FDT.FDT;

/** 
 * The interface ColorDecorator is a hint for the visualizer to recognize 
 * that this particular node should be decorated by coloring its nodes
 * and/or references. This inteface contains a large number of methods, so
 * it is recommended that instead of implementing this interface, you should
 * extend the <code>StyleSheetAdapter</code> -class of stylesheet package. 
 * <code>StyleSheetAdapter</code> is the default implementation of the 
 * <code>StyleSheet</code> interface, which extends <code>ColorDecorator</code>.
 *
 * The colors in this interface have been divided to three different categories, each
 * of which has four colors in it.<p>
 * <TABLE border="0"> 
 * <TR><TH>The three categories</TH></TR>
 * <TR><TD>Background colors.</TD><TD>Background colors control how the contents of the
 * element (the background over which all visualised data is placed) are colored.</TD></TR>
 * <TR><TD>Border colors.</TD><TD>The border colors control how the borders of an element
 * (including references going from an element to another) are colored.</TD></TR>
 * <TR><TD>Pen colors.</TD><TD>The pen colors control how anything drawn in an element
 * (title, label, etc) are colored. These are the colors that are used to draw the
 * contents of elements.</TD></TR>
 * </TABLE>
 * <TABLE border="0">
 * <TR><TH>The four colors</HT></TR>
 * <TR><TD>These colors are in each of the three categories.</TD></TR>
 * <TR><TD>Default color.</TD><TD>Default color is the color that is used
 * to paint an element by default, when there are no special conditions that
 * required the use of any other color.</TD></TR>
 * <TR><TD>Focused color.</TD><TD>Focused color is used when the mouse pointer
 * is over an element, or when the element is "on focus" in some other manner.</TD></TR>
 * <TR><TD>Visited color.</TD><TD>Visited color is used when an element has been
 * visited by an algorithm, when a reference has been moved, or an element has been modified
 * in some other way that Matrix can visualize.</TD></TR>
 * <TR><TD>Visited focused color.</TD><TD>Visited focused color is used when the focus is on
 * a visited element.</TD></TR></TABLE>
 * 
 * 
 * 
 *
 * @author Ari Korhonen, Jussi Nikander
 * @version $Revision: 1.7 $
 */

public interface ColorDecorator {

    /** Sets default background color.
     * @param c the new color
     */
    public void setDefaultBackgroundColor(Color c);
    /**
     * Sets focused background color.
     * @param c the new color
     */
    public void setFocusedBackgroundColor(Color c);
    /**
     * Sets visited background color.
     * @param c the new color
     */
    public void setVisitedBackgroundColor(Color c);

    /**
     * Sets visited focused background color.
     * @param c the new color
     */
    public void setVisitedFocusedBackgroundColor(Color c);

    /**
     * Gets default background color.
     * @return the color
     */
    public Color getDefaultBackgroundColor();

    /**
     * Gets focused background color.
     * @return the color
     */
    public Color getFocusedBackgroundColor();

    /**
     * Gets visited background color.
     * @return the color
     */
    public Color getVisitedBackgroundColor();

    /**
     * Gets visited focused background color.
     * @return the color
     */
    public Color getVisitedFocusedBackgroundColor();
    
    // Colors the paint the border of the element.
    
    /**
     * Sets default border color
     * @param c the new color
     */
    public void setDefaultBorderColor(Color c);

    /**
     * Sets focused border color
     * @param c the new color
     */
    public void setFocusedBorderColor(Color c);

    /**
     * Sets visited border color
     * @param c the new color
     */
    public void setVisitedBorderColor(Color c);

    /**
     * Sets visited focused border color
     * @param c the new color
     */
    public void setVisitedFocusedBorderColor(Color c);
    
    /**
     * gets default border color
     * @return the color
     */
    public Color getDefaultBorderColor();

    /**
     * gets focused border color
     * @return the color
     */
    public Color getFocusedBorderColor();

    /**
     * gets visited border color
     * @return the color
     */
    public Color getVisitedBorderColor();

    /**
     * gets visited focused border color
     * @return the color
     */
    public Color getVisitedFocusedBorderColor();
    
    // Colors that paint whatever is written or drawn inside the element
    // (ie keyvalue, title, etc)
    
    /** 
     * Sets default pen color.
     * @param c the new color
     */
    public void setDefaultPenColor(Color c);
    /** 
     * Sets focused pen color.
     * @param c the new color
     */
    public void setFocusedPenColor(Color c);
    /** 
     * Sets visited pen color.
     * @param c the new color
     */
    public void setVisitedPenColor(Color c);
    /** 
     * Sets visited focused pen color.
     * @param c the new color
     */
    public void setVisitedFocusedPenColor(Color c);
    
    /** 
     * Gets default pen color.
     * @return the color
     */
    public Color getDefaultPenColor();

    /** 
     * Gets focused pen color.
     * @return the color
     */
    public Color getFocusedPenColor();

    /** 
     * Gets visited pen color.
     * @return the color
     */
    public Color getVisitedPenColor();

    /** 
     * Gets visited focused pen color.
     * @return the color
     */
    public Color getVisitedFocusedPenColor();

  /**
   * Sets color for the reference from 
   * this to the parameter FDT.
   *
   * @param f the FDT the reference points to
   * @param c the new color of the reference
   */
  public void setReferenceColor(FDT f, Color c);

  /**
   * Sets color for focused reference.
   * 
   * @param f the FDT the reference points to
   * @param c the new color
   */
  public void setFocusedReferenceColor(FDT f, Color c);

  /**
   * Sets color for modified reference
   *
   * @param f the FDT the reference points to
   * @param c the new color
   */
  public void setModifiedReferenceColor(FDT f, Color c);

  /**
   * Gets color of the reference from this to
   * the parameter FDT.
   *
   * @param f the destination of the reference
   * @return the color
   */
  public Color getReferenceColor(FDT f);

  /**
   * Gets color of focused reference
   * 
   * @param f the FDT the reference points to
   * @return the color
   */
  public Color getFocusedReferenceColor(FDT f);

  /**
   * Gets color of modified reference
   *
   * @param f the FDT the reference points to
   * @return the color
   */
  public Color getModifiedReferenceColor(FDT f);

}




