/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package matrix.structures.CDT.probe;

import java.util.Stack;

import matrix.decoration.*;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.other.Rotations;
import matrix.util.Note;

/**
 * Labeled AVL-tree with node's height in node's label
 *
 * @author Petri Tenhunen
 */

public class AVLTree extends BinSearchTree {

  protected class AVLNode extends BinTree implements Styled {

    protected VirtualInteger height = new VirtualInteger(0);
    protected StyleSheet myStyle = null;

    public AVLNode() {
      super();
    }

    public AVLNode(Element e) {
      super(e);
    }

    public StyleSheet getStyleSheet() {
      if (myStyle != null)
        return myStyle;
      else {
        myStyle = new StyleSheetAdapter() {
                public boolean isInfoEnabled() { return false; }
                //public String getInfo() { return "hui hai ja tm on sitten toivottavasti niin"+
                //	      "pitk juttu, ettei se mahdukkaan yhdele riville niin vain"; }
          public boolean isLabelEnabled() { return !empty(); }
          public String getLabel() { return ""+getHeight(); }
        };
        return myStyle;
      }
    }

    public int getHeight() {
      return height.eval();
    }

    public void setHeight(int value) {
      height.assign(value);
    }

    public boolean empty() { return getElement() == null; }

    protected AVLNode AVLInsert(Element e) {
      BinaryTree prev;
      BinaryTree node = this;
      Stack route = new Stack();

      do {
        prev = node;
        route.push(prev);
        if (((Element)node.getElement()).gt(e))
          node = node.getLeft();
        else
          node = node.getRight();
      } while (node != null);
      node = new AVLNode(e);
      if (((Element)prev.getElement()).gt(e)) {
        prev.setLeft(node);
      }else {
        prev.setRight(node);
      }
      return this;
    }

  }

  public AVLTree() {
    super();
  }

  public AVLTree(Element e) {
    super(e);
  }

  public FDT getNewStructure() {
    return new AVLNode() ;
  }

  public Tree getNewNode(Object key) {
    return new AVLNode((Element)key);
  }

  // Methods for interface CDT

  public FDT insert(Object o) {
      //Note.out(this, "INSERT" + o);

    if (o == null) {
	return this;
    } else if (o instanceof Element) {
      if (empty()) {
        setElement((Element)o);
      } else
        setElement(AVLInsert((AVLNode)getElement(), (Element)o));

    } else if (o instanceof Table) {
      Table t = (Table)o;
      for (int i=0; i < t.size(); i++)
          insert(t.getObject(i));
    } else {
      return insert(new Key(o));
    }
    return this;
  }


  public FDT delete(Object o) {
    Note.show(this, "Cannot delete, Delete is not implemented");
    return this;
  }

  protected  AVLNode AVLInsert(AVLNode t, Element e) {
    if (t == null || t.getElement() == null) {
      t = new AVLNode();
      t.setHeight(0);
      t.setElement(e);
    } else if (((Element)t.getElement()).gt(e)) {
      t.setLeft(AVLInsert((AVLNode)t.getLeft(), e));
      if (getHeight((AVLNode)t.getLeft()) - getHeight((AVLNode)t.getRight()) == 2) {
        if (((Element)t.getLeft().getElement()).gt(e))
          t = (AVLNode)Rotations.withLeftChild(t);
        else
          t = (AVLNode)Rotations.doubleWithLeftChild(t);
      }

    } else {
      t.setRight(AVLInsert((AVLNode)t.getRight(), e));
      if (getHeight((AVLNode)t.getRight()) - getHeight((AVLNode)t.getLeft()) == 2) {
        if (!((Element)t.getRight().getElement()).gt(e))
            t = (AVLNode)Rotations.withRightChild(t);
        else
          t =(AVLNode) Rotations.doubleWithRightChild(t);
      }
    }

    int h_right = evalHeight((AVLNode)t.getRight());
    int h_left = evalHeight((AVLNode)t.getLeft());

    evalHeight(t);
    return t;
  }

  protected static int evalHeight(AVLNode t) {
    if (t ==null) return -1;
    if (t.getLeft()==null && t.getRight()==null) {
      t.setHeight(0);
      return 0;
    }
    int h = Math.max( evalHeight((AVLNode)t.getLeft()),
                      evalHeight((AVLNode)t.getRight()) ) + 1;
    t.setHeight(h);
    return h;

  }

  public static int getHeight(AVLNode t) {
    return (t == null) ? -1 : t.getHeight();
  }


  static final long serialVersionUID = -6160673953183778860L;
}
