/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: QueueImpl.java,v 1.4 2004/05/26 08:05:50 vkaravir Exp $ */

package matrix.structures.CDT.probe;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;

/**
 * Implementation of a queue. This queue is implemented with doubly linked list.
 * Thus, this queue is a collection of nodes that contains elements of the queue.
 * Every node contains also pointers to the previous node and the next node.
 * The node which pointer to previous node points to null is a first node of 
 * the queue.
 *
 */
public class QueueImpl extends LinkedListImpl implements CDT, DoublyLinkedList {
    

    /** previous node in the stack */
    private DoublyLinkedList prev;


    /**
     * Constructs QueueImpl.
     *
     */
    public QueueImpl() {
	super();
    }

    /**
     * Constructs QueueImpl with given key.
     *
     * @param key the key of the new QueueImpl
     */
    public QueueImpl(Object key) {
	super();
	setElement(key);
    }

    /**
     * Returns new (Queue) node with a given key. 
     *
     * @param key key of new node
     * @return new (Queue) Node
     */    
    public LinkedList getNewNode(Object key) {
	return new QueueImpl(key);
	}


    /**
     * Inserts a new key to the last position of the queue. This insert operation just calls 
     * the enqueue operation.
     *
     * @param key new key
     * @return queue after insertion
     */
    public FDT insert(Object key) {
	return enqueue(key);
    }


    /**
     * Enqueues a new key to the last position of the queue.
     *
     * @param key new key
     * @return queue after insertion of the key
     */
   public FDT enqueue(Object key) {
	if (key == null)
	    return getFirstOfQueue();
	if (key instanceof Key)
		if (((Key)key).toString() == "")
			return getFirstOfQueue();
	if (key instanceof Element) {
	    QueueImpl newNode = new QueueImpl();
	    newNode.setElement(key);
	    setLastOfQueue(newNode);
	    return getFirstOfQueue();
	}
	if (key instanceof Table) {
	    Table t = (Table)key;
	    for(int i = 0; i < t.size(); i++)
		insert(t.getObject(i));
	    return getFirstOfQueue();
	}
	Key k = new Key(key);
	QueueImpl newNode = new QueueImpl();
	newNode.setElement(key);
	setLastOfQueue(newNode);
	return getFirstOfQueue();
    }


    /**
     * Sets the given node to the last position of the queue.
     *
     * @param newNode new (Queue) node to be added to the last position of 
     * the queue
     */
    protected void setLastOfQueue(QueueImpl newNode) {
	if (getElement() == null && hasNext() == false) {
	    setElement(newNode.getElement());
	}
	else if (hasNext() == false) {
	    newNode.setPrev(this);	    
	    setNext(newNode);
	}
	else {
	    ((QueueImpl)getNext()).setLastOfQueue(newNode);
	}
    }

    /**
     * Returns the first node of the queue
     *
     * @return node from the first of the queue
     */
    protected QueueImpl getFirstOfQueue() {
	if (hasPrev() == false){
	    return this;}
	else
	    return ((QueueImpl)prev).getFirstOfQueue();
    }


    /**
     * This method deletes the node containing the given key and all
     * its predecessors from the queue and returns 
     * the first node of queue after deletion.
     *
     * @param o key to be deleted
     * @return returns always the first node of the queue after 
     * deletion.
     */
    public FDT delete(Object o) {
	boolean isInStack = false;
	QueueImpl temp = this;
	while(temp != null) {
	    if(o instanceof Key) {
		if(temp.getElement() == ((Key)o).getOrig())
		    isInStack = true;
		else if(temp.getElement() == o)
			isInStack = true;
	    }
	    else {
		if(temp.getElement() == o)
		    isInStack = true;
	    }
	    temp = (QueueImpl)temp.getNext();
	}
	if(isInStack)
	    return dequeue(o);
	return getFirstOfQueue();
    }


    /**
     * This method dequeues the node containing the given key and all its
     * predecessors from the queue and returns the first
     * node of the queue after dequeue operation.
     *
     * @return first node of the queue after dequeue operation
     */
   private FDT dequeue(Object o) {
       if (hasPrev() == false && getNext() == null) { //only one node in the queue
	    setElement(null);
	    return this;
       }
       else if (getPrev() == null) { //this node is the first node 
	   QueueImpl temp = (QueueImpl)getNext();
	   setNext(null);
	   temp.setPrev(null);
	   //return temp;
	   if(o instanceof Key) {
	       if(getElement() == ((Key)o).getOrig())
		   	return temp;
		   else if(getElement() == o)
		   	return temp;
	       else
		   return temp.delete(o);
	   }
	   else {
	       if(getElement() == o)
		   return temp;
	       else
		   return temp.delete(o);
	   }
       }
       else {
	   //return ((QueueImpl)getPrev()).dequeue();
	   return ((QueueImpl)getPrev()).delete(o);
       }
   }
    

    /**
     * Sets the previous node.
     *
     * @param o previous node
     */
    public void setPrev(DoublyLinkedList o) {
	prev = o;
    }


    /**
     * Returns the previous node.
     *
     * @return the previous node
     */
    public DoublyLinkedList getPrev() {
	return prev;
    }


    /**
     * Tests whether there is a previous node.
     *
     * @return true, if there is a previous node
     */
    public boolean hasPrev() {
	return (getPrev() != null);
    } 


    public Object search(Object o) {
	return first();
    }

    /**
     * Returns the first node of the queue.
     *
     * @return first node of the queue
     */
    public FDT first() {
	return getFirstOfQueue();
    }

    static final long serialVersionUID = -6085331632609828008L;
}
