/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.structures.CDT.probe;

import matrix.structures.FDT.*;
import matrix.structures.memory.*;
import matrix.util.Note;

/**
 * <p>Implementation of the RedBlackTree. This is implemented by encapsulating
 * <code>RedBlackNodes</code> inside this <code>RBTree</code>.
 * RebBlackTree is balanced binary search tree with following properties:
 * <ul>
 * <li> Root is black.</li>
 * <li> Every route from root to leaf contains same number of black nodes. </li>
 * <li> Red node can't have red childrens.</li>
 * </ul></p>
 *
 * <p>Every node has parent information so when using
 * methods inherited from <code>BinSearchTree</code> parent information may also need
 * some updating. <code>Rotations</code> class can be used with this class if parents
 * fixing method is called before rotation. These methods are:
 * <ul>
 * <li><code>newParentsZigRight()</code></li>
 * <li><code>newParentsZigLeft()</code></li>
 * <li><code>newParentsZigZagRight()</code></li>
 * <li><code> newParentsZigZagLeft()</code></li>
 * </ul></p>
 *
 * @see matrix.decoration.Styled
 * @see matrix.structures.CDT.probe.BinSearchTree
 * @see matrix.structures.other.Rotations
 * @see RedBlackNode
 *
 * @author Petri Tenhunen
 */
public class RBTree extends BinSearchTree {

  private VirtualObject root;

  /** Constructs a new RBTree */
  public RBTree() {
    this(null);
  }

  /**
   * Constructs a new RBTree
   * @param o the key to the root node
   */
  public RBTree(Object o) {
    super(o);
  }

  public FDT getNewStructure() {
    return new RedBlackNode() ;
  }

  public Tree getNewNode(Object key) {
    return new RedBlackNode((Element)key);
  }

  // Methods for interface CDT

  public FDT insert(Element o) {
      setElement(((RedBlackNode)getElement()).insert(o));
      return this;
  }

  public FDT insert(Object o) {
      if (o instanceof Element)
          return insert((Element)o);
      setElement(((RedBlackNode)getElement()).insert(new Key(o)));
      return this;
  }

  public FDT delete(Object o) {
    Note.show(this, "Cannot delete, Delete is not implemented");
    return this;
  }

  public boolean equals(Object o) {
      if (o instanceof RBTree)
          return getElement().equals(((RBTree)o).getElement());
      else return super.equals(o);
  }



  static final long serialVersionUID = 2860991368734163098L;
}
