/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package matrix.structures.CDT.probe;
import java.awt.Color;

import matrix.decoration.*;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;
import matrix.structures.other.Rotations;
import matrix.util.Note;

/**
 * <p>Implementation of the RedBlackNode in RedBlackTree.
 *
 * @see RBTree
 * @author Petri Tenhunen
 */

public class RedBlackNode extends BinTree implements Styled {

    private VirtualInteger red = new VirtualInteger();
    private VirtualArray parent;

    private StyleSheet myStyle = null;

    /** default constructor for empty tree */
    public RedBlackNode() {
        super();
        parent = new VirtualArray();
        red.assign(1);
    }


    /** Constructor for tree with one element */
    public RedBlackNode(Element e) {
        super(e);
        parent = new VirtualArray();
        red.assign(1);
    }

    /**
     * Tells if a tree is empty or not
     * @return true if the tree is empty, false otherwise
     */
    public boolean empty() {
        return getElement() == null;
    }


    public StyleSheet getStyleSheet() {
        if (myStyle != null)
            return myStyle;
        else {
            myStyle = new StyleSheetAdapter() {
                    public Color getDefaultBorderColor() {
                        return empty() || !isRed() ? Color.black: Color.red;
                    }
                    public Color getDefaultBackgroundColor() {
                        return isRed() ? Color.red : Color.black;
                    }
                    public Color getFocusedBackgroundColor() {
                        return empty() || !isRed() ? Color.black: Color.red;
                    }
                    public Color getFocusedBorderColor() {
                        return empty() || !isRed() ? Color.red : Color.black;
                    }
                };
            return myStyle;
        }
    }

    public FDT insert(Object o) {

        Note.out(this, "INSERT "+o+" to "+this);

        if (o == null)
            return this;
        if (o instanceof Table) {         //recursive insertion to Tables
            Table t = (Table)o;
            RedBlackNode r = this; //to store new roots
            for (int i=0; i<t.size(); i++)
                r = (RedBlackNode)r.insert(t.getObject(i));
            return r;
        } else if (o instanceof Element) {
            if (empty()) {                       //inserted element is only element
                setElement((Element)o);
                this.setRed(false);
                return this;
            }
            else {
                /**
                BinaryTree node = new RedBlackNode((Element)o);
                node.setRight(this);
                return node;
                */
                BinaryTree node=this, prev=this;
                do {
                    prev = node;
                    if (((Element)node.getElement()).gt((Element)o))
                        node = node.getLeft();
                    else
                        node = node.getRight();
                } while (node != null);  //new leaf will now be prev's child
                node = new RedBlackNode((Element)o);
                ((RedBlackNode)node).setParent((RedBlackNode)prev);
                if (((Element)prev.getElement()).gt((Element)o)) {
                    prev.setLeft(node);
                }else {
                    prev.setRight(node);
                }
                //((RedBlackNode)node).head = this.head;

                return rotateUp((RedBlackNode)node);
            }
        } else Note.show(this, "Cannot insert object of type " + o.getClass());
        return this;

    }


    /**
     * Rotations for bottom up insertion. Normal binary search tree
     * insertion must applied first. This method is called from
     * <code>public FDT insert(Object o)</code>.
     *
     * @param x newly inserted leaf before balansing.
     * @return Root of the RedBlackNode after all rotations are performed.
     */
    public  RedBlackNode rotateUp(RedBlackNode x) {
        x.setRed(true);
        Note.out(this,"Starting rotations");
        boolean contin = true;
        while ( x.getParent() != null && x.getParent().isRed() && x.isRed() ) {  //(not root) and red
            Note.out(this,"in loop");
            Note.out(this,"x: "+x);
            if (  x.getParent() == x.getParent().getParent().getLeft() ) { //K1
                Note.out(this,"left branch");
                RedBlackNode y = (RedBlackNode)x.getParent().getParent().getRight();
                RedBlackNode pp = x.getParent().getParent();
                RedBlackNode ppp = pp.getParent(); //parent of rotated subtree
                RedBlackNode tmp;

                if ( y != null && y.isRed()) { //only swap colors
                    Note.out(this,"color swap");
                    x.getParent().setRed(false);
                    y.setRed(false);
                    x.getParent().getParent().setRed(true);
                    x = x.getParent().getParent();


                } else if ( x == x.getParent().getRight()) {     //zig-zag
                    Note.out(this,"zig-zag");
                    x.setRed(false);
                    x.getParent().getParent().setRed(true);
                    x.newParentsZigZagRight();
                    tmp = (RedBlackNode)Rotations.doubleWithLeftChild(pp);
                    if (ppp == null) {
                        return tmp;
                    } else if (ppp.getLeft() == pp) {
                        ppp.setLeft(tmp);
                    } else {
                        ppp.setRight(tmp);
                    }

                } else {
                    Note.out(this,"zig");
                    x.getParent().getParent().setRed(true);
                    x.getParent().setRed(false);
                    x.newParentsZigRight();
                    tmp = (RedBlackNode) Rotations.withLeftChild(pp);

                    if (ppp == null) {
                        tmp.setRed(false);
                        tmp.setParent(null);
                        return tmp;
                    } else if (ppp.getRight() == pp) {
                        ppp.setRight(tmp);
                    } else {
                        ppp.setLeft(tmp);
                    }
                    tmp.setParent(ppp);
                }

            } else {
                Note.out(this,"right branch");
                RedBlackNode y = (RedBlackNode)x.getParent().getParent().getLeft();
                RedBlackNode pp = x.getParent().getParent();
                RedBlackNode ppp = pp.getParent(); //parent of rotated subtree
                RedBlackNode tmp;

                if ( y != null && y.isRed()) { //only swap colors
                    Note.out(this,"color swap");
                    x.getParent().setRed(false);
                    y.setRed(false);
                    x.getParent().getParent().setRed(true);
                    x = x.getParent().getParent();


                } else if ( x == x.getParent().getLeft()) {     //zig-zag
                    Note.out(this,"zig-zag");
                    x.setRed(false);
                    x.getParent().getParent().setRed(true);
                    x.newParentsZigZagLeft();
                    tmp = (RedBlackNode)Rotations.doubleWithRightChild(pp);
                    if (ppp == null) {
                        tmp.setParent(null);
                        return tmp;
                    } else if (ppp.getLeft() == pp) {
                        ppp.setLeft(tmp);
                    } else {
                        ppp.setRight(tmp);
                    }

                } else {
                    Note.out(this,"zig");
                    x.getParent().getParent().setRed(true);
                    x.getParent().setRed(false);
                    x.newParentsZigLeft();
                    tmp = (RedBlackNode) Rotations.withRightChild(pp);
                    if (ppp == null) {
                        tmp.setRed(false);
                        tmp.setParent(null);
                        return tmp;
                    } else if (ppp.getRight() == pp) {
                        ppp.setRight(tmp);
                    }else {
                        ppp.setLeft(tmp);
                    }
                    tmp.setParent(ppp);

                }
            }

        } // end while (x is not root and red)
        Note.out(this,"new root is "+getRoot().getElement().toString());
        x.getRoot().setRed(false);    //set root to black
        return x.getRoot();
    }

    /**
     * This method is called before right-rotation is performed. If rotation is performed
     * to node <code>n</code> the method is called to node <code>n.getLeft()<code>
     */
    protected void newParentsZigRight() {
        this.getParent().getParent().setParent(this.getParent());
        if (this.getParent().getRight() != null)
            ((RedBlackNode)this.getParent().getRight()).setParent(this.getParent().getParent());
        this.getParent().setParent(this.getParent().getParent().getParent());
    }

    /**
     * This method is called before left-rotation is performed. If rotation is performed
     * to node <code>n</code> the method is called to node <code>n.getRight()<code>
     */
    protected void newParentsZigLeft() {
        this.getParent().getParent().setParent(this.getParent());
        if (this.getParent().getLeft() != null)
            ((RedBlackNode)this.getParent().getLeft()).setParent(this.getParent().getParent());
        this.getParent().setParent(this.getParent().getParent().getParent());
    }

    /**
     * This method is called before double-rotation to right is performed.
     * If rotation is performed to node <code>n</code> the method is called to node
     * <code>n.getLeftt().getRight()<code>
     */
    protected void newParentsZigZagRight() {
        RedBlackNode a = this.getParent();
        RedBlackNode b = a.getParent();
        if (this.getRight() != null)
            ((RedBlackNode)this.getRight()).setParent(b);
        if (this.getLeft() != null)
            ((RedBlackNode)this.getLeft()).setParent(a);
        this.setParent(b.getParent());
        a.setParent(this);
        b.setParent(this);
    }

    /**
     * This method is called before double-rotation to left is performed.
     * If rotation is performed to node <code>n</code> the method is called to node
     * <code>n.getRight().getLeft()<code>
     */
    protected void  newParentsZigZagLeft() {
        RedBlackNode a = this.getParent();
        RedBlackNode b = a.getParent();
        if (this.getRight() != null)
            ((RedBlackNode)this.getRight()).setParent(a);
        if (this.getLeft() != null)
            ((RedBlackNode)this.getLeft()).setParent(b);
        this.setParent(b.getParent());
        a.setParent(this);
        b.setParent(this);
    }

    /** debug */
    private String printAllParents() {
        String tmp = getElement() +" : "+ getParent() +"\n";
        if (getRight() != null)
            tmp += ((RedBlackNode)getRight()).printAllParents();
        if (getLeft() != null)
            tmp += ((RedBlackNode)getLeft()).printAllParents();
        return tmp;
    }




    /**
     * Returns root node of this redBlack-tree.
     *
     * @return root node from tree where current node is acting as subtree
     */
    public RedBlackNode getRoot() {
        RedBlackNode root = this;
        //return (RedBlackNode)head.getRight();
        while(root.getParent() != null)
            root = root.getParent();
        return root;
    }


    /**
     * Gets father of this subtree
     *
     * @return parent of this subtree or null, if this is the root.
     **/
    public RedBlackNode getParent() {
        return (RedBlackNode)this.parent.getObject(0);
    }


    /**
     * Set father information.
     *
     * @param p father for this subtree
     */
    public void setParent(RedBlackNode p) {
        this.parent.setObject(p,0);
    }


    /**
     * Set color information.
     *
     * @param red If true color is set to red. If false the color is
     *            set to black
     */
    public void setRed(boolean red) {
        if (red)
            this.red.assign(1);
        else
            this.red.assign(0);
    }


    /**
     * Color of the node.
     *
     * @return true if node is red
     */
    public boolean isRed() { return red.eval()==1; }


    public String toString() {
        return empty() ? "empty" : getElement().toString() + ", isRed: "+isRed(); }


    public boolean equals(Object o) {
        //System.out.println(""+this+" joo "+o);
      //Note.out(this,"equals");
        if (o instanceof RedBlackNode)
            return (compareIdentical(this, (RedBlackNode)o));
        return false;
    }

    protected static boolean compareIdentical(Tree a, Tree b) {

        if(a == b) {
          //Note.out(null,"case 1");
          return true;
        }
        if(a == null || b == null) return false;
        Object aObj = a.getElement();
        boolean aRed = ((RedBlackNode)a).isRed();
        Object bObj = b.getElement();
        boolean bRed = ((RedBlackNode)b).isRed();
        if(aObj == bObj && aObj == null) {
          //Note.out(null,"case 2");
          return true;
        }
        if(aObj == null || bObj == null) {
          //Note.out(null,"case 2");
            return false;
        }
        if(((!(aObj.equals(bObj))) && (!(bObj.equals(aObj)))) || (aRed != bRed) )
            return false;
        Tree[] aTree = a.getSubTrees();
        Tree[] bTree = b.getSubTrees();
        if(aTree.length != bTree.length)
            return false;
        for(int i = 0; i < aTree.length; i++)
            if(!compareIdentical(aTree[i], bTree[i]))
                return false;
        return true;
    }

    private boolean compareChilds(RedBlackNode other) {
        boolean right = false, left = false;

        if ( getRight() != null )
            right = getRight().equals(other.getRight());
        else if (other.getRight() == null)
            right = true;

        if ( getLeft() != null )
            left = getLeft().equals(other.getLeft());
        else if (other.getLeft() == null)
            left = true;

        return (left && right);
    }


    static final long serialVersionUID = -1653377563227996332L;
}
