/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.structures.CDT.probe;

import java.util.Stack;

import matrix.structures.FDT.FDT;
import matrix.structures.FDT.probe.BinaryTree;
import matrix.structures.memory.*;
import matrix.structures.other.Rotations;
import matrix.util.Note;

/**
 * Splay tree is a data structure where the operations adjust the
 * data structure to ensure O(log(n)) amortized running time.
 * Adjustment is done by splaying operations.
 *
 * @author Jussi Nikander
 */
public class SplayTree extends BinSearchTree {

    public SplayTree() {
        this(null);
    }

    public SplayTree(Object o) {
        super(o);
    }

    /**
     * Method that does the splay operation
     */
    private BinaryTree splay(Element key, BinaryTree tree) {
        if(tree == null)
            return null;
        if(key == null)
            return tree;
        Stack route = findRoute(key, tree);
        BinaryTree tmp = rotate(route);
        if(tmp == null)
            return tree;
        return tmp;
    }

    // Finds the route to the node required by splay. The required
    // node is at the top of the stack in the end
    private Stack findRoute(Element key, BinaryTree tree) {
        Stack route = new Stack();
        BinaryTree current = tree;
        BinaryTree required = tree;
        // tells if there are keys smaller than the arg. Element in tree
        boolean smaller = key.gt((Element)current.getElement());
        while(current != null) {
            route.push(current);
            if(key.lt((Element)current.getElement())) {
                if(!smaller)
                    if(((Element)current.getElement()).lt((Element)required.getElement()))
                        required = current;
                current = current.getLeft();
            } else if(key.eq((Element)current.getElement())) {
                required = current;
                break;
            } else {
                if(!smaller) {
                    if(((Element)current.getElement()).lt((Element)required.getElement()))
                        required = current;
                } else if(((Element)current.getElement()).gt((Element)required.getElement()))
                    required = current;
                smaller = true;
                current = current.getRight();
            }
        }
        while(!route.empty() && required != route.peek())
            route.pop();
        Note.out(this, "Required element found: " + required.getElement());
        return route;
    }

    // Does the rotations needed in the splaying operation.
    private BinaryTree rotate(Stack route) {
        BinaryTree current = null;
        BinaryTree parent = null;
        BinaryTree grandParent = null;
        BinaryTree greatGrandParent = null;
        if(!route.empty()) {
            current = (BinaryTree)route.pop();
            if(!route.empty()) {
                parent = (BinaryTree)route.pop();
                if(!route.empty()) {
                    grandParent = (BinaryTree)route.pop();
                    if(!route.empty())
                        greatGrandParent = (BinaryTree)route.pop();
                } else
                    return simpleRotation(parent, current);
            } else
                return current;
        } else {
            Note.out(this, "Empty route.");
            return null;
        }
        boolean cont = true;
        // This loop does the rotations. Current shouldn't change during it...
        do {
            if(parent == grandParent.getRight()) {
                if(current == parent.getRight()) {
                    Note.out(this, "splayWithRight");
                    current = Rotations.splayWithRight(grandParent);
                } else {
                    Note.out(this, "doubleWithRight");
                    current = Rotations.doubleWithRightChild(grandParent);
                }
            } else {
                if(current == parent.getRight()) {
                    Note.out(this, "doubleWithLeft");
                    current = Rotations.doubleWithLeftChild(grandParent);
                } else {
                    Note.out(this, "splayWithLeft");
                    current = Rotations.splayWithLeft(grandParent);
                }
            }
            if(greatGrandParent != null) {
                if(grandParent == greatGrandParent.getLeft())
                    greatGrandParent.setLeft(current);
                else
                    greatGrandParent.setRight(current);
                parent = greatGrandParent;
                if(!route.empty()) {
                    grandParent = (BinaryTree)route.pop();
                    if(!route.empty())
                        greatGrandParent = (BinaryTree)route.pop();
                    else
                        greatGrandParent = null;
                } else
                    return simpleRotation(parent, current);
            } else
                return current;
        }while(parent != null);
        return current;
    }

    private BinaryTree simpleRotation(BinaryTree parent, BinaryTree child) {
        if(child == parent.getLeft()) {
            Note.out(this, "simple with left");
            return Rotations.withLeftChild(parent);
        }
        Note.out(this, "simple with right");
        return Rotations.withRightChild(parent);
    }

    //Implementation of operations defined for splay tree starts here

    /**
     * Tests if the parameter <code>Element</code> is in the parameter
     * <code>BinaryTree</code> by using the splay operation.
     *
     * @param k the <code>Element</code> searched
     * @param tree The <code>BinaryTree</code> where the element is searched from
     * @return true if the <code>Element</code> is in the tree, false otherwise
     */
    public boolean isInTree(Element k, BinaryTree tree) {
        tree = splay(k, tree);
        if(tree.getElement().equals(k))
            return true;
        return false;
    }

    /**
     * Combines two binary trees into one.
     *
     * @param s1 The first BinaryTree
     * @param s2 the second binary tree
     * @return the combined tree
     */
    public BinaryTree combine(BinaryTree s1, BinaryTree s2) {
        s1 = splay(LexInf.getInstance(), s1);
        if(s1 == null)
            return s2;
        s1.setRight(s2);
        return s1;
    }

    /**
     * Divides two binary trees by splaying the param tree, and
     * and then splitting the splayed tree. One tree contains the
     * root and the left subtree of the splayed tree, the other contains
     * the right subtree of the splayed tree
     *
     * @param key the <code>Element</code> used to splay the tree
     * @param tree the <code>BinaryTree</code> to be divided
     * @return an array containing the two trees
     */
    public BinaryTree[] divide(Element key, BinaryTree tree) {
        if(tree == null)
            return null;
        BinaryTree[] t = new BinaryTree[2];
        t[0] = splay(key, tree);
        t[1] = t[0].getRight();
        t[0].setRight(null);
        return t;
    }

    /**
     * Deletes the param <code>Element</code> from the param <code>BinaryTree</code>
     *
     * @param key the <code>Element</code> to be deleted
     * @param tree the <code>BinaryTree</code> where the <code>Element</code> should be deleted from
     * @return the modified tree
     */
    public BinaryTree delete(Element key, BinaryTree tree) {
        BinaryTree[] t = divide(key, tree);
        t[0] = t[0].getLeft();
        return combine(t[0], t[1]);
    }

    /**
     * Inserts the parameter key to the parameter tree
     *
     * @param key the <code>Element</code> to be inserted
     * @param tree the <code>BinaryTree</code> where the key should be inserted in
     * @return the modified tree
     */
    public BinaryTree insert(Element key, BinaryTree tree) {
        if(tree == null) {
            Note.out(this, "Null tree encountered!");
            return (BinaryTree)getNewNode(key);
        }
        if(tree.getElement() == null) {
            Note.out(this, "Null key encountered during insert.");
            tree.setElement(key);
            return tree;
        }
        BinaryTree[] t = divide(key, tree);
        if(t[0].getElement().equals(key)) {
            Note.out(this, "Duplicate, nothing will be done");
            tree = t[0];
            tree.setRight(t[1]);
        } else {
            if(((Element)t[0].getElement()).gt(key)) {
                Note.out(this, "key < root(" + t[0].getElement() + "), switching root to right subtree");
                t[0].setRight(t[1]);
                tree = t[0];
                t[0] = (BinaryTree)tree.getLeft();
                tree.setLeft(null);
                t[1] = tree;
            }
            tree = (BinaryTree)getNewNode(key);
            tree.setLeft(t[0]);
            tree.setRight(t[1]);
        }
        return tree;
    }

    //--- *** CDT interface *** ---

    public FDT insert(Object o) {
        if(o == null)
            return this;
        if(o instanceof Element) {
            setElement(insert((Element)o, (BinaryTree)getElement()));
            return this;
        }
        Key k = new Key(o);
        setElement(insert(k, (BinaryTree)getElement()));
        return this;
    }

    public FDT delete(Object o) {
        if(o == null)
            return this;
        if(o instanceof Element) {
            setElement(delete((Element)o, (BinaryTree)getElement()));
            return this;
        }
        Note.show(this, "Cannot delete, not a Key");
        return this;
    }


    static final long serialVersionUID = 5859577296862056523L;
}
