/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: StackImpl.java,v 1.4 2004/05/26 08:05:52 vkaravir Exp $ */

package matrix.structures.CDT.probe;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.memory.*;

/**
 * Implementation of a stack. This stack is implemented with doubly linked list.
 * Thus, this stack is a collection of nodes that contains elements of the stack.
 * Every node contains also pointers to the previous node and the next node.
 * The node which pointer to previous node points to null is a top node of 
 * the stack.
 *
 */
public class StackImpl extends LinkedListImpl implements CDT, DoublyLinkedList {
    

    /** previous node in the stack */
    private DoublyLinkedList prev;


    /**
     * Constructs StackImpl.
     *
     */
    public StackImpl() {
	super();
    }

    /**
     * Constructs StackImpl with given key.
     *
     * @param key the key of the new StackImpl
     */
    public StackImpl(Object key) {
	super();
	setElement(key);
    }

    /**
     * Returns new (Stack) node with a given key. 
     *
     * @param key key of new node
     * @return new (Stack) Node
     */    
    public LinkedList getNewNode(Object key) {
	return new StackImpl(key);
	}


    /**
     * Inserts a new key to the top of the stack. This insert operation just calls 
     * the push operation.
     *
     * @param key new key
     * @return stack after insertion
     */
    public FDT insert(Object key) {
	return push(key);
    }


    /**
     * Pushes a new key to the top of the stack.
     *
     * @param key new key
     * @return stack after insertion of the key
     */
   public FDT push(Object key) {
	if (key == null)
	    return getTopOfStack();
	if (key instanceof Key)
		if (((Key)key).toString() == "")
			return getTopOfStack();
	if (key instanceof Element) {
	    StackImpl newNode = new StackImpl();
	    newNode.setElement(key);
	    setTopOfStack(newNode);
	    return getTopOfStack();
	}
	if (key instanceof Table) {
	    Table t = (Table)key;
	    for(int i = 0; i < t.size(); i++)
		insert(t.getObject(i));
	    return getTopOfStack();
	}
	Key k = new Key(key);
	StackImpl newNode = new StackImpl();
	newNode.setElement(key);
	setTopOfStack(newNode);
	return getTopOfStack();
    }


    /**
     * Sets the given node to the top of the stack.
     *
     * @param newNode new (Stack) node to be added to the top of the stack
     */
    private void setTopOfStack(StackImpl newNode) {
	if (getElement() == null && hasPrev() == false) {
	    setElement(newNode.getElement());
	}
	else if (hasPrev() == false) {
	    newNode.setNext(this);	    
	    setPrev(newNode);
	}
	else {
	    ((StackImpl)prev).setTopOfStack(newNode);
	}
    }

    /**
     * Returns the top node of the stack
     *
     * @return node from the top of the stack
     */
    private StackImpl getTopOfStack() {
	if (hasPrev() == false){
	    return this;}
	else
	    return ((StackImpl)prev).getTopOfStack();
    }


    /**
     * This method deletes the node containing the given key
     * and all the nodes before it from the stack and returns 
     * the first node of stack after deletion.
     *
     * @param o key to be deleted
     * @return returns always the top node of the stack after deletion
     */
    public FDT delete(Object o) {
	boolean isInStack = false;
	StackImpl temp = this;
	while(temp != null) {
	    if(o instanceof Key){
			if(temp.getElement() == ((Key)o).getOrig())
		    	isInStack = true;
		    else if(temp.getElement() == o)
		    	isInStack = true;
	    }
	    else {
			if(temp.getElement() == o)
		    	isInStack = true;
	    }
	    temp = (StackImpl)temp.getNext();
	}
	if(isInStack)
	    return pop(o);
	return getTopOfStack();
    }

    /**
     * This method pops the node containing the given key and all its
     * predecessors from the stack and returns the first
     * node of the stack after deletion
     *
     * @return top node of the stack after deletion
     */
   public FDT pop(Object o) {
       if (hasPrev() == false && getNext() == null) { //only one node in the stack
	    setElement(null);
	    return this;
       }
       else if (hasPrev() == false) { //this node is the top node 
	   StackImpl temp = (StackImpl)getNext();
	   setNext(null);
	   temp.setPrev(null);
	   //   return temp;
	   if(o instanceof Key) {
	       if(getElement() == ((Key)o).getOrig())
			   return temp;
	       else if(getElement() == o)
		       return temp;
	       else
			   return temp.delete(o);
	   }
	   else {
			if(getElement() == o)
		   		return temp;
	       	else
		   		return temp.delete(o);
	   }
       }
       else {
	   //return ((StackImpl)prev).pop(o);
	   return ((StackImpl)prev).delete(o);
       }
   }
    

    /**
     * Sets the previous node.
     *
     * @param o previous node
     */
    public void setPrev(DoublyLinkedList o) {
	prev = o;
    }


    /**
     * Returns the previous node.
     *
     * @return the previous node
     */
    public DoublyLinkedList getPrev() {
	return prev;
    }


    /**
     * Tests whether there is a previous node.
     *
     * @return true, if there is a previous node
     */
    public boolean hasPrev() {
	return (getPrev() != null);
    } 


    public Object search(Object o) {
	return top();
    }

    /**
     * Returns the top node of the stack.
     *
     * @return top node of the stack
     */
    public FDT top() {
	return getTopOfStack();
    }


    static final long serialVersionUID = 4008244601878734431L;
}
