/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: LabeledBinTree.java,v 1.8 2004/05/26 08:05:54 vkaravir Exp $ */

package matrix.structures.FDT.probe;

import java.util.BitSet;

import matrix.decoration.LabelDecorator;
import matrix.structures.FDT.Tree;
import matrix.structures.memory.VirtualArray;
import matrix.util.*;

/**
 * A node of a binary tree. This node implemenets the <code>LabelDecorator</code> -interface
 * and each node is labeled with the bit representation of its key. Bitrepresentations are
 * created with the matrix.util.KeyBuilder -class.
 *
 * @author Jussi Nikander
 */
public class LabeledBinTree extends BinTree implements LabelDecorator {

    /**
     *Virtual array containing the class variables
     * <ul>
     * <li>bt[0] is the <code>String</code> key</li>
     * <li>bt[1] is the label</li>
     * </ul>
     */
    private VirtualArray bt;


    /** Constructs a new <code>LabeledBinTree</code> */

    public LabeledBinTree() {
        super();
        bt = new VirtualArray();
    }

    /**
     * Constructs a new <code>LabeledBinTree</code>
     * @param key Contents and key of the node
     */
    public LabeledBinTree(String key) {
        this(key, key);
    }

    /**
     * Constructs a new <code>LabeledBinTree</code>.
     * @param contents Contents of the node.
     * @param key Key of the node.
     */
    public LabeledBinTree(Object contents, String key) {
        super(contents);
        bt = new VirtualArray();
        bt.setObject(key, 0);
        if(key != "")
            makeLabel();
        else
            bt.setObject("", 1);
    }

    /**
     * Constructs a new <code>LabeledBinTree</code>
     * @param contents the contents of the node
     */
    public LabeledBinTree(Object contents) {
        super(contents);
        bt = new VirtualArray();
        if(contents != null)
            bt.setObject(contents.toString(), 0);
        if(bt.getObject(0) != null && bt.getObject(0) != "")
            makeLabel();
        else
            bt.setObject("", 1);
    }


    /**
     * Compares <code>This</code> to another <code>LabeledBinTree</code>. The two are
     * equal only if they are the same object
     * @return true if the two nodes are equal, false otherwise.

    public boolean equals(Object node) {
        if(getElement() == null)
            return false;
        if(this == node)
            return true;
        return false;
    }
     */
    /**
     * Sets the contents of this node to a new value.
     * @param contents the new contents
     * @param key the new key
     */
    public void setContents(Object contents, String key) {
        setElement(contents);
        bt.setObject(key, 0);
        makeLabel();
        return;
    }

    /**
     * Returns the contents of this node.
     * @return <code>Object</code> which is this node's contents
     */
    public Object getContents() {
        return bt.getObject(0);
    }

    /**
     * Returns the key of this node.
     * @return <code>String</code> which is the key of this node
     */
    public String getKey() {
        return (String) bt.getObject(0);
    }

    // methods for interface matrix.structures.FDT.FDT


    /**
     * Sets the contents of this node.
     * @param o the new contents
     */
    public void setElement(Object o) {
        super.setElement(o);
        if(o != null)
            bt.setObject(o.toString(), 0);
        else
            bt.setObject("", 0);
        makeLabel();
    }


    // methods for interface matrix.structures.FDT.BinaryTree

    /**
     * Sets the left child of this node.
     * @param p the <code>BinaryTree</code> to be set as the left child
     */
    public void setLeft(BinaryTree p) {
        if(p == null) {
            super.setLeft(p);
        } else if(p instanceof LabeledBinTree) {
            super.setLeft(p);
        }
        else {
            Object obj = p.getElement();
            LabeledBinTree leftChild = new LabeledBinTree(obj, obj.toString());
            leftChild.setRight(p.getRight());
            leftChild.setLeft(p.getLeft());
            super.setLeft(leftChild);
        }
    }

    /**
     * Sets the right child of this node.
     * @param p the <code>BinaryTree</code> to be set as the right child
     */
    public void setRight(BinaryTree p) {
        if(p == null) {
            super.setRight(p);
        }else if(p instanceof LabeledBinTree) {
            super.setRight(p);
        }
        else {
            Object obj = p.getElement();
            LabeledBinTree rightChild = new LabeledBinTree(obj, obj.toString());
            rightChild.setRight(p.getRight());
            rightChild.setLeft(p.getLeft());
            super.setRight(rightChild);
        }
    }


    // Methods for interface matrix.structures.FDT.SimulationTree;


    /**
     * Sets the i'th child of this node.
     * @param t the new child
     * @param child the number i
     */
    public void setSubTree(Tree t, int child) {
        if(t instanceof LabeledBinTree) {
            LabeledBinTree tmp = (LabeledBinTree) t;
            if(child == 0)
                setLeft(tmp);
            else
                setRight(tmp);
        } else
            Note.out(this, "parameter " + t + " not compatible");
    }

    /**
     * Returns a new node.
     * @param key the contents of the new node
     * @return the new <code>Tree</code> node.
     */
    public Tree getNewNode(Object key) {
        return new LabeledBinTree(key, key.toString());
    }

    // Methods for interface LabelDecorator

    /**
     * Tells if labeling is enabled for this node.
     * @return true if labeling is enabled, false otherwise
     */
    public boolean isLabelEnabled() {
        String key = (String) bt.getObject(0);
        if(key != "" && key != null)
            return true;
        else
            return false;
    }

    /**
     * Sets the label of this node.
     * @param s the new label
     */
    public void setLabel(String s) {
        bt.setObject(s, 1);
        return;
    }

    /**
     * Returns the label of this node.
     * @return the label
     */
    public String getLabel() {
        return (String) bt.getObject(1);
    }

    /**
     * Tells if references of this node can be labeled.
     * @return false, since reference labeling is not supported
     */
    public boolean isReferenceLabelEnabled() {
        return false;
    }

    /**
     * Does nothing. This method is needed for implementing the <code>LabelDecorator</code> interface.
     */
    public void setReferenceLabel(int i, String s) {
        return;
    }

    /**
     * Does nothing. This method is needed for implementing the <code>LabelDecorator</code> interface.
     * @return an empty <code>String</code>
     */
    public String getReferenceLabel(int i) {
        return "";
    }

    // Method for creating a a bit representation of the key

    private void makeLabel() {
        String to = "";
        String key = (String) bt.getObject(0);
        Object contents = getElement();
        if(key == "" || key == null || (contents instanceof String && (contents == "" || contents == null))) {
            to = "";
        } else {
            BitSet tmp = KeyBuilder.getDigitalKey(key);
            int length = KeyBuilder.getDigitalKeyLength(key);
            for(int i = 0; i < length; i++) {
                if(tmp.get(i))
                    to += "1";
                else
                    to += "0";
            }

        }
        setLabel(to);
        return;
    }


    static final long serialVersionUID = -2410153543089085951L;
}
