/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.structures.other;

import java.util.*;

import matrix.structures.FDT.*;
import matrix.structures.FDT.LinkedList;

/**
 * Contains methods for comparing the equality of different kinds of data structures.
 * All the comparison methods are static, so there should never be any reason to create
 * an instance of this class.
 */

public class Comparer {

  /**
   * Compares two arrays. 
   * @param a the first array to be compared
   * @param b the second array to be compared
   * @return true if and only if the states of the two arrays are equal
   */
  public static boolean compareArray(Array a, Array b) {
 
    if(a == b) return true;
    if(a == null || b == null) return false;
   
    int aFirst = a.getFirst();
    int aLast = a.getLast();
   
    if(aFirst != b.getFirst() || aLast != b.getLast())
      return false;

    Object aObj;
    Object bObj;

    for(int i = aFirst; i < aLast; i++) {
	aObj = a.getObject(i);
	bObj = b.getObject(i);      
	if(aObj == bObj) continue;
	if(aObj != null && !(aObj.equals(bObj)))
	    return false;
	if(aObj == null || bObj == null)
	    return false;
    }
    if (aFirst == aLast) {
	aObj = a.getObject(aFirst);
	bObj = b.getObject(aFirst);
	if(aObj == bObj) return true;
	if(aObj != null && !(aObj.equals(bObj)))
	    return false;
	if(aObj == null || bObj == null)
	    return false;
    }
    return true;
  }

    public static boolean compareList(LinkedList a, LinkedList b) {
	if(a == b) return true;
	if(a == null || b == null) return false;
	Object aObj = a.getElement();
	Object bObj = b.getElement();
	if((aObj == null && bObj != null) || (aObj != null && bObj == null))
	    return false;
	if(aObj != null && !(aObj.equals(bObj)))
	    return false;
	if(a.hasNext() && b.hasNext())
	    return compareList(a.getNext(), b.getNext());
	else if(!a.hasNext() && !b.hasNext())
	    return true;
	return false;
    }


    /**
     * Compares two static lists.
     * @param a the first list to be compared
     * @param b the second list to be compared
     * @return true if and only if the contents of the two lists are equal
     */
    public static boolean compareList(StaticLinkedList a, StaticLinkedList b) {

	//Note.out(null, "List compare start");
	if(a == b) return true;
	if(a == null || b == null) return false;
	a.setFirst();
	b.setFirst();
	Object aObj;
	Object bObj;
	while(a.isInList() && b.isInList()) {
	    aObj = a.currValue();
	    bObj = b.currValue();
	    //Note.out(null, "Object a: " + aObj + "Object b: " + bObj);
	    if(aObj == bObj) {
		a.next();
		b.next();
		continue;
	    }
	    if(aObj == null || bObj == null)
		return false;
	    if(!(aObj.equals(bObj)))
		return false;
	    a.next();
	    b.next();
	}
	if(a.isInList() || b.isInList())
	    return false;
	return true;
    }

  /**
   * Compares two common trees. This is an identical comparison, so the children of
   * the nodes must be in the same order in both trees.
   * @param a the first tree to be compared
   * @param b the second tree to be compared
   * @return true if and only if the contents of the two trees are equal
   */
  public static boolean compareIdentical(Tree a, Tree b) {
      // System.out.println(a+":"+b);
    if(a == b) return true;
    if(a == null || b == null) return false;
    Object aObj = a.getElement();
    Object bObj = b.getElement();
    //if(aObj == bObj) return true;
    if((aObj == null && bObj != null)  || (bObj == null && aObj != null))
      return false;
    if(aObj != null && !(aObj.equals(bObj)))
      return false;
    Tree[] aTree = a.getSubTrees();
    Tree[] bTree = b.getSubTrees();
    if(aTree.length != bTree.length)
      return false;
    for(int i = 0; i < aTree.length; i++)
      if(!compareIdentical(aTree[i], bTree[i]))
	return false;
    return true;
  }

  /**
   * Compares two common trees. This is an isomorphic comparison, so the children of
   * the nodes need not be in the same order in both trees.
   * @param a the first tree to be compared
   * @param b the second tree to be compared
   * @return true if and only if the contents of the two trees are equal
   */
  public static boolean compareIsomorphic(Tree a, Tree b) {
    if(a == b) return true;
    if(a == null || b == null) return false;
    Object aObj = a.getElement();
    Object bObj = b.getElement();
    //if(aObj == bObj) return true;
    if((aObj == null && bObj != null)  || (bObj == null && aObj != null))
      return false;
    if(aObj != null && !(aObj.equals(bObj)))
      return false;
    Tree[] aTree = a.getSubTrees();
    Tree[] bTree = b.getSubTrees();
    if(aTree.length != bTree.length)
      return false;
    BitSet matches = new BitSet(aTree.length);
    for(int i = 0; i < aTree.length; i++) {
      for(int j = 0; j < bTree.length; i++) {
	if(!matches.get(j)) 
	  if(compareIsomorphic(aTree[i], bTree[j]))
	    matches.set(j);
      }
    }
    for(int i = 0; i < aTree.length; i++) 
      if(!matches.get(i))
	return false;
    return true;
  }

  /**
   * Compares two graphs.
   * @param a the first graph to be compared
   * @param b the second graph to be compared
   * @return true if and only if the contents of the two graphs are equal
   */
  public static boolean compareGraph(SimulationGraph a, SimulationGraph b) {
    if(a == b) return true;
    if(a == null || b == null) return false;
    Vertex[] aVertices = a.getVertices();
    Vertex[] bVertices = b.getVertices();
    Object ao;
    Object bo;
    for(int i = 0; i < aVertices.length; i++) {
      ao = aVertices[i].getElement();
      for(int j = 0; j < bVertices.length; i++) {
	bo = bVertices[j].getElement();
	if(ao.equals(bo)) {
	  Hashtable hasha = new Hashtable();
	  Hashtable hashb = new Hashtable();
	  if(compareVertex(aVertices[i], bVertices[j], hasha, hashb))
	    return true;
	}
      }
    }
    return false;
  }

  // Helper function for graph comparison. Two vertices are considered equal if and only if
  // they contain the same key and have identical successors. 
  private static boolean compareVertex(Vertex a, Vertex b, Hashtable ha, Hashtable hb) {
    if(!(ha.containsKey(a)) && !(hb.contains(b))) { //Should the contains() -method be used here?
      Object aObj = a.getElement();
      Object bObj = b.getElement();
      if(aObj.equals(bObj)) {
	ha.put(a, a);
	hb.put(b, b);
	Vertex[] aSucc = a.getSuccessors();
	Vertex[] bSucc = b.getSuccessors();
	if(aSucc.length == bSucc.length) {
	  BitSet bits = new BitSet(aSucc.length);
	Loop:
	  for(int i = 0; i < aSucc.length; i++) {
	    for(int j = 0; j < aSucc.length; j++) {
	      if(!bits.get(j))
		if(compareVertex(aSucc[i], bSucc[j], ha, hb)) {
		  bits.set(j);
		  continue Loop;
		}
	    }
	  }
	  for(int i = 0; i < aSucc.length; i++) {
	    if(!bits.get(i)) {
	      ha.remove(a);
	      hb.remove(b);
	      return false;
	    }
	  }
	  return true;
	}
	ha.remove(a);
	hb.remove(b);
	return false;
      }
      return false;
    }
      if(ha.containsKey(a) && hb.containsKey(b)) 
	if(a.equals(b))
	   return true;
    return false;
  }

  /*
   * Compares two binary trees. This comparison is identical, so the children of each
   * node must be in the same order.
   * @param a the first tree to be compared
   * @param b the second tree to be compared
   * @return true if and only if the contents of the two trees are equal

  public static boolean compareIndentical(BinaryTree a, BinaryTree b) {
    if(a == b) return true;
    if(a == b && a == null) return true;
    if(a == null || b == null) return false;
    Object aObj = a.getElement();
    Object bObj = b.getElement();
    if(aObj == bObj && aObj == null) return true;
    if(aObj == null || bObj == null)
      return false;
    if(!(aObj.equals(bObj)))
      return false;
    if(!compareIdentical(a.getLeft(), b.getLeft())) return false;
    if(!compareIdentical(a.getRight(), b.getRight())) return false;
    return true;
  }
   */
  /*
   * Compares two binary trees. The comparison is isomorphic, so the children of 
   * nodes don't need to be in the same order.
   * @param a the first tree to be compared
   * @param b the second tree to be compared
   * @return true if and only if the contents of the two trees are equal

  public static boolean compareIsomorphic(BinaryTree a, BinaryTree b) {
    if(a == b && a == null) return true;
    if(a == null || b == null) return false;
    Object aObj = a.getElement();
    Object bObj = b.getElement();
    if(aObj == bObj && aObj == null) return true;
    if(aObj != null && !(aObj.equals(bObj)))
      return false;
    if(aObj == null || bObj == null)
      return false;
    return (compareIsomorphic(a.getLeft(), b.getLeft()) && compareIsomorphic(a.getRight(), b.getRight()))
      || (compareIsomorphic(a.getLeft(), b.getRight()) && compareIsomorphic(a.getRight(), b.getLeft()));
  }
   */


}



