/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2002  Ari Korhonen and others
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package matrix.util.configuration;

import java.io.*;
import java.lang.reflect.Field;
import java.net.*;
import java.util.Vector;
import java.util.jar.JarFile;

import javax.xml.parsers.*;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import matrix.uitools.MenuFactory;
import matrix.util.Note;
import matrix.visual.RepresentationFactory;
/**
 * This class parses a matrix configuration file and creates a configuration,
 * including all menus, for matrix. The configuration file must be an XML
 * file confirming to the matrix.dtd. The parsing is started by calling the
 * <code>parse()</code> -method.
 *
 * <code>MatrixConfiguration</code> is a singleton class. To get the instance
 * of this class, use the <code>getInstance()</code> -method.
 *
 * @author Jussi Nikander
 */

public class MatrixConfiguration extends DefaultHandler {

    // ---- static fields ----

    private static MatrixConfiguration theInstance = null;
    private static String confFile = null;
    private static String url = null; 
    private static String codebaseURL = null; 
    private static boolean validatingParser = false;
    private static ConfItem configuration = null;
    private static String jarFile = null;
    private static InputStream in = null;

    private static final String DEFAULT = "matrixconf.xml";
    private static final String JARFILE = "matrix.jar";
    private static final String JARCONFFILE = "matrixconf.xml";


    // ---- static methods ----

    /**
     * Retrieves the singleton instance of this class.
     */
    public static MatrixConfiguration getInstance() {
	if(theInstance == null) {
	    theInstance = new MatrixConfiguration();
	}
	return theInstance;
    }

    /**
     * Sets the local path from which the configuration
     * file is read. It is best to use the full, absolute
     * path to the configuration file in order to make sure
     * that it is found. Most Java implementations are capable
     * of reading relative paths, but not necessarely all.
     *
     * @param s the path to the configuration file
     */
    public static void setConfPath(String s) {
	confFile = s;
    }

    public static String getConfPath() {
	return confFile;
    }

    /**
     * Sets the URL to the configuration file, making it
     * possible to read the configuration file from the
     * Internet. This is used by the applet version
     * of the framework to search for the configuration
     * file.
     *
     * @param s the URL to the configuration file
     */
    public static void setUrl(String s) {
	url = s;
    }

    public static String getUrl() {
	return url;
    }

    public static void setJarFile(String s) {
	jarFile = s;
    }

    public static String getJarFile() {
	return jarFile;
    }


    /*
     * XXX not needed -jtn
     * needed by applets -vkaravir
     */
    public static void setCodebaseURL(String s) {
	codebaseURL = s;
    }

    public static String getCodebaseURL() {
	return codebaseURL;
    }
    

    public static InputStream getInputStream() {
        return in;
    }

    public static void setInputStream(InputStream stream) {
        in = stream;
    }


    /**
     * Sets the XML parser to validate the configuration
     * file using the DTD specified in the XML.
     * 
     * @param v if true, the XML will be validated, otherwise
     */
    public static void setValidatingParser(boolean v) {
	validatingParser = v;
    }

    public static boolean getValidatingParser() {
	return validatingParser;
    }

    /**
     * Returns the root item of the whole Matrix
     * Configuration.
     *
     * @return The ConfItem corresponding to the &lt;configuration&gt; -element
     */
    public static ConfItem getConfiguration() {
	return configuration;
    }

    // ---- fields ----

    private ConfStack stack;
    private Class currentAnchor;

    // ---- constructor ----

    private MatrixConfiguration() {
	stack = new ConfStack();
	currentAnchor = null;
	parse();
    }

    // ---- XML parsing methods ----

    // parse has been lifted line-by-line from old MatrixConfiguration. There is 
    // no need for changes in this method.

    /**
     * Tries to open configuration file for reading and parse it. If Configuration file location has
     * been set using <code>setConfPath</code> that location is tried. If 
     * configuration file cannot be opened from specified location, the program tries
     * to find file "matrixconf.xml" in the current working directory. 
     * <p>
     * If configuration file is not set or could not be opened for reading, the 
     * program will try to open configuration file from an URL, if configuration file
     * URL has been set using <code>setUrl</code>. 
     * <p>
     * If neither configuration file location nor URL has been set, or configuration could not
     * successfully be parsed from either, the program will try to see if it finds the
     * Matrix jar package holding the configuration file. 
     * <p>
     * If configuration could not be read from jar, the program will give up and 
     * exit.
     * <p>
     */
    public void parse() {
	Note.out(this, "Starting parsing");
	SAXParserFactory factory = SAXParserFactory.newInstance();
	factory.setValidating(getValidatingParser());
	SAXParser parser = null;
	try {
	    parser = factory.newSAXParser();
	} catch(Exception e) {
	    throw new MatrixConfigurationException("Could not create SAX Parser; cannot read configuration.");
	}
	File conf = null;
	// First, if confFile is set, we try to open it.
	if(confFile != null) {
	    try {
		conf = new File(confFile);
		if(!(conf.exists()) || !(conf.isFile())) {
		    Note.out(this,"Configuration file not found, trying default.");
		    conf = null;
		}
	    } catch(SecurityException e) {
		Note.out(this,"Could not open configuration file, trying default.");
		conf = null;
	    }
	}
	// if confFile could not be opened, we try the default.
	if(conf == null) {
	    try {
		conf = new File(DEFAULT);
		if(!(conf.exists()) || !(conf.isFile())) {
		    Note.out(this,"Default configuration file not found.");
		    conf = null;
		}
	    } catch(SecurityException e) {
		Note.out(this,"Could not open default configuration file.");
		conf = null;
	    }
	}
	
	// Now we try to parse the configuration, if we have an InputStream for the
	// confFile or an URL where the configuration file can be found at.
	try {
	    if(conf != null) {
		Note.out(this, "Parsing...");
		parser.parse(conf, this);
		Note.out(this, "Ended parsing");
		return;
	    } else if(url != null) {
		Note.out(this, "Parsing URL...");
		parser.parse(url, this);
		Note.out(this, "Ended parsing");
		return;
	    }
	} catch(SAXException e) {
	    Note.out(this,"Error occurred when trying to parse configuration file.");
	    e.printStackTrace();
	} catch(IOException e) {
	    Note.out(this,"Could not parse configuration file.");
	    e.printStackTrace();
	}
        // added 8.6.2004 -- if InputStream is set, use it - vkaravir
        if (in != null) {
            Note.out(this, "Using InputStream");
            try {
                parser.parse(in, this);
                return;
            } catch(SAXException e) {
               Note.out(this,"Error occurred when trying to parse configuration file.");
                e.printStackTrace();
            } catch(IOException e) {
                Note.out(this,"Could not parse configuration file.");
                e.printStackTrace();
            }
        }
	// If no confFile or URL was present or could not be parsed, we try to find
	// the Matrix jar and open the file from it. 
	try {
	    Note.out(this, "Trying jar");
	    if (codebaseURL != null) {
		String jarURL = "jar:" + codebaseURL + ((jarFile == null) ? JARFILE : jarFile) + "!/";
		Note.out(this, "Reading configuration file from " + jarURL);
		URL u = new URL(jarURL);
		JarURLConnection conn = (JarURLConnection)u.openConnection();
		JarFile jar = conn.getJarFile();
		parser.parse(jar.getInputStream(jar.getEntry(JARCONFFILE)),  this);
		return;
	    } else {
		JarFile jar = new JarFile((jarFile == null) ? JARFILE : jarFile);
                Note.out(this, "Reading configuration file from " + ((jarFile == null) ? JARFILE : jarFile));
		parser.parse(jar.getInputStream(jar.getEntry(JARCONFFILE)),  this);
		return;
	    }
	} catch(Exception e) {
	    e.printStackTrace();
	}
	// Give up, if configuration could not be created
	throw new MatrixConfigurationException("Could not configure Matrix");
    }
    
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
	//Note.out(this, "Starting element " + qName);
	try {
	    if(qName.equals("configuration")) {
		// configuration has no attributes
		startConfiguration();
	    } else if(qName.equals("menubar")) {
		startMenuBar(attributes);
	    } else if(qName.equals("menu")) {
		startMenu(attributes);
	    } else if(qName.equals("menuitem")) {
		startMenuItem(attributes);
	    } else if(qName.equals("checked")) {
		// checked has no attributes
		startChecked();
	    } else if(qName.equals("menuseparator")) {
		//menuseparator has no attributes
		startMenuSeparator();
	    } else if(qName.equals("field")) {
		startField(attributes);
	    } else if(qName.equals("constructor")) {
		startConstructor(attributes);
	    } else if(qName.equals("method")) {
		startMethod(attributes);
	    } else if(qName.equals("int")) {
		startInt(attributes);
	    } else if(qName.equals("string")) {
		startString(attributes);
	    } else if(qName.equals("boolean")) {
		startBoolean(attributes);
	    } else if(qName.equals("char")) {
		startChar(attributes);
	    } else if(qName.equals("popupmenu")) {
		startPopupMenu(attributes);
	    } else if(qName.equals("initialization")) {
		// initialization has no attributes
		startInitialization();
	    } else if(qName.equals("visualizations")) {
		// visualizations has no attributes
		startVisualizations();
	    } else if(qName.equals("structure")) {
		startStructure(attributes);
	    } else if(qName.equals("visualization")) {
		// not to be confused with visualizations
		startVisualization(attributes);
	    } else if(qName.equals("picked-visualizations")) {
		// picked-visualizations has no attributes
		startPickedVisualizations();
	    }
	} catch(MatrixConfigurationException e) {
	    Note.out(this, "Exception caught while opening element " + qName);
	    e.printStackTrace();
	}

    }

    private void startConfiguration(){
	ConfItem item = new Configuration();
	stack.push(item);
	return;
    }

    private void startMenuBar(Attributes attributes){
	// MenuBar has only one legal attribute
	String anchorName = attributes.getValue(0);
	Class anchor = null;
	try {
	    anchor = Class.forName(anchorName);
	} catch(ClassNotFoundException e) {
	    throw new MatrixConfigurationException("unable to create anchor " + anchorName + " for MenuBar");
	}
	currentAnchor = anchor;
	ConfItem mb = new CMenuBar(anchor);
	stack.push(mb);
    }

    private void startMenu(Attributes attributes){
	String name = attributes.getValue(0);
	ConfItem menu = new CMenu(name);
	stack.push(menu);
    }

    private void startMenuItem(Attributes attributes){
	CMenuItem mi = new CMenuItem();
	for(int i = 0; i < attributes.getLength(); i++) {
	    String attr = attributes.getQName(i);
	    String val = attributes.getValue(i);
	    if(attr.equals("name")){
		mi.setName(val);
	    } else if(attr.equals("disabled")) {
		if(val.toLowerCase().equals("true")) {
		    mi.setDisabled(true);
		} else {
		    mi.setDisabled(false);
		}
	    } else if(attr.equals("type")) {
		mi.setType(val);
	    } else if(attr.equals("checked")) {
		if(val.toLowerCase().equals("true")) {
		    mi.setChecked(true);
		} else {
		    mi.setChecked(false);
		}
	    } else if(attr.equals("accesskey")) {
		Class keys = null;
		Field key = null;
		try {
		    keys = Class.forName("java.awt.event.KeyEvent");
		    key = keys.getField(val);
		    mi.setAccessKey(key.getInt(null));
		} catch(Exception e) {
		    e.printStackTrace();
		    Note.out(this, "Could not create access key " + val);
		}
	    } else if(attr.equals("group")) {
		mi.setGroup(val);
	    }
	}
	stack.push(mi);
    }

    private void startChecked(){
	stack.push(new CChecked());
    }
    private void startMenuSeparator(){
	stack.push(new CMenuSeparator());
    }
    private void startField(Attributes attributes){
	CField f = new CField();
	Class anchor = currentAnchor;
	String fName = null;
	for(int i = 0; i < attributes.getLength(); i++) {
	    String aName = attributes.getQName(i);
	    String val = attributes.getValue(i);
	    if(aName.equals("name")) {
		fName = val;
	    } else if(aName.equals("class")) {
		try {
		    anchor = Class.forName(val);
		} catch(ClassNotFoundException e) {
		    e.printStackTrace();
		    throw new MatrixConfigurationException("Could not initalize static anchor for field");
		}
	    }
	}
  	try {
  	    f.setField(anchor.getField(fName));
  	} catch(Exception e) {
  	    e.printStackTrace();
  	    throw new MatrixConfigurationException("Could not create required field" + fName);
  	}
	stack.push(f);
    }

    private void startConstructor(Attributes attributes){
	String cName = attributes.getValue(0);
	Class c = null;
	try {
	    c = Class.forName(cName);
	} catch(ClassNotFoundException e) {
	    e.printStackTrace();
	    throw new MatrixConfigurationException("Class to be instantiated not found");
	}
	ConfItem ci = new CConstructor(c);
	stack.push(ci);
    }

    private void startMethod(Attributes attributes){
	String mName = null;
	String cName = null;
	for(int i = 0; i < attributes.getLength(); i++) {
	    String curr = attributes.getQName(i);
	    if(curr.equals("name")) {
		mName = attributes.getValue(i);
	    } else if(curr.equals("class")) {
		cName = attributes.getValue(i);
	    }
	}
	Class anchor = null;
	if(cName != null) {
	    try {
		anchor = Class.forName(cName);
	    } catch(ClassNotFoundException e) {
		e.printStackTrace();
		throw new MatrixConfigurationException("Class not found");
	    }
	}
	ConfItem ci;
	if(cName != null)
	    ci = new CMethod(anchor, mName, true);
	else
	    ci = new CMethod(anchor, mName);
	stack.push(ci);
    }

    private void startInt(Attributes attributes){
	String intValue = attributes.getValue(0);
	ConfItem ci = new CInt(Integer.parseInt(intValue));
	stack.push(ci);
    }

    private void startString(Attributes attributes){
	String str = attributes.getValue(0);
	ConfItem ci = new CString(str);
	stack.push(ci);
    }
    private void startBoolean(Attributes attributes){
	String bVal = attributes.getValue(0);
	ConfItem ci = null;
	if(bVal.toLowerCase().equals("true")) {
	    ci = new CBoolean(true);
	} else {
	    ci = new CBoolean(false);
	}
	stack.push(ci);
    }
    private void startChar(Attributes attributes){
	String cVal = attributes.getValue(0);
	ConfItem ci = new CChar(cVal.charAt(0));
	stack.push(ci);
    }
    private void startPopupMenu(Attributes attributes){
	String aName = attributes.getValue(0);
	Class anchor = null;
	try {
	    anchor = Class.forName(aName);
	} catch(ClassNotFoundException e) {
	    e.printStackTrace();
	    throw new MatrixConfigurationException("Could not find class " + aName);
	}
	currentAnchor = anchor;
	ConfItem ci = new CPopupMenu(anchor);
	stack.push(ci);
    }
    private void startInitialization(){
	stack.push(new CInitialization());
    }

    private void startVisualizations(){
	stack.push(new CVisualizations());
    }
    private void startStructure(Attributes attributes){
	String sName = attributes.getValue(0);
	Class str = null;
	try {
	    str = Class.forName(sName);
	} catch(ClassNotFoundException e) {
	    e.printStackTrace();
	    throw new MatrixConfigurationException("No class " + sName + " found");
	}
	stack.push(new CStructure(str));
    }
    private void startVisualization(Attributes attributes){
	String name = null;
	String cName = null;
	for(int i = 0; i < attributes.getLength(); i++) {
	    String attr = attributes.getQName(i);
	    if(attr.equals("name")) {
		name = attributes.getValue(i);
	    } else if(attr.equals("class")) {
		cName = attributes.getValue(i);
	    }
	}
	Class visual = null;
	try {
	    visual = Class.forName(cName);
	} catch(ClassNotFoundException e) {
	    e.printStackTrace();
	    throw new MatrixConfigurationException("Could not find class " + cName);
	}
	ConfItem ci = new CVisualization(name,visual);
	stack.push(ci);
    }

    private void startPickedVisualizations(){
	stack.push(new CPickedVisualizations());
    }
	
    public void endElement(String uri, String localName, String qName) throws SAXException {
	//Note.out(this, "Ending element " + qName);
	try {
	    if(qName.equals("configuration")) {
		endConfiguration();
	    } else if(qName.equals("menubar")) {
		endMenuBar();
	    } else if(qName.equals("menu")) {
		endMenu();
	    } else if(qName.equals("menuitem")) {
		endMenuItem();
	    } else if(qName.equals("checked")) {
		endChecked();
	    } else if(qName.equals("menuseparator")) {
		endMenuSeparator();
	    } else if(qName.equals("field")) {
		endField();
	    } else if(qName.equals("constructor")) {
		endConstructor();
	    } else if(qName.equals("method")) {
		endMethod();
	    } else if(qName.equals("int")) {
		endInt();
	    } else if(qName.equals("string")) {
		endString();
	    } else if(qName.equals("boolean")) {
		endBoolean();
	    } else if(qName.equals("char")) {
		endChar();
	    } else if(qName.equals("popupmenu")) {
		endPopupMenu();
	    } else if(qName.equals("initialization")) {
		// initialization has no 
		endInitialization();
	    } else if(qName.equals("visualizations")) {
		endVisualizations();
	    } else if(qName.equals("structure")) {
		endStructure();
	    } else if(qName.equals("visualization")) {
		// not to be confused with visualizations
		endVisualization();
	    } else if(qName.equals("picked-visualizations")) {
		endPickedVisualizations();
	    }
	} catch(MatrixConfigurationException e) {
	    Note.out(this, "Exception caught while closing element " + qName);
	    e.printStackTrace();
	}
    }

    private void endConfiguration(){
	ConfItem ci = stack.pop();
	if(ci.getItemType() != ConfItem.CONFIGURATION) {
	    unbalancedStack(ci.getItemType(), ConfItem.CONFIGURATION);
	}
	configuration = ci;
    }

    private void endMenuBar(){
	ConfItem ci = stack.pop();
	if(ci.getItemType() != ConfItem.MENUBAR) {
	    unbalancedStack(ci.getItemType(), ConfItem.MENUBAR);
	}
	ConfItem parent = stack.peek();
	parent.addChild(ci);
	MenuFactory mf = MenuFactory.getInstance();
	mf.addMenuBar((CMenuBar)ci);
    }

    private void endMenu(){
	endCommonElement(ConfItem.MENU);
    }

    private void endMenuItem(){
	endCommonElement(ConfItem.MENUITEM);
    }

    private void endChecked(){
	endCommonElement(ConfItem.CHECKED);
    }

    private void endMenuSeparator(){
	endCommonElement(ConfItem.MENUSEPARATOR);
    }

    private void endField(){
	endCommonElement(ConfItem.FIELD);
    }
    private void endConstructor(){
	endCommonElement(ConfItem.CONSTRUCTOR);
    }

    private void endMethod(){
	ConfItem ci = stack.pop();
	if(ci.getItemType() != ConfItem.METHOD) {
	    unbalancedStack(ci.getItemType(),ConfItem.METHOD);
	}
	ConfItem parent = stack.peek();
	parent.addChild(ci);
	if(((CMethod)ci).isStatic() && 
	   parent.getItemType() == ConfItem.MENUITEM) {
	    ((CMenuItem)parent).setGlobal(true);
	}
    }

    private void endInt(){
	endCommonElement(ConfItem.INT);
    }

    private void endString(){
	endCommonElement(ConfItem.STRING);
    }
    private void endBoolean(){
	endCommonElement(ConfItem.BOOLEAN);
    }
    private void endChar(){
	endCommonElement(ConfItem.CHAR);
    }
    private void endPopupMenu(){
	ConfItem ci = stack.pop();
	if(ci.getItemType() != ConfItem.POPUPMENU) {
	    unbalancedStack(ci.getItemType(), ConfItem.POPUPMENU);
	}
	ConfItem parent = stack.peek();
	parent.addChild(ci);
	MenuFactory mf = MenuFactory.getInstance();
	mf.addPopupMenu((CPopupMenu)ci);
    }

    private void endInitialization(){
	ConfItem ci = stack.pop();
	if(ci.getItemType() != ConfItem.INITIALIZATION) {
	    unbalancedStack(ci.getItemType(), ConfItem.INITIALIZATION);
	}
	ConfItem parent = stack.peek();
	parent.addChild(ci);
	// Initializations that are immediate children of the configuration -element
	// are run immediately when they've been completely parsed
	if(parent.getItemType() == ConfItem.CONFIGURATION) {
	    Initializator.getInstance().initialize((CInitialization)ci,null);
	}
    }
    private void endVisualizations(){
	ConfItem ci = stack.pop();
	if(ci.getItemType() != ConfItem.VISUALIZATIONS) {
	    unbalancedStack(ci.getItemType(), ConfItem.VISUALIZATIONS);
	}
	ConfItem parent = stack.peek();
	parent.addChild(ci);
	RepresentationFactory rf = RepresentationFactory.getInstance();
	rf.addVisualizations((CVisualizations)ci);
    }

    private void endStructure(){
	endCommonElement(ConfItem.STRUCTURE);
    }

    private void endVisualization(){
	endCommonElement(ConfItem.VISUALIZATION);
    }


    private void endPickedVisualizations(){
	ConfItem ci = stack.pop();
	if(ci.getItemType() != ConfItem.PICKEDVISUALIZATIONS) {
	    unbalancedStack(ci.getItemType(), ConfItem.PICKEDVISUALIZATIONS);
	}
	ConfItem parent = stack.peek();
	parent.addChild(ci);
	RepresentationFactory rf = RepresentationFactory.getInstance();
	rf.addPickedVisualizations((CPickedVisualizations)ci);
    }

    private void endCommonElement(int type) {
	ConfItem ci = stack.pop();
	if(ci.getItemType() != type) {
	    unbalancedStack(ci.getItemType(),type);
	}
	ConfItem parent = stack.peek();
	parent.addChild(ci);
    }


    // ---- debug methods ----

    /**
     * Prints the intermediate representation tree to System.out. Used for
     * debug purposes.
     */
    public void printTree(ConfItem ci) {
	printTree(ci, 0);
    }

    private void printTree(ConfItem ci, int offset) {
	String hdr = "";
	for(int i = 0; i < offset; i++) {
	    hdr += "   ";
	}
	System.out.println(hdr + ci.toString());
	for(int i = 0; i < ci.getChildCount(); i++) {
	    printTree(ci.getChild(i), offset+1);
	}
	if(ci.getChildCount() > 0) {
	    System.out.println(hdr+"/" + ci.toString());
	}
    }

    // ---- miscallenous methods ----

    private void unbalancedStack(int ci, int req) {
	    throw new MatrixConfigurationException("Stack error: type " + ci + " item popped from stack. Should have been "
					       + "type " + req);
    }

    // ---- inner classes ----

    class ConfStack extends Vector {
	public ConfItem peek() {
	    return (ConfItem) lastElement();
	}
	public ConfItem pop() {
	    return (ConfItem) remove(size()-1);
	}

	public void push(ConfItem c) {
	    add(c);
	}
    }
}
