/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: MG_Group.java,v 1.15 2004/05/26 08:06:03 vkaravir Exp $ */

package matrix.util.export;

import java.util.*;

/**
 * Stores elements and subgroups for Matrix export.
 */
public class MG_Group {

    /** Contents of this group. */
    private Vector contents;
    /** Subgroups of this group. */
    private Vector groups;
    /** Father of this group. */
    private MG_Group father = null;
    /** VisualType Id. */
    private String vtId = null;
    /** FDT Id. */
    private String fdtId = null;
    /** Name. */
    private String name = "";

    /** 
     * Constructs a new <code>MG_Group</code> instance whose father is
     * itself. 
     */
    public MG_Group() {
	contents = new Vector();
	groups = new Vector();
	father = this;
    }

    /** 
     * Constructs a new <code>MG_Group</code> instance whose father is
     * the given group.
     *
     * @param f Father of the new group.
     */
    public MG_Group(MG_Group f) {
	this();
	father = f;
    }

    /**
     * Returns the father of this group.
     */
    public MG_Group getFather() {
	return father;
    }

    /**
     * Adds a new group as a subgroup of this group. This new group is
     * then returned.
     *
     * @return The new group.
     */
    public MG_Group addGroup() {
	MG_Group newG = new MG_Group(this); //set this as father
	groups.addElement(newG);
	return newG;
    }

    /** 
     * Adds the given element to this group. 
     *
     * @param elem The element to be added.
     * @param step The step that the element is to be added.
     */
    public void add(MG_Element elem, int step) {
	if (step == 0)
	    contents.addElement(elem);
	else { 
	    // Look for the contents to find an element of the same
	    // type than the given element that has the previous step.
	    // If it is found, a new step is added to it instead of
	    // just adding the element to groups contents.
	    Enumeration en = contents.elements();
	    while (en.hasMoreElements()) {
		MG_Element element = (MG_Element)en.nextElement();
		if (elem.getType() == element.getType() && 
		    !element.hasStep(step) && element.hasStep(step - 1)) {
		    element.addStep(elem, step);
		    return;
		}
	    }
	    // origin not found so just add it
	    addNewElement(elem, step);
	}
    }

    /** 
     * Adds the given element to this groups contents. 
     *
     * @param elem The element to be added.
     * @param step The step that the element is to be added.
     */
    protected void addNewElement(MG_Element elem, int step) {
	if (elem == null)
	    return;
	int type = elem.getType();
	MG_Element newElem;
	// Create a new element with elem's properties and make it 
	if (type == MG_Element.MG_TEXT) {
	    MG_Text temp = (MG_Text)elem;
	    newElem = new MG_Text(temp.getString(), temp.getCoord(),
					  temp.getColor(), temp.getFont(),
					  temp.getFontMetrics(), step - 1);
	    newElem.addStep(temp, step);
	} else if (type == MG_Element.MG_LINE) {
	    MG_Line temp = (MG_Line)elem;
	    newElem = new MG_Line(temp.getCoord(), temp.getColor(), 
					  step - 1);
	    newElem.addStep(temp, step);
	} else if (type == MG_Element.MG_RECT) {
	    MG_Rect temp = (MG_Rect)elem;
	    newElem = new MG_Rect(temp.getBounds(), temp.getArcSize(),
					  temp.getColor(), temp.getFillColor(),
					  step - 1);
	    newElem.addStep(temp, step);
	} else if (type == MG_Element.MG_OVAL) {
	    MG_Oval temp = (MG_Oval)elem;
	    newElem = new MG_Oval(temp.getBounds(), temp.getColor(), 
					  temp.getFillColor(), step - 1);
	    newElem.addStep(temp, step);
	} else if (type == MG_Element.MG_ARC) {
	    MG_Arc temp = (MG_Arc)elem;
	    newElem = new MG_Arc(temp.getBounds(), temp.getStartAngle(),
				 temp.getArcAngle(), temp.getColor(), 
				 temp.getFillColor(), step - 1);
	    newElem.addStep(temp, step);
	} else if (type == MG_Element.MG_POLYGON) {
	    MG_Polygon temp = (MG_Polygon)elem;
	    newElem = new MG_Polygon(temp.getPoints()[0], 
				   temp.getPoints()[1], temp.getColor(), 
				   temp.getFillColor(), step - 1);
	    newElem.addStep(temp, step);
	} else
	    return;
	contents.addElement(newElem);
    } 
    

    /**
     * Returns the contents of this group.
     *
     * @return <code>Enumeration</code> of the <code>MG_Element</code>
     * objects this group contains.
     */
    public Enumeration getContents() {
	return contents.elements();
    }

    /**
     * Returns the subgroups of this group. 
     *
     * @return <code>Enumeration</code> of the <code>MG_Group</code>
     * objects this group contains.
     */
    public Enumeration getGroups() {
	return groups.elements();
    }

    /**
     * Sets the Id of this group.
     *
     * @param id The new Id. Index 0 is VisualType Id and index 1 is 
     * FDT Id.
     */
    public void setId(String[] id) {
	vtId = id[0];
	fdtId = id[1];
    }

    /**
     * Returns the Id of this group.
     *
     * @return This group's Id. Index 0 is VisualType Id and index 1 is 
     * FDT Id.
     */
    public String[] getId() {
	return new String[] { vtId, fdtId };
    }

    /**
     * Returns the group with the given VisualType Id. Note, that this
     * method only searches from it's subgroups.
     *
     * @param vtId The VisualType Id that is looked for.
     * 
     * @return The group having the given VisualType Id. Null if no 
     * suitable group is found.  
     */
    public MG_Group getGroupByVtId(String id) {
	if (vtId.equals(id))
	    return this;
	else {
	    int i=0;
	    while (i < groups.size()) {
		MG_Group temp = (MG_Group)groups.elementAt(i);
		if (temp.getId()[0] != null)
		    if (id.startsWith(temp.getId()[0]))
			return temp.getGroupByVtId(id);
		i++;
	    }
	}
	return null;
    }

    /**
     * Returns the group with the given fdtId. The group's VisualType
     * ID must be different than the given vtId and the group must
     * have the given step. Note, that this method only searches from
     * it's subgroups.
     *
     * @param fdtId The FDT Id that is looked for.
     * @param vtId The VisualType Id that the group can't have.
     * @param step The step the group must have.
     * 
     * @return The group matching the given properties. Null if no 
     * suitable group is found.
     */
    public MG_Group getGroupByFdtId(String fdtId, String vtId, int step) {
	if (this.fdtId.equals(fdtId) && !this.vtId.equals(vtId))
	    return this;
	else {
	    MG_Group[] grps = new MG_Group[groups.size()];
	    int i = 0;
	    while (i < groups.size()) {
		int pos = groups.size() - i - 1;
		grps[i] = 
		    ((MG_Group)groups.elementAt(pos)).getGroupByFdtId(fdtId,
					     vtId, step);
		boolean ok = false;
		if (grps[i] != null) {
		    try {
			MG_Element elem = 
			    (MG_Element)grps[i].getContents().nextElement();
			if (elem.hasStep(step))
			    return grps[i];
		    } catch(NoSuchElementException nsee) {
			
		    }
		}
		i++;
	    }
	}
	return null;
    }
    

    /**
     * Returns a <code>String</code> representation of this group.
     */
    public String toString() {
	return "MG_Group " + name + "vtId:" + vtId + " fdtId:" + fdtId + 
	    " " + groups;
    }

    /**
     * Sets the name of the object that is currently drawn. 
     *
     * @param name Name of the object.
     */
    public void setName(String name) {
	this.name = name;
    }

    /**
     * Returns the name of this group.
     */
    public String getName() {
	return name;
    }
}
