/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: MatrixGraphics.java,v 1.20 2004/05/26 08:06:03 vkaravir Exp $ */

package matrix.util.export;

import java.awt.*;
import java.awt.image.ImageObserver;
import java.util.Stack;

import matrix.util.Note;

/**
 * This class collects all shapes drawn to a panel/frame and
 * makes a hierarcial model of them. Just give an instance of 
 * this class as a parameter to, for example, a paint method.
 * After this you can get all the objects drawn by the getContents()
 * method. 
 *
 * Note:
 * <ul><li> <code>drawImage(..)</code> methods not implemented</li>
 * <li>draws outside the current clip, so the result might look a
 * bit different than what Java shows (except for non-filled 
 * rectangles, ovals and strings)</li>
 * <li>clearRect(..) not implemented</li></ul>
 * @see java.awt.Graphics
 * @author Ville Karavirta
 */
public class MatrixGraphics extends Graphics {

    /** Coordinates of the current origo in the original coordination */
    private int x, y;
    /** Current clip */
    private Rectangle clip;
    /** Original clip */
    private Rectangle origClip;
    /** Current color */
    private Color color = Color.black;
    /** Graphics instance to handle fonts */
    private Graphics fontMgr;
    /** Saves all shapes drawn to this graphics */
    private MatrixGraphicsContents contents = 
	      new MatrixGraphicsContents();
    /** Translations done to this graphics */
    private Stack translations = new Stack();
    /** Clips */
    private Stack clips = new Stack();

    /**
     * Constructs a new MatrixGraphics instance.
     *
     * @param clip Rectangle representing the area of the new
     * MatrixGraphics instance
     * @param g Instance of java.awt.Graphics
     */
    public MatrixGraphics(Rectangle clip, Graphics g) {
	x = 0;
	y = 0;
	fontMgr = g;
	this.clip = clip;
	origClip = new Rectangle(clip);
	clips.push(clip);
	translations.push(new Integer(0));
    }

    /** 
     * Should clear the given area, now it does nothing. If you
     * want to clear contents, just create new MatrixGraphics 
     * instance.
     */
    public void clearRect(int x, int y, int width, int height) {
    }

    /**
     * Intersects the current clip with the specified rectangle. The 
     * resulting clipping area is the intersection of the current clipping
     * area and the specified rectangle.
     *
     * @param x the x coordinate of the rectangle to intersect the clip with 
     * @param y the y coordinate of the rectangle to intersect the clip with
     * @param width the width of the rectangle to intersect the clip with
     * @param height  the height of the rectangle to intersect the clip with
     */
    public void clipRect(int x, int y, int width, int height) {
	clip = clip.intersection(new Rectangle(x + this.x, y + this.y, 
					       width, height));
    }

    /** Not implemented, do not use. */
    public void copyArea(int x, int y, int width, int height, int dx, 
			 int dy) {
    }

    public Graphics create() { 
	contents.levelDown();
	clips.push(clip);
	translations.push(new Integer(0));
	return this;
    }

    public Graphics create(int x, int y, int width, int height) {
	contents.levelDown();
	clips.push(clip); // save the old clip
	Rectangle newClip = new Rectangle(x + this.x, y + this.y, 
					  width, height);
	if (!clip.intersects(newClip))
	    clip = new Rectangle(0, 0, 0, 0);
	else
	    clip = clip.intersection(newClip);
	translations.push(new Integer(0));
	translate(x, y);
	return this;
    }

    public void dispose() {
	if (clips.isEmpty())
	    return;
	clip = (Rectangle)clips.pop();
	contents.levelUp();
	int trans = ((Integer)translations.pop()).intValue();
	int i = 0;
	while (i < trans) { //undo all translations
	    int[] pop = (int[])translations.pop();
	    this.x = this.x - pop[0];
	    this.y = this.y - pop[1];
	    trans--;
	}
    }

    /*
      public void draw3DRect(int x, int y, int width, int height, 
      boolean raised) { // XXX Should be implemented
      }
    */

    public void drawArc(int x, int y, int width, int height, int startAngle,
			int arcAngle) {
	if (width < 0 || height < 0 || arcAngle == 0) 
	    return;
	x = x + this.x;
	y = y + this.y;
	contents.addArc(x, y, width, height, startAngle, arcAngle, 
			color, null);
    }

    /** Not implemented, do not use. */
    public boolean drawImage(Image img, int x, int y, Color bgcolor, 
			     ImageObserver observer) {
	Note.out(this, "Method drawImage not implemented!");
	return false;
    }

    /** Not implemented, do not use. */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer) {
	Note.out(this, "Method drawImage not implemented!");
	return false;
    }

    /** Not implemented, do not use. */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     Color bgcolor, ImageObserver observer) {
	Note.out(this, "Method drawImage not implemented!");
	return false;
    }

    /** Not implemented, do not use. */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     ImageObserver observer) {
	Note.out(this, "Method drawImage not implemented!");
	return false;
    }

    /** Not implemented, do not use. */
    public boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2, Color bgcolor,
			     ImageObserver observer) {
	Note.out(this, "Method drawImage not implemented!");
	return false;
    }

    /** Not implemented, do not use. */
    public boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer) {
	Note.out(this, "Method drawImage not implemented!");
	return false;
    }

    /**
     * Draws a line from point (x1, y1) to point (x2, y2) using
     * the current color.
     *
     * @param x1 x coordinate of the starting point
     * @param y1 y coordinate of the starting point
     * @param x2 x coordinate of the end point
     * @param y2 y coordinate of the end point
     */
    public void drawLine(int x1, int y1, int x2, int y2) {
	x1 = x1 + this.x;
	y1 = y1 + this.y;
	x2 = x2 + this.x;
	y2 = y2 + this.y;
	if (!rectCheck(Math.min(x1, x2), Math.min(y1, y2), Math.abs(x2 - x1),
		       Math.abs(y2 - y1))) {
	    int[] newLine = rectCut(x1, y1, x2 - x1, y2 - y1);
	    if (newLine[2] == 0 && newLine[3] == 0)
		return;
	    x1 = newLine[0];
	    y1 = newLine[1];
	    x2 = x1 + newLine[2];
	    y2 = y1 + newLine[3];
	}
	contents.addLine(x1, y1, x2, y2, color);
    }

    /**
     * Draws the outline of an oval. 
     * 
     * @param x the x coordinate of the upper left corner of the oval.
     * @param y the y coordinate of the upper left corner of the oval.
     * @param width the width of the oval
     * @param height the height of the oval
     */
    public void drawOval(int x, int y, int width, int height) {
	x = x + this.x;
	y = y + this.y;
	if (!rectCheck(x, y, width, height) || width < 0 || height < 0) 
	    return;
	contents.addOval(x, y, width, height, color, null);
    }
    
    public void drawPolygon(int[] xPoints, int[] yPoints, int nPoints) {
	// XXX Should check if inside current clip!!!
	for (int i=0; i < nPoints; i++) {
	    xPoints[i] = xPoints[i] + this.x;
	    yPoints[i] = yPoints[i] + this.y;
	}
	contents.addPolygon(xPoints, yPoints, nPoints, color, null);
    }

    /** Not implemented, do not use */
    public void drawPolyline(int [] xPoints, int[] yPoints, int nPoints) {
	// XXX should be implemented
	Note.out(this, "drawPolyline(..) not implemented");
    }

    /** Draws the outline of the specified rectangle. The left and right
     * edges of the rectangle are at x and x + width. The top and bottom
     * edges of the rectangle are at y and y + height. The rectangle is
     * drawn using the current color.
     *
     * @param x the x coordinate of the left edge
     * @param y the y coordinate of the top edge
     * @param width the width of the rectangle
     * @param height the height of the rectangle
     */    
    public void drawRect(int x, int y, int width, int height) {
	x = x + this.x;
	y = y + this.y;
	if (width < 0 || height < 0)
	    return;
	if (!rectCheck(x, y, width, height)) {
	    int[] newRect = rectCut(x, y, width, height);
	    if (newRect[2] == 0 && newRect[3] == 0)
		return;
	    x = newRect[0];
	    y = newRect[1];
	    width = newRect[2];
	    height = newRect[3];
	}
	contents.addRect(x, y, width, height, 0, 0, color, null);
    }
    
    
    public void drawRoundRect(int x, int y, int width, int height, 
			      int arcWidth, int arcHeight) {
	if (width < 0 || height < 0)
	    return;
	x = x + this.x;
	y = y + this.y;
	contents.addRect(x, y, width, height, Math.abs(arcWidth), 
			 Math.abs(arcHeight), color, null);
    }

    /** Not implemented, do not use. */
    public void drawString(java.text.AttributedCharacterIterator iterator, 
			   int x, int y) {
    }

    /** Draws the text given by the specified string, using current
     * font and color. The baseline of the first character is at 
     * position (x, y) in the current coordinate system.
     *
     * @param str the string to be drawn
     * @param x the x coordinate
     * @param y the y coordinate
     */
    public void drawString(String str, int x, int y) {
	if (str.trim().equals("")) 
	    return;
	int width = getFontMetrics().stringWidth(str);
	int height = getFontMetrics().getHeight();
	x = x + this.x;
	y = y + this.y;
	if (!clip.intersects(new Rectangle(x, y - height + 3, width, height)))
	    return; // if entirely outside the current clip, don't draw
	contents.addText(str, x, y, color, fontMgr.getFont(), 
			 fontMgr.getFontMetrics());
    }

    /*
      public void fill3DRect(int x, int y, int width, int height, 
      boolean raised) {  //XXX Should be implemented??!!
      }
    */


    public void fillArc(int x, int y, int width, int height, int startAngle,
			int arcAngle) {
	if (width < 0 || height < 0 || arcAngle == 0) 
	    return;
	x = x + this.x;
	y = y + this.y;
	contents.addArc(x, y, width, height, startAngle, arcAngle, 
			color, color);
    }

    public void fillOval(int x, int y, int width, int height) {
	x = x + this.x;
	y = y + this.y;
	if (!rectCheck(x, y, width, height) || width <= 0 || height <= 0) 
	    return;
	contents.addOval(x, y, width, height, color, color);
    }

    public void fillPolygon(int[] xPoints, int[] yPoints, int nPoints) {
	// XXX Should check if inside current clip!!!
	for (int i=0; i < nPoints; i++) {
	    xPoints[i] = xPoints[i] + this.x;
	    yPoints[i] = yPoints[i] + this.y;
	}
	contents.addPolygon(xPoints, yPoints, nPoints, color, color);
    }
    /*
      public void fillPolygon(Polygon p) {
    }*/

    public void fillRect(int x, int y, int width, int height) {
	x = x + this.x;
	y = y + this.y;
	if (width <= 0 || height <= 0)
	    return;
	if (!rectCheck(x, y, width, height)) {
	    int[] newRect = rectCut(x, y, width, height);
	    if (newRect[2] == 0 && newRect[3] == 0)
		return;
	    x = newRect[0];
	    y = newRect[1];
	    width = newRect[2];
	    height = newRect[3];
	}
	contents.addRect(x, y, width, height, 0, 0, color, color);
    }

    public void fillRoundRect(int x, int y, int width, int height, 
			      int arcWidth, int arcHeight) {
	if (width < 0 || height < 0) 
	    return;
	if (Math.abs(arcWidth) > width)
	    arcWidth = width;
	if (Math.abs(arcHeight) > height) 
	    arcHeight = height;
	x = x + this.x;
	y = y + this.y;
	contents.addRect(x, y, width, height, Math.abs(arcWidth), 
			 Math.abs(arcHeight), color, color);
    }

    
    public Shape getClip() {
	// return the original clip, for some reason it doesn't work
	// if this returns the current clip
	return origClip;
    }
    
    public Rectangle getClipBounds() {
	return origClip;
    }
    
    public Rectangle getClipBounds(Rectangle r) {
 	return r;
    }

    public Color getColor() {
	return color;
    }
    
    public Font getFont() {
 	return fontMgr.getFont();
    }

    public FontMetrics getFontMetrics() {
	return fontMgr.getFontMetrics();
    }

    public FontMetrics getFontMetrics(Font f) {
	return fontMgr.getFontMetrics(f);
    }

    /*
      public boolean hitClip(int x, int y, int width, int height) {
      }
    */

    /** Not implemented, do not use */
    public void setClip(int x, int y, int width, int height) {
    }

    /** Not implemented, do not use */
    public void setClip(Shape clip) {
    }

    public void setColor(Color c) {
	this.color = c;
    }

    public void setFont(Font font) {
	fontMgr.setFont(font);
    }

    /** Not implemented, do not use. */
    public void setPaintMode() {
    }

    /** Not implemented, do not use. */
    public void setXORMode(Color c1) {
    }

    public void translate(int x, int y) {
	int trans = ((Integer)translations.pop()).intValue();
	trans++;
	this.x = this.x + x;
	this.y = this.y + y;  
	translations.push(new int[] {x, y});
	translations.push(new Integer(trans));
    }

    /**
     * Returns the contens of this instance.
     */
    public MatrixGraphicsContents getContents() {
	return contents;
    }
    
    /**
     * Check if the given area is inside the current clip.
     *
     * @param x x coordinate of the area (upper left corner)
     * @param y y coordinate of the area (upper left corner)
     * @param w width of the area
     * @param h height of the area
     * 
     * @return true, if the area is inside the current clip, false 
     * otherwise
     */
    private boolean rectCheck(int x, int y, int w, int h) {
	return contains(clip, x, y, w, h);
    }

    /**
     * Modifies the given area so that it is inside the current clip.
     *
     * @param x x coordinate of the area (upper left corner)
     * @param y y coordinate of the area (upper left corner)
     * @param w width of the area
     * @param h height of the area
     * 
     * @return Int array containing the new coordinates and size. New
     * x coodinate is at index 0, y at 1, width at 2 and height at 3.
     */ 
    private int[] rectCut(int x, int y, int w, int h) {
 	int clipx = clip.x;
	int clipy = clip.y;
	int clipw = clip.width;
	int cliph = clip.height;
	if (!clip.intersects(new Rectangle(x, y, w, h)))
	    return new int[] {0, 0, 0, 0};
	if (x < clipx) {
	    w = w - (clipx - x);
	    x = clipx;
	}
	if ((x + w) > clipx + clipw)
	    w = clipx + clipw - x;
	if (y < clipy) { 
	    h -= clipy - y; 
	    y = clipy; 
	} 
	if ((y + h) > clipy + cliph)
	    h = clipy + cliph - y;
	return new int[] {x, y, w, h};
    }

    /**
     * Checks if the given <code>Rectangle</code> contains the given 
     * area. Returns true if it does, false otherwise.
     *
     * @param rect The <code>Rectangle</code>
     * @param x x-coordinate of the area to be tested
     * @param y y-coordinate of the area to be tested
     * @param w width of the area to be tested
     * @param h height of the area to be tested
     * 
     * @return True if the given area is inside the <code>Rectangle</code>,
     * false otherwise.
     */
    private boolean contains(Rectangle rect, int x, int y, int w, int h) {
	//if rectangle contains all the corners, it must contain the
	//whole area
	boolean corner1 = rect.contains(x, y);
	boolean corner2 = rect.contains(x, y + h);
	boolean corner3 = rect.contains(x + w, y);
	boolean corner4 = rect.contains(x + w, y + h);
	return (corner1 && corner2 && corner3 && corner4);
    }

    /**
     * Sets the Id of the object that is currently drawn. These are needed
     * if SVG-export is to be made. 
     *
     * @param id The new Id. Index 0 is VisualType Id, index 1 is FDT Id
     * and index 2 is Original FDT's Id.
     */
    public void setId(int[] id) {
	contents.setId(id);
    }

    /**
     * Moves to the next step in animation. Use this method to inform
     * <code>MatrixGraphics</code> that another step in the animation
     * is beginning.
     */
    public void nextStep() {
	color = Color.black;
	translations = new Stack();
	translations.push(new Integer(0));
	clips = new Stack();
	clips.push(origClip);
	contents.nextStep();
    }

    /**
     * Sets the name of the object that is currently drawn. 
     *
     * @param name Name of the object.
     */
    public void setName(String name) {
	contents.setName(name);
    }

    /**
     * Sets the scale used in export. Initially it is 1.
     *
     * @param scale The new scale.
     */
    public void setScale(double scale) {
	contents.setScale(scale);
    }
}
