/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: SVGUtil.java,v 1.8 2004/05/25 13:28:30 kstalnac Exp $ */

package matrix.util.export;

public class SVGUtil {

    /** Length of one smooth transition in animation. In seconds. */
    private float stepLength;
    /** Pause between steps. In seconds. */
    private float pause;
    /** Number of the last step in the animation. */
    private int maxStep;
    /** Whether to add scripts and animator panel to the animation. */
    private boolean panel;
    /** Array of steps used in animation creation. */
    // Used in method animate(....)
    private boolean[] forward;

    public SVGUtil(float stepLength, float pause, int maxStep,
		   boolean panel) {
	this.stepLength = stepLength;
	this.pause = pause;
	this.maxStep = maxStep;
	this.panel = panel;
	forward = new boolean[maxStep+1];
    }

    /** Returns the length of one step. Time is in seconds. */
    public float getStepLength() { return stepLength; }
    /** Returns the length of the pause between steps. In seconds. */
    public float getPause() { return pause; }
    /** Returns the boolean representiong whether or not to add acripts
	and the animator panel to the SVG animation. */
    public boolean addPanel() { return panel; }
    /** Returns the number of the last step in this animation. */
    public int getMaxStep() { return maxStep; }

    /**
     * Returns the number of the last step used in this
     * animation. This is not the same as the number obtained with
     * method <code>getMaxStep()</code>.  This method returns the
     * number of the last step where something actually happens. Steps
     * where nothing happens are left out from the animation.
     *
     * @return Number of the last step used in this animation.
     */
    public int getLastStep() {
	int i = maxStep;
	while (!forward[i]) i--;
	return i+1;
    }

    /**
     * Returns the number of the first step used. This isn't
     * necessarily 0, since steps where nothing happens are left out
     * from the animation. 
     * @return Number of the first step used.  
     */
    public int getFirstStep() {
	int i = 0;
	while(!forward[i]) i++;
	return i;
    }

    /**
     * Returns all the steps used in the animation. Steps where
     * nothing happens are left out of the animation. Steps are
     * returned as a <code>String</code> which is in format "-1, 1, 2,
     * 3". There is always the extra -1 in the string.
     *
     * @return All the steps used in the animation as String.
     */
    public String getUsedSteps() {
	String steps = "-1";
	for (int i=0; i < forward.length; i++)
	    if (forward[i])
		steps += "," + i;
	return steps;
    }

    /** 
     * Creates an SVG element as String from the given values. 
     *
     * @param attribName Name of the attribute to animate.
     * @param step Number of the step that is animated.
     * @param from The starting value of the attribute.
     * @param to The end value of the attribute.
     * @return String representing an animate element of the
     * given values.
     */
    public String animate(String attribName, int step, String from,
			  String to) {
	return animate("animate", attribName, step, from, to, null);
    }

    /** 
     * Creates an SVG element as String from the given values. You can
     * specify the name of the element.
     *
     * @param tagName Name of the element to be created.
     * @param attribName Name of the attribute to animate.
     * @param step Number of the step that is animated.
     * @param from The starting value of the attribute.
     * @param to The end value of the attribute.
     * @return String representing an animate element of the
     * given values.
     */
    public String animate(String tagName, String attribName, int step, 
			  String from, String to) {
	return animate(tagName, attribName, step, from, to, null);
    }

    /** 
     * Creates an SVG animate element as String from the given
     * values. You can specify additional attributes for the element
     * by using the attrs parameter. However, this parameter shouldn't
     * be used for attributes from, to, attributeName and dur.
     *
     * @param attribName Name of the attribute to animate.
     * @param step Number of the step that is animated.
     * @param from The starting value of the attribute.
     * @param to The end value of the attribute.
     * @param attrs Additional parameters to the element.
     * @return String representing an animate element of the
     * given values.
     */
    public String animate(String attribName, int step, String from, 
			  String to, String attrs) {
	return animate("animate", attribName, step, from, to, attrs);
    }

    /** 
     * Creates an SVG element as String from the given values. You can
     * specify additional attributes for the element by using the
     * attrs parameter. However, this parameter shouldn't be used for
     * attributes from, to, attributeName and dur. You can also
     * specify name of the element.
     *
     * @param tagName Name of the element to be created.
     * @param attribName Name of the attribute to animate.
     * @param step Number of the step that is animated.
     * @param from The starting value of the attribute.
     * @param to The end value of the attribute.
     * @param attrs Additional parameters to the element.
     * @return String representing an animate element of the
     * given values. 
     */
    public String animate(String tagName, String attribName, int step,
			  String from, String to, String attrs) {
	if (step > maxStep + 1)
	    return "";
	String animate = "<" + tagName + " attributeName=\"" + attribName + 
	    "\" fill=\"freeze\" dur=\"" + stepLength + "s\" ";
	String animB = animate;
	animate += ((from != null)?"from=\"" + from + "\"" : "") + " to=\"" + 
	    to + "\"";
	if (attrs != null)
	    animate += " " + attrs;
	animate += " begin=\"";
	if (panel) {
	    animB += "from=\"" + to + "\" to=\"" + from + "\" begin=\"";
	    if (forward[step]) {
		animate += "anim" + step + ".begin\"";
		animB += "animb" + step + ".begin\"";
	    } else {
		animate += "indefinite\" id=\"anim" + step + "\" " +
		    "onend=\"animationEnd()\"";
		animB += "indefinite\" id=\"animb" + step + "\" " +
		    "onend=\"animationEnd()\""; 
		forward[step] = true;
	    }
	    animate += "/>" + animB;
	} else {
	    float beginTime = (step)*(pause + stepLength) + pause;
	    animate += beginTime + "s\"";
	}
	animate += "/>";
	return animate;
    }

    /** 
     * Creates an SVG animate element as String from the given
     * values.
     * @param attribName name of the attribute to animate
     * @param begin time of the begin (in seconds)
     * @param dur the duration of the animation (in seconds)
     * @param from the starting value of the attribute
     * @param to the end value of the attribute
     * @return string representing an animate element of the
     * given values */
    public static String animateElement(String attribName, float begin, 
					float dur, String from, String to) {
	String animate = "<animate attributeName=\"" + attribName +
	    "\" begin=\"" + begin + "s\" dur=\"" + dur +
	    "s\" end=\"" + (begin+dur) + "s\" from=\"" + from + "\" to=\"" + 
	    to + "\" fill=\"freeze\"/>";
	return animate;
    }

    /** 
     * Creates an SVG animateColor element as String from the
     * given values.
     * @param attribName name of the attribute to animate
     * @param begin time of the begin (in seconds)
     * @param dur the duration of the animation (in seconds)
     * @param from the starting color
     * @param to the end color
     * @return string representing an animateColor element of the
     * given values
     */
    public static String animateColor(String attribName, float begin, 
				      float dur, String from, String to) {
	String animate = "<animateColor attributeName=\"" + attribName + 
	    "\" begin=\"" + begin + "s\" dur=\"" + dur + "s\" end=\"" +
	    (begin+dur) + "s\" from=\"" + 
	    from + "\" to=\"" + to + "\" fill=\"freeze\"/>";
	return animate;
    }

    /** 
     * Creates an SVG animate element as String from the
     * given values.
     * @param attribName name of the attribute to animate
     * @param begin time of the begin (in seconds)
     * @param dur the duration of the animation (in seconds)
     * @param to the end color
     * @param additive controls whether or not the animation is additive
     * @return string representing an animateColor element of the
     * given values
     */
    public static String animateAdditive(String attribName, float begin, 
				      float dur, String to, String additive) {
	String animate = "<animate attributeName=\"" + attribName + 
	    "\" begin=\"" + begin + "s\" dur=\"" + dur + "s\" end=\"" +
	    (begin+dur) + "s\" to=\"" +
	    to + "\" fill=\"freeze\" additive=\"" + additive + "\"/>";
	return animate;
    } 

}
