/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: ASCIISaver.java,v 1.13 2004/06/17 12:11:09 kstalnac Exp $ */

package matrix.util.io;

import java.awt.Component;
import java.io.*;
import java.util.*;

import matrix.visual.*;

/**
 * Contains methods to save structures as an ASCII file.
 * This saves also all the inner structures. All the structures are saved in 
 * a same file. 
 */
public class ASCIISaver {

    private Component[] links;
    private Component[] items;
    private File file;
    private String saveName;

    /**
     * Constructs ASCIISaver which saves the given structure as an ASCII file.
     *
     * @param links references of the structure
     * @param items nodes of the structure 
     * @param saveFile file where to save the structure
     */
    public ASCIISaver(Component[] links, Component[] items, File saveFile) {
	this.links = links;
	this.items = items;
	file = saveFile;
	this.saveName = file.getName();
    }


    /**
     * Constructs ASCIISaver which purpose is to produce string representation
     * of the structure. This should be called only from 
     * getSaveNameAndDoInnerSave-method or from getTableInASCIIFormat-method.
     * When you have to save certain structure in a file use the other constructor.
     * This is used when saving the inner structures and this only produces
     * the string representation of the structure and no new files are created. 
     * 
     * @param links references of the structure
     * @param items nodes of the structure
     * @param saveName name for the structure which is going to be saved
     */
    public ASCIISaver(Component[] links, Component[] items, String saveName) {
	this.links = links;
	this.items = items;
	this.saveName = saveName;
    }


    Component[] cs = null; // components to be saved
    
    /**
     * Constructs ASCIISaver which saves all the given structures to the same
     * ASCII-file.
     *
     * @param cs components which contain outermost structures
     * @param saveFile file where to save structures
     */
    public ASCIISaver(Component[] cs, File saveFile) {
	this.cs = cs;
	file = saveFile;
	this.saveName = file.getName();
    }


    VisualContainer visualcontainer = null;
    /**
     * Constructs ASCIISaver which saves the structure in the given component 
     * (it is infact VisualContainer).
     * This is used because in this way we can obtain also information about
     * the visualization of the container.
     *
     * @param c VisualContainer which needs to be saved
     * @param saveName name for the structure which is going to be saved
     */
    public ASCIISaver(Component c, String saveName) {
	this.visualcontainer = (VisualContainer)c;
	this.items = visualcontainer.getItems();
	this.links = visualcontainer.getLinks();
	this.saveName = saveName;
    }
    

    /**
     * Saves all the outermost structures which are in Component[] cs to 
     * the same file. All the inner structures are also saved.
     *
     */
    public void saveStructuresToFile() {
	String structuresInASCII = "";
	VisualType vt = null;
	String structureName = saveName;
	int numberOfFoundStructures = 0;
	
	for (int i=0; i < cs.length; i++) {
	    //System.out.println(cs[i]);
	    if(cs[i] instanceof VisualType) {
		vt = (VisualType)cs[i];
		if(vt instanceof VisualContainer) {
		    structureName = saveName.concat("#" + (numberOfFoundStructures + 1));
		    //ASCIISaver saver = new ASCIISaver(vt.getLinks(), 
		    //			      vt.getItems(), 
		    //			      structureName);
		    ASCIISaver saver = new ASCIISaver(vt, structureName);
		    structuresInASCII = structuresInASCII.concat(structureName + "\n");
		    if(vt instanceof VisualArray) 
			structuresInASCII = structuresInASCII.concat(saver.getTableInASCIIFormat());
		    else
			structuresInASCII = structuresInASCII.concat(saver.getGraphInASCIIFormat());
		
		    numberOfFoundStructures++; //found new outermost structure
		}
	    }
	}
	try {
	    if(structuresInASCII != "") { 
		FileWriter fw = new FileWriter(file);
		BufferedWriter bw = new BufferedWriter(fw);
		PrintWriter out = new PrintWriter(bw, true);
		out.println("#matrix structures");
		StringTokenizer st = new StringTokenizer(structuresInASCII, "\n", false);
		while(st.hasMoreTokens()) {
		    out.println(st.nextToken());
		}
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    
    /**
     * Saves nodes of the graph as adjacency-list format into ASCII file.
     * Saves also all the inner structures.
     */
    public void saveGraphToASCIIFile() {
	try {
	    String gia = getGraphInASCIIFormat();
	    FileWriter fw = new FileWriter(file);
	    BufferedWriter bw = new BufferedWriter(fw);
	    PrintWriter out = new PrintWriter(bw, true);

	    StringTokenizer st = new StringTokenizer(gia, "\n", false);
	    while(st.hasMoreTokens()) {
		out.println(st.nextToken()); //printing the string in a file one "line" (token) at a time
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    LinkedList structureNames = new LinkedList();
    //string representations of the structures
    HashMap structuresInASCII = new HashMap();


    /**
     * Returns the graph in a string. String contains also information about 
     * visualization of the structure. Saves also all the inner structures.
     *
     * @return the graph in a string.
     */
    public String getGraphInASCIIFormat() {
	String gia = null; // graph in ascii

	VisualReference vr = null;
	VisualComponent source = null;
	VisualComponent target = null;
	String key1;
	String key2;

	//Finding the name of the representation of the structure.
	String repreTemp = visualcontainer.getNameOfRepresentation();
	if(repreTemp == "")
	    repreTemp = visualcontainer.getRepresentationName();
	if(repreTemp.equals("Graph"))
	   repreTemp = "layered graph";
	if(repreTemp.equals("Layered Tree"))
	    repreTemp = "layered tree";

	//Adding information about visualization of the structure.
	gia = "#representation " + repreTemp + "\n";
	gia = gia.concat("#rotated " + visualcontainer.isRotated() + "\n");
	gia = gia.concat("#minimized " + visualcontainer.isMinimized() + "\n");

	gia = gia.concat("#matrix graph adjacency-list\n");
	int arraySize = items.length;
	LinkedList[] array;
	int numberOfNulls = 0;

	// counting number of components pointing null
	for(int i=0; i < arraySize; i++) {
	    if(((VisualComponent)items[i]).getKey() == null)
		numberOfNulls++;
	}
	// Creates an array containing LinkedLists for nodes and their 
	// adjacents. Size of the array is number of non-null components.
	array = new LinkedList[arraySize-numberOfNulls];
	//this is used for saving the keys in a right place in the array
	int foundNulls = 0;
	for(int i=0; i < arraySize; i++) {
	    if(((VisualComponent)items[i]).getKey() != null){		
		VisualComponent vc = (VisualComponent)items[i];
		array[i-foundNulls] = new LinkedList();
		array[i-foundNulls].add(vc);
	    }
	    else {
		foundNulls++; // found component which points to null
	    }
	}
   
	for(int i=0; i < links.length; i++) {
	    vr = (VisualReference)links[i];
	    source = vr.getSource();
	    target = vr.getTarget();
	    if (source.getKey() != null && target.getKey() != null) {
		//finding the right place to add adjacent
		for(int j=0; j < array.length; j++)
		    if (array[j].getFirst() == source)
			array[j].add(target);
	    }
	}

	// writing nodes into adjacency-list format in an ASCII file
	for(int i=0; i < array.length; i++) {
	    if (!(((VisualComponent)array[i].getFirst()).getKey().toString().equals("Forest"))) {
		String line = null;
		//System.out.println("line number " + i);
		line = getSaveNameAndDoInnerSave((VisualComponent)array[i].getFirst()) + ":";

		ListIterator listiter = array[i].listIterator();
		listiter.next();
		while(listiter.hasNext()) {
		    line = line + getSaveNameAndDoInnerSave((VisualComponent)listiter.next());
		    if (listiter.hasNext())
			line = line + " ";
		}
		
		if(line.endsWith(":")) {
		    int index = line.indexOf(":");
		    line = line.substring(0, index);
		}
		gia = gia.concat(line + "\n");
	    }
	}
	gia = gia.concat("#EOS\n"); // structure separator
	
	//adding all the inner structures in the end of the string
	ListIterator listiter = structureNames.listIterator();
	while(listiter.hasNext()) {
	    String sname = (String)listiter.next();
	    gia = gia.concat("   " + sname + "\n"); //indentation and name of the structure
	    String temp = (String)structuresInASCII.get(sname);
	    
	    StringTokenizer st = new StringTokenizer(temp, "\n", false);
	    while(st.hasMoreTokens()) {
		//space characters are indentation
		gia = gia.concat("   " + st.nextToken() + "\n");
	    }
	}
	return gia;
    }



    int postfix = 1;

    /**
     * Returns name of the component and if the component contains inner structures,
     * they are also saved into structuresInASCII.
     *
     *
     * @param vc VisualComponent which is going to be saved
     * @return name of the VisualComponent
     */
    private String getSaveNameAndDoInnerSave(VisualComponent vc) {
	String name = null;
	VisualType vt = null;
	File saveFile = null;

	vt = (VisualType)vc.getOther(0); // getting item from VisualComponent
	if (vt instanceof VisualKey) { // item is just a normal Key
	    name = getSaveNameForKey(vc); // getting name for the key
	    return name;
	}
	else {
	    //item contained inner structure
	    name = getSaveNameForStructure(vc); //getting name for the inner structure
	    if(name == null) { // first occurrence of this structure
		name = saveName + "_" + postfix; // deciding new name
		setSaveNameForStructure(vc, name); // saving the new name
		postfix++;
	    }
	    //ASCIISaver saver2 = new ASCIISaver(vt.getLinks(), vt.getItems(), name);
	    ASCIISaver saver2 = new ASCIISaver(vt, name);
	    if (!(structureNames.contains(name))) {
		structureNames.add(name); //adding name of structure, each
		                          //name is added only one time
	    }

	    if (vt instanceof VisualArray) {
		//System.out.println("vt was VisualArray: getsavenameand...");
		String temp = saver2.getTableInASCIIFormat(); //getting table in ascii-string
		structuresInASCII.put(name, temp); //saving the string
	    }
	    else {
		//System.out.println("was neither VisualKey nor VisualArray " +
		//		   "getsavenamand..");
		String temp = saver2.getGraphInASCIIFormat(); //getting graph in ascii-string
		structuresInASCII.put(name, temp); //saving the string
	    }
	    return name;
	}
    }


    HashMap structureNameMap = new HashMap();

    /**
     * Returns the name of the structure.
     *
     * @param vc visualcomponent which contains the structure
     * @return name of the structure
     */
    private String getSaveNameForStructure(VisualComponent vc) {
	String tempName = null;
	Object temp = structureNameMap.get(vc);
	if (temp != null)
	    tempName = (String)temp;

	return tempName;
    }

    /**
     * Sets the name of the structure.
     *
     * @param vc visualcomponent which contains the structure
     * @param name name of the structure
     */
    private void setSaveNameForStructure(VisualComponent vc, String name) {
	structureNameMap.put(vc, name);
    }


    HashMap componentMap = new HashMap();

    /**
     * Returns the name of the key. If the same key has been found earlier but the
     * visualcomponents are not the same, then an index is added to the end of the name
     * of the key so that the keys can be differentiated when opening the file.
     *
     * @param vc visualcomponent which contains the key
     * @return name of the key
     */
    private String getSaveNameForKey(VisualComponent vc) {
	int index = 1;	
	String name2 = null;
	String tempName = null;

	if(componentMap.get(vc.getKey().toString()) == null) { // first occurrence of this key
	    //System.out.println("first attempt " + vc.getKey());
//	    componentMap.put(vc.getKey().toString(), vc);
		componentMap.put(vc.getKey().toString(), vc.getStructure());	    
	    //name2 = vc.getKey().toString();
	    tempName = vc.getKey().toString();
	    name2 = processName(tempName);
	}
//	else if(componentMap.get(vc.getKey().toString()) == vc) { // this visualcomponent is already in the hashmap
	else if(componentMap.get(vc.getKey().toString()) == vc.getStructure()) { // this visualcomponentn is already in the hashmap
	    //System.out.println("found from the hashmap " + vc.getKey());
	    //name2 = vc.getKey().toString();
	    tempName = vc.getKey().toString();
	    name2 = processName(tempName);
	}
	else {
	    index++;
	    // finding the name for the key in this visualcomponent 
	    while((componentMap.get(vc.getKey() + "_" + index) != null) && 
//		(componentMap.get(vc.getKey() + "_" + index) != vc)) {
		(componentMap.get(vc.getKey() + "_" + index) != vc.getStructure())) {
		//System.out.println("was neither empty nor right " + 
		//		   vc.getKey() + "_" + index);
		index++;
	    }
	    if(componentMap.get(vc.getKey() + "_" + index) == null) {// found a name for the key of this component
		//System.out.println("new place was empty " +
		//		   vc.getKey() + "_" + index);
//		componentMap.put(vc.getKey() + "_" + index, vc);
		componentMap.put(vc.getKey() + "_" + index, vc.getStructure());
	    }
	//System.out.println(vc.getKey() + "_" + index);
	    //name2 = vc.getKey() + "_" + index;  
	    tempName = vc.getKey().toString();
	    name2 = processName(tempName) + "_" + index;
	}
	
	return name2;
    }


    /**
     * Processes the name of the key. This makes it possible that the key
     * can have special characters (" ", "_", "#", "'") and they are still
     * saved and opened correctly. This method adds escape character "'" in
     * front of special character so that entire key can be correctly recognized 
     * when opening the keys of the saved structure.
     *
     * @param name name of the original key to be saved
     * @return name of the key in a saving format
     */
    private String processName(String name) {
	String tempName = "";
	String temp = null;
	int index;
	if(name != "") {
	    for(index = 0; index < name.length(); index++) {
		if(index == (name.length() - 1))
		    temp = name.substring(index);
		else
		    temp = name.substring(index, index + 1);
		
		if(temp.equals("'"))
		    tempName = tempName.concat("''");
		else if (temp.equals("_"))
		    tempName = tempName.concat("'_");
		else if (temp.equals(" "))
		    tempName = tempName.concat("' ");
		else if (temp.equals("#"))
		    tempName = tempName.concat("'#");
		else
		    tempName = tempName.concat(temp);
	    }
	}
	return tempName;
    }



    /**
     * Saves components of the table as an ASCII file.
     * Saves also all the inner structures.
     *
     */
    public void saveTableToASCIIFile() {
	try {
	    String tableInASCII = getTableInASCIIFormat();
	    FileWriter fw = new FileWriter(file);
	    BufferedWriter bw = new BufferedWriter(fw);
	    PrintWriter out = new PrintWriter(bw, true);
	    
	    StringTokenizer st = new StringTokenizer(tableInASCII, "\n", false);
	    while(st.hasMoreTokens()) {
		out.println(st.nextToken()); // printing the string in a file one "line" (token) at a time
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }


    /**
     * Returns the table in a string. Saves also all the inner structures.
     * 
     * @return the table in a string
     */
    public String getTableInASCIIFormat() {
	
	String tableInASCII = null;

	//Finding the name of the representation of the structure. 	
	String repreTemp = visualcontainer.getNameOfRepresentation();
	if(repreTemp == "")
	    repreTemp = visualcontainer.getRepresentationName();
	if(repreTemp.equals("Graph"))
	   repreTemp = "layered graph";
	if(repreTemp.equals("Layered Tree"))
	    repreTemp = "layered tree";

	//Adding information about visualization of the structure.
	tableInASCII = "#representation " + repreTemp + "\n";
	tableInASCII = tableInASCII.concat("#rotated " + visualcontainer.isRotated() + "\n");
	tableInASCII = tableInASCII.concat("#minimized " + visualcontainer.isMinimized() + "\n");
	//tableInASCII = tableInASCII.concat("#matrix table\n");
	tableInASCII = tableInASCII.concat("#matrix array\n");
	    
	int arraySize = items.length;
	VisualComponent vc = null;
	VisualType vt = null;
	int postfix = 1;
	LinkedList structureNames = new LinkedList();
	//string representations of the structures
	HashMap structuresInASCII = new HashMap();

	for(int i=0; i < arraySize; i++) {
	    //System.out.println(i);
	    if (((VisualComponent)items[i]).getKey()== null) {
		//System.out.println("getKey() is null");
		tableInASCII = tableInASCII.concat("' \n"); //empty line if component points to null
	    }
	    else {
		vc = (VisualComponent)items[i];
		vt = (VisualType)vc.getOther(0); // getting item from visualcomponent
		
		if (vt instanceof VisualKey) { // item is a normal key
		    //System.out.println("vt was VisualKey");
		    if(((VisualKey)vt).getKey() == "")
			tableInASCII = tableInASCII.concat("' \n");
		    else {
			//writes the key and marks the end of the key (line)
			//tableInASCII = tableInASCII.concat(((VisualKey)vt).getKey() + "\n"); 
			String temp = processName(((VisualKey)vt).getKey().toString());
			tableInASCII = tableInASCII.concat(temp + "\n");
		    }
		}
		else {
		    //item contained inner structure
		    String structureName = saveName + "_" + postfix; //deciding the name of the inner structure
		    postfix++;
		    tableInASCII = tableInASCII.concat(structureName + "\n");
		    //ASCIISaver saver2 = new ASCIISaver(vt.getLinks(), vt.getItems(), structureName);
		    ASCIISaver saver2 = new ASCIISaver(vt, structureName);
		    structureNames.add(structureName); //adding name of the structure

		    if (vt instanceof VisualArray) {
			//System.out.println("vt was VisualArray");
			String temp = saver2.getTableInASCIIFormat(); //getting table in ascii-string
			structuresInASCII.put(structureName, temp); //saving the string
		    }
		    else {
			//System.out.println("was neither VisualKey nor VisualArray");
			String temp = saver2.getGraphInASCIIFormat(); //getting graph in ascii-string
			structuresInASCII.put(structureName, temp);
		    }
		}
	    }
	}

	tableInASCII = tableInASCII.concat("#EOS\n"); //structure separator

	// adding all the inner structures in the end of the string
	ListIterator listiter = structureNames.listIterator(0);
	while(listiter.hasNext()) {
	    String sname = (String)listiter.next();
	    
	    //name of the structure, space characters indent the inner structure
	    tableInASCII = tableInASCII.concat("   " + sname + "\n");
	    
	    String temp = (String)structuresInASCII.get(sname);//getting the string representation of the structure

	    StringTokenizer st = new StringTokenizer(temp, "\n", false);
	    while (st.hasMoreTokens()) {
		//space characters indent the inner structure
		tableInASCII = tableInASCII.concat("   " + st.nextToken() + "\n");
	    }
	}
	return tableInASCII;
    }
    
}
