/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.util.io;

import java.awt.Frame;
import java.io.*;
import java.net.*;
import java.util.Vector;

import matrix.animation.Animator;
import matrix.structures.FDT.FDT;
import matrix.util.*;

/**
 * a libraty class to save and load matrix animations from and to a file. Instance of
 * this class is created internally to store the animator and the FDT Vector. This 
 * instance is then serialized into a file. 
 *
 * @author Petri Tenhunen
 */
public class MatrixSerialization implements Serializable{

    private Animator animator;
    private FDT[] savedFDT;
  
    /** 
     * Instance of this class is only created to encapsulate animator and fdt vector
     * together. This capsulation is serialized into a file when sace is called.
     *
     * @param a animator to save
     * @param v fdt vector to save
     */
    protected MatrixSerialization(Animator a, Vector v) {
	animator = a;
	this.savedFDT = new FDT[v.size()];
	v.copyInto(savedFDT);
    }

  /** 
     * Instance of this class is only created to encapsulate animator and fdt vector
     * together. This capsulation is serialized into a file when sace is called.
     *
     * @param a animator to save
     * @param ss fdt array to save
     */
    protected MatrixSerialization(Animator a, FDT[] ss) {
        animator = a;
        savedFDT = ss;
    }

    /**
     * If there is a saved matrix animation in given file a new Application window 
     * is opened with saved animation.
     *
     * @param f file to load
     * @see matrix.util.Application
     * @see matrix.util.ApplicationAdapter
     */
    public static boolean load(File f) {
	MatrixSerialization loaded;
	try { 
	    FileInputStream istream = new FileInputStream(f);
	    ObjectInputStream p = new ObjectInputStream(istream);
      
	    loaded = (MatrixSerialization)p.readObject();
	    istream.close();
	} 
	catch (IOException e) { Note.warning(null,e.toString()); return false; }
	catch (ClassNotFoundException e) {  Note.warning(null,e.toString()); return false; }

	Application app = ApplicationAdapter.getApplication().getNewInstance(loaded.savedFDT, 
							   loaded.animator);
	if (app instanceof Frame) {
	  ((Frame)app).show();
	} else {
	  app.getNewWindow().show();
	}
	return true;
    }
    
    
    /**
     * Save animator and fdt vector together into the argument file.
     *
     * @param a animator to save
     * @param v fdt vector to save
     * @param f saving file
     * @deprecated replaced by save(Animator, FDT[], File)
     */
    public static void save(Animator a, Vector v, File f) {
      FDT[] tmp = new FDT[v.size()];
      for(int i = 0; i < tmp.length; i++) {
	tmp[i] = (FDT) v.elementAt(i);
      }
      save(a, tmp, f);
    }

    public static void save(Animator a, FDT[] ss, File f) {
	MatrixSerialization s = new MatrixSerialization(a,ss);

	try {
	    FileOutputStream ostream = new FileOutputStream(f);
	    ObjectOutputStream p = new ObjectOutputStream(ostream);
	    p.writeObject(s);
	    p.flush();
	    ostream.close();
	} catch (IOException e) { 
	    Note.out(e,"Error: " + e.toString()); 
	    e.printStackTrace();
	}    
    }


    public static MatrixSerialization loadFromFile(File f) {
	MatrixSerialization loaded;
	try { 
	    FileInputStream istream = new FileInputStream(f);
	    ObjectInputStream p = new ObjectInputStream(istream);
	    
	    loaded = (MatrixSerialization)p.readObject();
	    istream.close();
	} 
	catch (IOException e) { Note.warning(null,e.toString()); return null; }
	catch (ClassNotFoundException e) {  Note.warning(null,e.toString()); return null; }
	
	return loaded;
    }



    public static MatrixSerialization loadFromNet(URL url) {
	MatrixSerialization loaded;
	try {
	    URLConnection c = url.openConnection();
	    InputStream istream = c.getInputStream();
	    ObjectInputStream p = new ObjectInputStream(istream);
	    
	    loaded = (MatrixSerialization)p.readObject();
	    istream.close();
	} 
	catch (IOException e) { 
	    Note.warning(null,e.toString()); 
	    return null; 
	}
	catch (ClassNotFoundException e) {  
	    Note.warning(null,e.toString()); 
	    return null; 
	}
	return loaded;
    }
    



    public Animator getAnimator() {
	return animator;
    }


    public Vector getFDTs() {
      Vector v = new Vector(savedFDT.length);
      for (int i = 0; i < savedFDT.length; i++) {
	v.addElement(savedFDT[i]);
      }
      return v;
    }


    public FDT[] getStructures() {
      return savedFDT;
    }


    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append("MatrixSerialization:\n");
	buf.append("Animator=" + animator);
	buf.append("\nFDTs\n");
	java.util.Enumeration e = getFDTs().elements();
	while (e.hasMoreElements()) {
	    FDT fdt = (FDT)e.nextElement();
	    buf.append(fdt.toString());
	    buf.append("\n");
	}
	return buf.toString();
    }
    

   static final long serialVersionUID = 983793253973548179L;
 

}

