/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: StructuresFromFile.java,v 1.12 2004/05/26 12:47:22 kstalnac Exp $ */

package matrix.util.io;

import java.awt.Component;
import java.io.*;
import java.util.*;

import matrix.structures.FDT.FDT;
import matrix.uitools.StructurePanel;
import matrix.visual.*;

/**
 * Creates structures from data in a file. This can open several structures
 * that are saved in a same file. Structures can also contain inner structures
 * which are also saved in a same file.<br>
 * File that contains several structures starts with:
 * #matrix structures
 * <p>
 * example:<br>
 * 
 * #matrix structures<br>
 * structure#1<br>
 * #matrix graph adjacency-list<br>
 * key1:key2 key3<br>
 * key2:structure#1_1<br>
 * key3<br>
 * #EOS<br>
 * structure#1_1<br>
 * #matrix array<br>
 * key4<br>
 * key5<br>
 * key6<br>
 * #EOS<br>
 * structure#2<br>
 * #matrix array<br>
 * a<br>
 * b<br>
 * c<br>
 * #EOS<br>
 * etc.<br>
 * <p>
 * Moreover, file can contain information about the visualization of the structures 
 * and this information is used when opening the file. At the moment supported
 * visualization settings can be seen from the example below.
 * <p>
 * example:<br>
 * ...<br>
 * structure#1<br>
 * #representation layered graph<br>
 * #rotated true<br>
 * #minimized false<br>
 * #matrix graph adjacency-list<br>
 * ...<br>
 * <p>
 * Lines which contains visualization settings are not necessary, so they can be 
 * omitted.
 * <p>
 * This type of file can be created by saving all the structures in an 
 * animation window with "Save To Ascii" in a File-menu.
 * <p>
 * Moreover, lines can contain additional comments, which are skipped when loading the file.
 * Comments must start with string "//". Space characters are also removed from the end of 
 * each line. Empty characters before the comment must be space characters.
 * <p>
 * example:<br>
 * <br>
 * #matrix structures     //this is the header of the file<br>
 * structure#1            // name of the first structure<br>
 * #matrix graph adjacency-list     //type of the first structure<br>
 * ...<br>
 */
public class StructuresFromFile {

    public static final String STRUCTURES_HEADER = "#matrix structures";

    boolean hasStructures = false;
    private FDT[] structures = null;    

    //after parsing the file this contains names of the outermost structures
    private LinkedList structureNames = new LinkedList();
    //after parsing the file this contains all the structures in strings
    private HashMap structuresInASCII = new HashMap(); //Structures In ASCII

    /**
     * Constructor
     *
     * @param fileName file, which contains keys of the table
     */
    public StructuresFromFile(String fileName) {
	try {
	    parseFile(fileName);
	    
	    ListIterator listiter = structureNames.listIterator();
	    structures = new FDT[structureNames.size()];
	    int i = 0;
	    String name = null;
	    FDT temp = null;
	    
	    while(listiter.hasNext()) {
		name = (String)listiter.next();
		temp = getStructureFromASCII(name);
		structures[i] = temp;
		i++;
	    }
	    
	} catch(Exception e) {
	    e.printStackTrace();
	    hasStructures = false;
	}
    }


    /**
     * Returns true, if the file contains structures.
     *
     * @param fileName file which needs to be checked
     * @return true, if the file contains structures
     */
    public static boolean hasStructures(String fileName) {
	try {
	    BufferedReader reader = new BufferedReader(new FileReader(fileName));
	    String line = reader.readLine();
	    line = handleLine(line);
	    while (line.equals("")) {
		line = reader.readLine();
		line = handleLine(line);
	    }
	    //file contains structures if the first header in
	    //the file is STRUCTURES_HEADER
	    return line.trim().equalsIgnoreCase(STRUCTURES_HEADER);
	} catch (Exception e) {
	    return false;
	}
    }




    /**
     * Parses a file and inserts the string representation of the each structure
     * into hashmap.
     *
     * @param fileName file, which contains the structures and needs to be parsed
     */
    private void parseFile(String fileName) {
	String structureName = null;
	try {
	    BufferedReader reader = new BufferedReader(new FileReader(fileName));
	    String sia = ""; //structure in ascii
	    String line = reader.readLine(); //first line
	    line = handleLine(line);
	    while (line.equals("")) {
		line = reader.readLine();
		line = handleLine(line);
	    }
	    hasStructures = line.trim().equalsIgnoreCase(STRUCTURES_HEADER);
	    if(hasStructures()) {
		structureName = reader.readLine();
		structureName = handleLine(structureName);
		while (structureName.equals("")) {
		    structureName = reader.readLine();
		    structureName = handleLine(structureName);
		}   
		//saving the names of the outermost structures
		if(isMainStructureName(structureName))
		    structureNames.add(structureName);

		while((line = reader.readLine()) != null) {
		    line = handleLine(line);
		    while (line.equals("")) {
			line = reader.readLine();
			if (line == null)
			    break;
			line = handleLine(line);
		    }
		    if (line == null)
			break;
		    if(line.trim().equals("#EOS")) {
			structuresInASCII.put(structureName, sia);
			line = reader.readLine();
			if(line != null) {
			    line = handleLine(line);
			    while (line.equals("")) {
				line = reader.readLine();
				if(line == null)
				    break;
				line = handleLine(line);
			    }
			    if (line == null)
				break;
			    structureName = line;
			    if(isMainStructureName(structureName))
				structureNames.add(structureName);
			    sia = "";
			}
		    }
		    //else if(line.startsWith("#representation") || 
		    //	    line.startsWith("#rotated") ||
		    //    line.startsWith("#minimized")) {
			//do nothing, only skip these lines
			//System.out.println("Skipped: " + line);
		    //}
		    else {
			if(line.equals(""))
			    sia = sia.concat(" \n");
			else
			    sia = sia.concat(line + "\n");
		    }
		}
		structuresInASCII.put(structureName, sia);
	    }
	    else {
		structureName = fileName;
		structureNames.add(structureName);
		sia = sia.concat(line + "\n");
		while((line = reader.readLine()) != null) {
		    line = handleLine(line);
		    while (line.equals("")) {
			line = reader.readLine();
			if (line == null)
			    break;
			line = handleLine(line);
		    }
		    if (line == null)
			break;
		    if(line.trim().equals("#EOS")) {
			structuresInASCII.put(structureName, sia);
			line = reader.readLine();
			if(line != null) {
			    line = handleLine(line);
			    while (line.equals("")) {
				line = reader.readLine();
				if (line == null)
				    break;
				line = handleLine(line);
			    }
			    if (line == null)
				break;
			    structureName = line;
			    sia = "";
			}
		    }
		    else {
			if(line.equals(""))
			    sia = sia.concat(" \n");
			else
			    sia = sia.concat(line + "\n");
		    }
		}
		structuresInASCII.put(structureName, sia);
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	    hasStructures = false;
	}
    }


    private static String handleLine(String str) {
	String temp = str;
	temp = removeSpacesFromBeginning(temp);
	temp = removeCommentsFromLine(temp);
	return temp;
    }

   /**
     * Removes space characters from the beginning of the string.
     *
     * @param str string which needs to be handled
     * @return string where the possible space characters are removed
     */
    private static String removeSpacesFromBeginning(String str) {
	String temp = str;
	while(temp.indexOf(" ") == 0) {
	    if (temp.length() == 1)
		return "";
	    temp = temp.substring(1);
	}
	return temp;
    }


    /**
     * Removes comments from the string. Comments starts with "//".
     *
     * @param str string which needs to be handled
     * @return string from where possible comments are removed
     */
    private static String removeCommentsFromLine(String str) {
	String temp = str;
	int indexOfComment = temp.indexOf("//");
	if (indexOfComment == -1)
	    ;
	else
	    temp = temp.substring(0, indexOfComment);
	temp = removeSpacesFromEnd(temp);
	return temp;
    }


    /**
     * Removes space characters from the end of a string.
     *
     * @param str string which needs to be handled
     * @return string from where possible space charachters are removed
     */
    private static String removeSpacesFromEnd (String str) {
	String temp = str;
	int indexOfLastSpace = temp.lastIndexOf(" ");
	while (indexOfLastSpace == (temp.length() - 1)) {
	    if (temp.length() == 0)
		return temp;
	    temp = temp.substring(0, indexOfLastSpace);
	    indexOfLastSpace = temp.lastIndexOf(" ");
	}
	if (temp.lastIndexOf("'") == (temp.length() - 1))
	    return temp.concat(" ");
	else
	    return temp;
    }
    

    /**
     * Checking if the structure is one of the outermost structures.
     * Outermost structures are marked with "#nro" in end of the their 
     * name. Example: structure#1
     *
     * @param name name of the structure which needs to be checked
     * @return true, if structure is one of the outermost structures
     */
    private boolean isMainStructureName(String name) {
	int index = name.lastIndexOf("#");
	if(index > 0) {
	    String temp = name.substring(index + 1);
	    try {
		if(Integer.parseInt(temp) >= 0) {
		    return true;
		}

	    } catch (Exception e) {

	    }
	}
	return false;
    }


    /**
     * Returns FDT which corresponds the given name of the structure.
     *
     * @param sname name of the structure
     * @return FDT which corresponds the name of the structure
     */
    private FDT getStructureFromASCII(String sname) {
	String temp = (String)structuresInASCII.get(sname);
	StringTokenizer st = new StringTokenizer(temp, "\n", false);
	String visualSettings = ""; //visualization settings
	String line = null;
	line = st.nextToken();
	while(!(line.toLowerCase().startsWith("#matrix"))) {
	    visualSettings = visualSettings.concat(line + "\n");
	    line = st.nextToken();
	}

	//String firstLine = st.nextToken();
	//if (firstLine.equals(TableFromFile.TABLE_HEADER)) {
	if(line.trim().equalsIgnoreCase(TableFromFile.TABLE_HEADER) ||
	   line.trim().equalsIgnoreCase(TableFromFile.TABLE_HEADER2)) {
	    TableFromFile tff = new TableFromFile(structuresInASCII, this);
	    FDT fdttemp = tff.getTableFromASCII(st);
	    //saves the visualization settings for the current fdt
	    setVisualSettingsForFDT(fdttemp, visualSettings);
	    return fdttemp;
	}
	else if(line.trim().equalsIgnoreCase(GraphFromFile.GRAPH_HEADER) ||
		line.trim().equalsIgnoreCase(GraphFromFile.GRAPH_HEADER + " " + GraphFromFile.EDGELIST)) {
	    //System.out.println("-- graph edge-list");
	    GraphFromFile gff = new GraphFromFile(structuresInASCII, this);
	    FDT fdttemp = gff.getGraphFromASCII2(st);
	    //saves the visualization settings for the current fdt
	    setVisualSettingsForFDT(fdttemp, visualSettings);
	    return fdttemp;
	}
	else {
	    //System.out.println("-- graph adjacency-list");
	    GraphFromFile gff = new GraphFromFile(structuresInASCII, this);
	    FDT fdttemp = gff.getGraphFromASCII(st);
	    //saves the visualization settings for the current fdt
	    setVisualSettingsForFDT(fdttemp, visualSettings);
	    return fdttemp;
	}
    }


    HashMap visualSettingsForFDTs = new HashMap();

    /**
     * Sets the visualization settings for the fdt
     *
     * @param fdt fdt
     * @param visualSettings visualization settings
     */
    public void setVisualSettingsForFDT(FDT fdt, String visualSettings) {
	visualSettingsForFDTs.put(fdt, visualSettings);
    }


    /**
     * Returns the visualization settings of the specific fdt.
     *
     * @param fdt fdt
     * @return visualization of the fdt
     */
    public String getVisualSettingsForFDT(FDT fdt) {
	return (String)visualSettingsForFDTs.get(fdt);
    }



    /**
     * Returns true, if the given file contained structures.
     *
     * @return true, if the given file contained structures.
     */
    public boolean hasStructures() {
	return hasStructures;
    }


    /**
     * Returns the FDTs which were saved in the file.
     *
     * @return FDTs which were saved in the file
     */
    public FDT[] getStructures() {
	return structures;
    }


    /**
     * Configures the VisualTypes to correspond the visualization settings.
     *
     * @param vts VisualTypes which should be configured
     * @param sp current structurepanel where the structures are
     */
    public void configureVisualTypes(VisualType[] vts, StructurePanel sp) {
	VisualType[] visualTypes = vts;
	for(int i = 0; i < visualTypes.length; i++) {
	    VisualType vttemp = visualTypes[i];
	    configureVisualTypeRepresentation(vttemp, sp);
	}

	//visualtypes have changed so next we get new visualtypes
	visualTypes = sp.getVisualComponents();
	for(int i = 0; i < visualTypes.length; i++) {
	    VisualType vttemp = visualTypes[i];
	    configureVisualType(vttemp, sp);
	}
    }


    /**
     * Configures the representation of the visualtypes.
     *
     * @param visualType visualtype which needs to be configured
     * @param sp current structurepanel which contains the visualtypes
     */
    private void configureVisualTypeRepresentation(VisualType visualType, StructurePanel sp) {
	VisualType vt = visualType;
	//getting visualization settings
	String visualSettings = getVisualSettingsForFDT(vt.getStructure());
	String representation = null;
	if(visualSettings != null) {
	    StringTokenizer st = new StringTokenizer(visualSettings, "\n", false);
	    String setting = "";
	    int index;
	    while(st.hasMoreTokens()) {
		setting = st.nextToken();
		//finding the name of the representation
		if(setting.startsWith("#representation")) {
		    index = setting.indexOf(" ");
		    representation = setting.substring(index +1).trim();
		}
		else {
		    //do nothing
		}
	    }
	}
	
	if(representation != null) {
	    //changing representation
	    sp.changeRepresentation((VisualContainer)vt, representation);
	}
    }

    /**
     * Configures the visualtype.
     *
     * @param visualType visualtype which needs to be configured
     * @param sp current structurepanel which contains the visualtypes
     */
    private void configureVisualType(VisualType visualType, StructurePanel sp) {
	VisualType vt = visualType;
	//getting visualization settings
	String visualSettings = getVisualSettingsForFDT(vt.getStructure());
	//String representation = null;
	boolean rotated = false;
	boolean minimized = false;
	
	if(visualSettings != null) {
	    StringTokenizer st = new StringTokenizer(visualSettings, "\n", false);
	    String setting = "";
	    int index;
	    while(st.hasMoreTokens()) {
		setting = st.nextToken();
		//if(setting.startsWith("#representation")) {
		//  index = setting.indexOf(" ");
		//  representation = setting.substring(index + 1).trim().toLowerCase();
		//}
		//else 
		if(setting.startsWith("#rotated")) {
		    index = setting.indexOf(" ");
		    setting = setting.substring(index + 1).trim();
		    rotated = setting.equals("true");
		}
		else if(setting.startsWith("#minimized")) {
		    index = setting.indexOf(" ");
		    setting = setting.substring(index + 1).trim();
		    minimized = setting.equals("true");
		}
		else {
		    //do nothing
		}
	    }
	}
	

	if(visualSettings != null) {
	    //setting the visualization settings
	    ((VisualContainer)vt).setRotated(rotated);
	    vt.setMinimized(minimized);
	    vt.setInvalid();
	}

	//next clauses handles the inner structures and their visualizations
	Component[] items = vt.getItems();
	VisualComponent vc = null;
	VisualType vttempOld = null;
	FDT fdttemp = null;
	String str = null;
	for(int i = 0; i < items.length; i++) {
	    vc = (VisualComponent)items[i];
	    vttempOld = (VisualType)vc.getOther(0);
	    fdttemp = vttempOld.getStructure();
	    str = getVisualSettingsForFDT(fdttemp);
	    if(str != null) {
		//configuring the visualtype of the inner structure
		configureVisualType(vttempOld, sp);
	    }
	}
    }

   
}
