/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualGraphComponent.java,v 1.30 2004/05/26 08:06:08 vkaravir Exp $ */

package matrix.visual;

import java.awt.Container;
import java.io.PrintStream;
import java.util.Vector;

import matrix.structures.FDT.*;
import matrix.util.Note;

/** This abstraction is used to represent the nodes of an abstract graph. 
 * A subclass of the VisualGraph could have its own concrete subclass of
 * the VisualGraphComponents in order to have a special look and feel of nodes.
 * @see VisualGraph
 *
 * @author Ari Korhonen
 * @version $Revision: 1.30 $
 */
public class VisualGraphComponent extends VisualNode {
    private Vector succ = new Vector();
    private Vector pred = new Vector();
    private VisualReference[] successor, predecessor;
    //    private MenuItem insertEdge, setSource;

    // debug
    public void dump(PrintStream out) {
        super.dump(out);
        out.println("succ="+getSuccessors().length+
                 ";#pred="+getPredecessors().length);
        for (int i=0; i<successor.length; i++)
            out.println("  ->"+successor[i]);
        for (int i=0; i<predecessor.length; i++)
            out.println("  <-"+predecessor[i]);
    }

    /** Construct a new representation for the specified Vertex. */
    public VisualGraphComponent(Vertex v) {
        super(v);   
        successor = new VisualReference[0];
        predecessor = new VisualReference[0];
	//addPopUpMenuItem(insertEdge = new MenuItem("insertEdge"));
	//addPopUpMenuItem(setSource = new MenuItem("setSource"));
    }
    
    // --- end of constructors ---

    /** This method should be invoked after all the successors and predecessors
     ** (references for those) are seen and they should be committed. 
     */
    public void endReferenceUpdate() {
        // update successors
        successor = new VisualReference[succ.size()];
        for (int i=0; i<succ.size(); i++)
            successor[i] = (VisualReference)succ.elementAt(i);
        succ.setSize(0);
        // update predecessors
        predecessor = new VisualReference[pred.size()];
        for (int i=0; i<pred.size(); i++)
            predecessor[i] = (VisualReference)pred.elementAt(i);
        pred.setSize(0);
    }

    /** This method is invoked in order to notify that the specified
     ** VisualReference is pointing from this node to some of the
     ** successors of this node
     * @see #endReferenceUpdate()
     */
    public void addSuccessor(VisualReference vr) {
        succ.addElement(vr);
    }

    /**
     * Removes the argument reference from the list of successor references.
     */
    public void removeSuccessor(VisualReference vr) {
	succ.removeElement(vr);
    }
    
    /** Return the successors that are committed to this node.
     * @see #endReferenceUpdate()
     */        
    public VisualReference[] getSuccessors() {
        return successor;
    }
    
    public VisualGraphComponent[] getSuccessorComponents() {
        VisualGraphComponent s[] =
            new VisualGraphComponent[successor.length];

        for (int i = 0; i < successor.length; i++)
            // XXX does cast always work?
            s[i] = (VisualGraphComponent) successor[i].getTarget();
        
        return s;
    }

    /** This method is invoked in order to notify that the specified
     ** VisualReference is pointing from some of the predecessors
     ** into this node.
     * @see #endReferenceUpdate()
     */
    public void addPredecessor(VisualReference vr) {
        pred.addElement(vr);
    }
    
    /**
     * Removes the parameter reference from the list of redecessor references.
     */
    public void removePredecessor(VisualReference vr) {
	pred.removeElement(vr);
    }

    /** Return the predecessors that are committed to this node.
     * @see #endReferenceUpdate()
     */        
    public VisualReference[] getPredecessors() {
        return predecessor;
    }

    public VisualGraphComponent[] getPredecessorComponents() {
        VisualGraphComponent s[] =
            new VisualGraphComponent[predecessor.length];
	    Object tmp1 = null;
	    Object tmp2 = null;
        for (int i = 0; i < predecessor.length; i++)
            // XXX does cast always work?
            s[i] = (VisualGraphComponent) predecessor[i].getSource();
        
        return s;
    }

    public int getReferenceNumber(VisualType to) {
	if(to instanceof VisualGraphComponent) {
	    Vertex v1 = (Vertex)getStructure();
	    Vertex v2 = (Vertex)to.getStructure();
	    Vertex[] sc = v1.getSuccessors();
	    for(int i = 0; i < sc.length; i++) {
		if(v2 == sc[i])
		    return i;
	    }
	    return -1;
	} else {
	    Note.err(this, "getReferenceNumber: Illegal argument " + to);
	    return -1;
	}
    }	    

    public void erase() {
	if(getParent() instanceof VisualType 
	   && ((VisualType)getParent()).getStructure() instanceof SimulationGraph) {
	    VisualGraphComponent[] pred = getPredecessorComponents();
	    SimulationGraph graph = (SimulationGraph)((VisualType)getParent()).getStructure();
	    SimulationVertex v;
	    // Should this be removed and just expect the underlying data structure to
	    // handle deletion of edges?
	    for(int i = 0; i < pred.length; i++) {
		try {
		    v = (SimulationVertex) pred[i].getStructure();
		}catch(ClassCastException e) {
		    Note.out(this, pred[i].getStructure() + " not a SimulationVertex, cannot remove link");
		    continue;
		}
		v.removeSuccessor((Vertex)getStructure());
	    }
	    graph.removeVertex((Vertex)getStructure());
	    ((VisualType)getParent()).setInvalid();
	} else
	    getApplication().showMessage("Sorry", "Cannot delete from a non-dynamic graph.");
    }

    public void moveReference(VisualReference r, 
			      VisualGraphComponent from, 
			      VisualGraphComponent to) {
	//Note.out(this, "new move reference");
	FDT origin = getStructure();
	FDT oldTarget = from.getStructure(); // FDT the edge goes to
	FDT newTarget = to.getStructure(); // FDT the edge is moved to
	if (origin instanceof SimulationVertex 
	    && oldTarget instanceof SimulationVertex 
	    && newTarget instanceof SimulationVertex) {
	    getAnimator().startOperation();		
	    ((SimulationVertex)origin).removeSuccessor((Vertex)oldTarget);
	    ((SimulationVertex)origin).addSuccessor((Vertex)newTarget);
	    getAnimator().endOperation();		
	    setInvalid();
	} else
	    getApplication().showMessage("Sorry", 
					 "Cannot move reference " + 
					 oldTarget + " to " + newTarget + 
					 " (simulation not allowed for this vertex");
    }

    public void moveReference(VisualReference r, VisualType from, VisualType to) {
	if ((from instanceof VisualGraphComponent) && (to instanceof VisualGraphComponent))
	    moveReference(r, (VisualGraphComponent)from, (VisualGraphComponent)to);
	else
	    getApplication().showMessage("Sorry", "VisualGraphComponent: Cannot move reference from " + from + " into " + to);	
    }

    // Moves a reference. Currently unable to color the moved reference.
    // See VisualTreeComponent ?
    /*
      void moveReference(VisualReference r, VisualType from, VisualType to) {
	Note.out(this, "old move reference "+from+"; "+to);
	VisualGraph vgraph;
	VisualComponent edgeStart = null;;
	VisualGraphComponent oldGoal = null;
	VisualGraphComponent newGoal = null;

	// seach the graph the old target belongs to
	VisualType temp = from;
	while(temp instanceof VisualType && !(temp instanceof VisualGraph)) 
	    temp = (VisualType) temp.getParent();
	if(temp instanceof VisualGraph)
	    vgraph = (VisualGraph)temp;
	else {
	    getApplication().showMessage("Sorry", "Cannot move reference " + from + " to " + to);
	    return;
	}

	while((from instanceof VisualType) &&
	      !(from instanceof VisualGraphComponent))
	    from = (VisualType)from.getParent();
	if(from instanceof VisualGraphComponent)
	    oldGoal = (VisualGraphComponent) from;
	while((to instanceof VisualType) &&
	      !(to instanceof VisualGraphComponent))
	    to = (VisualType)to.getParent();
	if(to instanceof VisualGraphComponent)
	    newGoal = (VisualGraphComponent)to;
	edgeStart = r.getSource();
	if(edgeStart instanceof VisualGraphComponent 
	   && oldGoal instanceof VisualGraphComponent 
	   && newGoal instanceof VisualGraphComponent) {
	    FDT tmp1 = edgeStart.getStructure(); // FDT the edge originates from
	    FDT tmp2 = oldGoal.getStructure(); // FDT the edge goes to
	    FDT tmp3 = newGoal.getStructure(); // FDT the edge is moved to
	    //FDT graph = vgraph.getStructure(); // the graph
	    if(tmp1 instanceof SimulationVertex 
	       && tmp2 instanceof SimulationVertex 
	       && tmp3 instanceof SimulationVertex) {
		getAnimator().startOperation();		
		((SimulationVertex)tmp1).removeSuccessor((Vertex)tmp2);
		((SimulationVertex)tmp1).addSuccessor((Vertex)tmp3);
		getAnimator().endOperation();		
		vgraph.setInvalid();
		return;
	    }
	}
	getApplication().showMessage("Sorry", "Cannot move reference " + from + " to " + to);
	return;
    }
    */
    public void insertEdge() {
	if(getStructure() instanceof SimulationVertex) {
	    Container tmp = getParent();
	    if(tmp instanceof VisualGraph)
		((VisualGraph)tmp).addEdge(this);
	    else
		getApplication().showMessage("Sorry", "Could not insert edge");
	} else
	    getApplication().showMessage("Sorry", "Not a SimulationVertex, cannot add successor.");
    }

    public void setSource() {
	Container tmp = getParent();
	if(tmp instanceof VisualGraph) {
	    FDT fdt = ((VisualGraph)tmp).getStructure();
	    if(fdt instanceof Graph) 
		fdt.setElement(getStructure());
	}
    }

    public void selectEntity(boolean isShiftDown) {
	Container c = getParent();
	if(c instanceof VisualGraph && ((VisualGraph)c).getEdgeAdd()) {
	    VisualGraph parent = (VisualGraph) c;
	    VisualGraphComponent vgc = parent.getEdgeFrom();
	    FDT source = vgc.getStructure();
	    FDT goal = getStructure();
	    if(source instanceof SimulationVertex && goal instanceof Vertex) {
		getAnimator().startOperation();
		((SimulationVertex)source).addSuccessor((Vertex)goal);
		getAnimator().endOperation();
	    }
	    parent.setEdgeAdd();
	} else if (isShiftDown)
	    insertEdge();
    }
}






