/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualNode.java,v 1.44 2004/05/26 08:06:11 vkaravir Exp $ */

package matrix.visual;

import java.awt.*;
import java.io.PrintStream;

import matrix.decoration.*;
import matrix.structures.FDT.FDT;
import matrix.util.Note;

/** This abstract class is the superclass for all nodes. 
 * A node is a single dynamic object in a more complex object structure
 * and it knows its predecessors and successors.
 */
public class VisualNode extends VisualComponent {
    private FDT fdt;
    private boolean circleMode;
    private Insets in, cin;
    private int drawMode = -1;

    private static final int circleMaxDiameter = 35;
    private static final int rrectSpace = 3;
    private static final int rrectDiam = 16;

    // constructor
    public VisualNode(FDT fdt) {
        super(fdt); 
        this.fdt = fdt;
        in = new Insets(0,0,0,0);
        cin = new Insets(0,0,0,0);
        labeled = false;
        if (fdt instanceof LabelDecorator) {
            LabelDecorator ld = (LabelDecorator)fdt;
            setName(ld.getLabel());
            if (ld.isLabelEnabled())
                labeled = true;
        } else if (fdt instanceof Styled) {
	    StyleSheet ss = ((Styled)fdt).getStyleSheet();
	    setName(ss.getLabel());
	    if (ss.isLabelEnabled())
		labeled = true;
	} else setName("Node");
    }
    
    /** @return the key of the object this visual object is representing. 
     * @see #getStructure()
     */
    public FDT getKey() {
        if (getStructure() == null) return null;
        return turnFDT(getStructure().getElement());
    }

    public void dump(PrintStream out) {
	super.dump(out);
	Component[] c = getComponents();
	out.println("VisualNode components:");
	for(int n = 0; n < c.length; n++)
	    out.println(" "+c[n].toString());
	out.println("End of components.");
    }

    /** Lay out a single node within this component border. 
     * First, super doLayout() is invoked which
     * recalculates the width and height of this
     * component depending on the size of the nested structure.
     * Finally, the correct positioning and size of the nested structure
     * is calculated and the bounds of the nested strucures are set in
     * case the default behaviour is not valid (e.g. in case of
     * minimized polygon).
     * @see VisualComponent#doLayout()
     */
    public void doLayout() {
        super.doLayout();
        if (isMinimized() || isDuplicate()) { // draw minimized polygon
            VisualType vt;
            if ((vt = getVisualType()) != null) {
                Insets in = getInsets();
                Rectangle d = vt.getBounds();
                int x = d.x;
                int y = d.y + (height - d.height + 1)/2 - in.top - in.bottom;
                vt.setLocation(x, y);
            } 
        }
    }
    
    /** By default this method returns true. 
     * Subclasses should override this method to return 
     * false if ordinary rectangle is preferred instead of rounded rectangle.
     * @see VisualCodeLine#isRoundRectangle()
     */
    public boolean isRoundRectangle() {
        return true;
    }
    
    /** Return preferred size for rounded components.
     * The dimension of this component is set as large as required
     * for representing the nested visualtype within an ellipse.
     */
    public Dimension getPreferredSize() {
        if (getVisualType() != null) {
            //Insets in = getInsets();
            Dimension d = getVisualType().getPreferredSize();
	    
            if (isMinimized() || isDuplicate()) { // draw minimized polygon
                drawMode = 0;
                // XXX calculate these correctly...
                width = 2*d.width;
                height = 2*d.height;
                in.left = in.right = in.top = in.bottom = 1;
	    } else {
                int diam = (int)Math.sqrt(d.width*d.width +
                                          d.height*d.height) + 2;
                if ((isRoundRectangle()) && (diam <= circleMaxDiameter)) { // draw a circle
                    drawMode = 1;
                    width = height = (int) diam;
                    in.left = in.right = (int)(diam - d.width)/2;
                    in.top = in.bottom = (int)(diam - d.height)/2;
                } else { // draw a (rounded) rectangle
                    drawMode = isRoundRectangle() ? 2 : 3;
                    width = d.width + 2 * rrectSpace;
                    height = d.height + 2 * rrectSpace;
                    in.left = in.right = in.top = in.bottom = rrectSpace;
                }
	    }
        }
	int addWidth = 0, addHeight = 0;
        if (labeled && getFont() != null)
	    addWidth = getFontMetrics(getFont()).stringWidth(getName());

	// for the automatically added node numbers
	VisualType vt = this;
	if (getParent() instanceof VisualType)
	    vt = (VisualType)getParent();
	// Check from Container whether or not add the numbers
	if ((vt instanceof VisualContainer) && ((VisualContainer)vt).isAutoLabeled() && getFont() != null) {
	    FontMetrics fm = getFontMetrics(getFont());
	    addHeight = Math.max(fm.getAscent() - height, 0);
	    addWidth = Math.max(addWidth, fm.stringWidth(getIndex((VisualContainer)vt)+""));
	}
	// node numbering code ended

	if (addWidth != 0 || addHeight != 0) {
	    cin.right = addWidth;
	    in.right += addWidth;
	    width += addWidth + addHeight;
	    height += addHeight;
	} else {
	    cin = new Insets(0, 0, 0, 0);
	    in = new Insets(0, 0, 0, 0);
	}
        return new Dimension(width, height);
    }

    public final VisualType getVisualType() { 
        return getComponentCount() > 0 ? (VisualType)getComponent(0) : null; 
    }

    public Insets getInsets() { 
        return in;
    }
    
    public Insets getComponentInsets() {
        return cin;
    }
    
    public boolean empty() {
        FDT fdt = getStructure();
        return ((fdt == null) || (fdt.getElement() == null));
    }

    // simulation

    public void assign() {
        getApplication().showMessage("Sorry", 
                               "Cannot assign " + read + " into " + this);
    }

    public void write() {
        if (read != null) {
            if ((read.getStructure() != null) && 
		(read.getStructure().equals(this.getStructure()))) { 
                getApplication().showMessage("Error", "Component cannot be it's own key");
                return;
            }
            FDT fdt = getStructure();
            if (fdt != null)
                fdt.setElement(read.getStructure()); 
            else
                getApplication().showMessage("Sorry", 
                          "VisualNode: Cannot write into an empty node.");
        }
    }

    public void remove() {
        getStructure().setElement(null);
    }
    
    public Polygon getMinimizedPolygon(int x, int y, int width, int height) {
        int xpoints[] = new int[4];
        int ypoints[] = new int[4];
        
        xpoints[0] = width/2-1; ypoints[0] = 1;
        xpoints[1] = width/2+1; ypoints[1] = 1;
        xpoints[2] = width; ypoints[2] = height;
        xpoints[3] = 1; ypoints[3] = height;
        
        return new Polygon(xpoints, ypoints, 4);
    }

    public void setMinimized(boolean val) {
        super.setMinimized(val);
        VisualType t = getVisualType();
        if (t != null) {
            t.setMinimized(val);
        }
    }
    
    private int getLabelHeight() {
        FontMetrics fm = getFontMetrics(getFont());
        return fm.getHeight() - fm.getMaxDescent();
    }

    private int getIndex(VisualContainer par) {
	for (int i=0; i<par.getItemCount();i++) 
	    if (par.getItem(i).equals(this))
		return i;
	return -1;
    }

    private void paintLabel(Graphics g, int x, int y, int w, int h) {
      //Note.out(this, "Labeled: " + labeled + " for " + this);
        if (labeled)
            // XXX calculate coordinates correctly!
            g.drawString(getName(), x + w + 1, y + getLabelHeight());
	VisualType vt = this;
	if (getParent() instanceof VisualType) // Should always be, but just
	    vt = (VisualType)getParent();      // checking to be sure
	if ((vt instanceof VisualContainer) && ((VisualContainer)vt).isAutoLabeled()) {
	    g.setColor(Color.black);
	    int num = getIndex((VisualContainer)vt);
	    g.drawString(num+"", x+w+1, y+h);
	}
    }

    public void paint(Graphics g) {
        // super.paint(g) is invoked at the end of this method 
	
        if (isHiding()) {
            return;
	}
        getPreferredSize(); // recalculate everything just in case
        Rectangle b = getComponentBoundsForPaint();
        b.x = b.y = 0;
	b.width--;
	b.height--;
        
        /*
          if (hasFocus())
          setColor(g);
          else setBackgroundColor(g);
        */
        setBackgroundColor(g);
        
        switch (drawMode) {
        case 0: // minimized polygon
            g.fillPolygon(getMinimizedPolygon(b.x, b.y, b.width, b.height));
            setBorderColor(g);
            g.drawPolygon(getMinimizedPolygon(b.x, b.y, b.width, b.height));
            break;
        case 1: // circle
            g.fillOval(b.x, b.y, b.width, b.height);
	    setBorderColor(g);
            g.drawOval(b.x, b.y, b.width, b.height);            
            break;
        case 2: // rounded rectangle
            g.fillRoundRect(b.x, b.y, b.width, b.height,
                            rrectDiam, rrectDiam);
            setBorderColor(g);
            g.drawRoundRect(b.x, b.y, b.width, b.height,
                            rrectDiam, rrectDiam);
            break;
        case 3: // rectangle
            g.fillRect(b.x, b.y, b.width, b.height);
            setBorderColor(g);
            g.drawRect(b.x, b.y, b.width, b.height);
            break;
        default:
            Note.err(this, "invalid drawMode");
            break;
        }
        setPenColor(g);
        paintLabel(g, b.x, b.y, b.width, b.height);
        //resetColor(g);
        super.paint(g);
    }

    public Point getBorderAtAngle(double angle) {
        Point c = getCenter();
        Rectangle b = getComponentBounds();

        switch (drawMode) {
        case 0: // minimized polygon
            return super.getBorderAtAngle(angle); // XXX
        case 1: // circle
            int ellA = b.width / 2, ellB = b.height / 2;
            return new Point((int) (c.x + Math.cos(angle) * ellA),
                             (int) (c.y - Math.sin(angle) * ellB));
        case 2: // rounded rectangle
            double x, y;
            // Calculate the point (x, y) for a rectangle
            // Copied from VisualType.getBorderAtAngle
            double urCornerA = Math.atan2(b.height, b.width);
            double ulCornerA = Math.PI - urCornerA;
            double lrCornerA = 2*Math.PI - urCornerA;
            double llCornerA = urCornerA + Math.PI;
            if (angle < urCornerA || angle > lrCornerA) { // on right side
                x = b.x + b.width;
                y = c.y - (b.width/2.0) * Math.tan(angle);
            } else if (angle > ulCornerA && angle < llCornerA) { // left
                x = b.x;
                y = c.y + (b.width/2.0) * Math.tan(angle - Math.PI);
            } else if (angle <= ulCornerA) { // top
                x = c.x + (b.height/2.0) / Math.tan(angle);
                y = b.y;
            } else { // on bottom side
                x = c.x - (b.height/2.0) / Math.tan(angle - Math.PI);
                y = b.y + b.height;
            }
            // Correct (x, y) for rounded corners
            double cx = 0, cy = 0;
            boolean inCornerX = false, inCornerY = false;
            if (x > b.x + b.width - rrectDiam/2) {
                cx = b.x + b.width - rrectDiam/2;
                inCornerX = true;
            } else if (x < b.x + rrectDiam/2) {
                cx = b.x + rrectDiam/2;
                inCornerX = true;
            }
            if (y > b.y + b.height - rrectDiam/2) {
                cy = b.y + b.height - rrectDiam/2;
                inCornerY = true;
            } else if (y < b.y + rrectDiam/2) {
                cy = b.y + rrectDiam/2;
                inCornerY = true;
            }                
            if (inCornerX && inCornerY) {
                double ang = 2*Math.PI - Math.atan2(y - cy, x - cx);
                x = cx + Math.cos(ang) * rrectDiam/2;
                y = cy - Math.sin(ang) * rrectDiam/2;
            }
            return new Point((int)x, (int)y);
        default:
            /* XXX this gets called before getPreferredSize() etc.
            Note.assert(false, "invalid drawMode");
            return null;
            */
            return super.getBorderAtAngle(angle);
        }
    }

    public void print(Graphics g) {
        // bugi bugi in Java???
        if (isHiding()) return;
        super.print(g);
    }

}
