/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualReference.java,v 1.61 2004/06/10 12:15:12 kstalnac Exp $ */

package matrix.visual;

import java.awt.*;
import java.io.PrintStream;
import java.util.*;

import matrix.decoration.*;
import matrix.structures.FDT.*;
import matrix.structures.memory.Key;
import matrix.util.Note;
import matrix.animation.Animator;

/** This abstraction is used to represent any kind of references.
 * A subclass of the VisualContainer could have its own concrete subclass of
 * the VisualReference in order to have a special look and feel of links.
 * @see VisualContainer
 *
 * @author Ari Korhonen, Riku Saikkonen
 * @version $Revision: 1.61 $
 */
public class VisualReference extends VisualType {
    protected VisualComponent from, to, movingTarget = null;
    protected boolean weighted = false;
    protected Point fromPoint, toPoint;

    private int pick_x, pick_y;
    private VisualType weight;
    private Vector/*<VisualComponent>*/ midPoints;
    private boolean midPointReversal;

    private static final double arrowHalfAngle = Math.PI / 6;
    private static final int arrowLength = 8;
    private static final double epsilon = 4.0;

    /**
     * Returns the centre point of an AWT component converted to the
     * coordinate system of the container of this component.
     * @param c Component to get centre of
     * @return Centre point of component
     */
    private Point getCorrectedCentre(VisualComponent c) {
	if (c!=movingTarget) {
	    Component cp=c.getParent();
	    Component tp=getParent();	
	
	    if ((cp!=tp)&&(cp!=null)&&(tp!=null)) {
		Point p=javax.swing.SwingUtilities.convertPoint(cp,
								c.getCenter(),
								tp);
		return p;
	    }
	}
	return c.getCenter();
    }

    /**
     * Returns the centre point of an AWT component converted to the
     * coordinate system of the container of this component.
     * @param c Component to get centre of
     * @return Centre point of component
     */
    private Point getCorrectedBorderAtAngle(VisualComponent c, double angle) {
	Component cp=c.getParent();
	Component tp=getParent();	

	if (c!=movingTarget) {
	    if ((cp!=tp)&&(cp!=null)&&(tp!=null)) {
		Point p=javax.swing.
		    SwingUtilities.convertPoint(cp,
						c.getBorderAtAngle(angle),
						tp);
		return p;
	    }	    
	}
	return c.getBorderAtAngle(angle);
    }

    // --- debug purposes only ---
    public void dump(PrintStream out) {
        super.dump(out);
        out.println("movingTarget " + movingTarget);
    }

    // --- constructors ---
    public VisualReference(VisualComponent from, VisualComponent to) {
        super(null);
        this.from = from;
        this.to = to;
        fromPoint=getCorrectedCentre(from); // just to have some initial values
        toPoint=getCorrectedCentre(to);
        forgetMidPoints();

	//Note.out(this, "Checking for reference labels..");
        if (from.getStructure() instanceof LabelDecorator) {
            LabelDecorator ld = (LabelDecorator)from.getStructure();
            this.weighted = ld.isReferenceLabelEnabled();
            setName(ld.getReferenceLabel(from.getReferenceNumber(to)));
	} else if (from.getStructure() instanceof Styled) {
	  StyleSheet ss = ((Styled)from.getStructure()).getStyleSheet();
	  this.weighted = ss.isReferenceLabelEnabled();
	  setName(ss.getReferenceLabel(from.getReferenceNumber(to)));
        } else this.weighted = false;

        //addPopUpMenuSeparator();
        //addPopUpCheckboxMenuItem("Weighted", weighted);
    }

    // --- end of constructors ---


    // -- visualization methods: COLOR --

  public boolean hasFocus() {
    return focus;
  }

    /** Toggle weight labels. */
    public void Weighted() {
        weighted = !weighted;
        weight.setVisible(weighted);
        //setInvalid();
    }

    /** Determine the weight flag. 
     * @return true if the reference is labeled by weight; false otherwise
     */
    public boolean isWeighted() {
        return weighted; 
    }

    /** Determine the directed flag. 
     * @return true if the reference is directed; false otherwise
     */
    public boolean isDirected() {
        if (getVisualContainer() instanceof VisualList)
            return true;
        if (getVisualContainer() instanceof VisualTree)
            return ((VisualTree)getVisualContainer()).isDirected();
        if (getVisualContainer() instanceof VisualGraph)
            return ((VisualGraph)getVisualContainer()).isDirected();
        return false;
    }

    /** Forgets previously added midpoints and reversing information. */
    public void forgetMidPoints() {
        midPointReversal = false;
        midPoints = new Vector();
    }

    public void setMidPointOrdering(boolean reversed) {
        midPointReversal = reversed;
    }

    /** Adds a midpoint to the reference.
     * The line is drawn through the point c.getCenter(). This adds a
     * midpoint after all previous midpoints (just before toPoint).
     * @param c a (possibly dummy) component placed at the midpoint
     */
    public void addMidPoint(VisualComponent c) {
        if (midPointReversal)
            midPoints.insertElementAt(c, 0);
        else
            midPoints.addElement(c);
    }

    public void addComponents() {
        if (weight == null)
            weight = new VisualLabel(new Key(getName()));
    }
    
    public void validateComponents() {
	if (weight != null)
	    weight.setVisible(isWeighted());
	if (isWeighted()) {
	    if (getComponentCount() == 0)
		add(weight);
	} else {
	    removeAll();
	}
    }

    public void doLayout() {
        super.doLayout();
        int cCount = getComponentCount();
        for (int i = 0; i < cCount; i++)
            getComponent(i).doLayout();
        if ((getSource() != null) && (getTarget() != null)) {
            fromPoint = getCorrectedCentre(getSource());
            toPoint = getCorrectedCentre(getTarget());

            if (fromPoint.equals(toPoint)) {
                //Note.err(this, 
                //       "source and target are on top of each other!");
                return;
            }

            Point secondPoint, secondToLastPoint;
            if (midPoints.size() > 0) {
                secondPoint=
                    getCorrectedCentre(((VisualComponent)midPoints.
					firstElement()));
                secondToLastPoint =
                    getCorrectedCentre(((VisualComponent)midPoints.
					lastElement()));
            } else {
                secondPoint = toPoint;
                secondToLastPoint = fromPoint;
            }

            double fromAngle =
                normalizeAngle(2*Math.PI -
                               Math.atan2(secondPoint.y - fromPoint.y,
                                          secondPoint.x - fromPoint.x));
            double toAngle = 
                normalizeAngle(2*Math.PI -
                               Math.atan2(secondToLastPoint.y - toPoint.y,
                                          secondToLastPoint.x - toPoint.x));

            fromPoint = getCorrectedBorderAtAngle(getSource(), fromAngle);
            toPoint = getCorrectedBorderAtAngle(getTarget(), toAngle);

            int x1, y1, x2, y2;

            if (fromPoint.x < toPoint.x) { x1 = fromPoint.x; x2 = toPoint.x; }
            else { x2 = fromPoint.x; x1 = toPoint.x; }

            if (fromPoint.y < toPoint.y) { y1 = fromPoint.y; y2 = toPoint.y; }
            else { y2 = fromPoint.y; y1 = toPoint.y; }

            Enumeration mpe = midPoints.elements();
            while (mpe.hasMoreElements()) {
                Point p = getCorrectedCentre(((VisualComponent)mpe.
					      nextElement()));
                if (p.x < x1) x1 = p.x;
                if (p.x > x2) x2 = p.x;
                if (p.y < y1) y1 = p.y;
                if (p.y > y2) y2 = p.y;
            }

            if (isDirected()) {
                // A pessimistic estimate of arrowhead size
                int arrowWidth = (int) (Math.sin(arrowHalfAngle) *
                                        2 * arrowLength);
                int arrowHeight = arrowWidth;
                // Add half of the arrowhead to all borders
                x1 -= arrowWidth / 2 + 1; x2 += arrowWidth / 2 + 1;
                y1 -= arrowHeight / 2 + 1; y2 += arrowHeight / 2 + 1;
            }

            // XXX check that these calculations are still correct!
            if (isWeighted() && (getFont() != null) && (getName() != null)) {
                int lWidth = getFontMetrics(getFont()).stringWidth(getName());
		int lHeight = getFontMetrics(getFont()).getHeight();
                if (x2 - x1 < 2*lWidth) {
                    int disp = (2*lWidth - (x2 - x1)) / 2 + 1;
                    x1 -= disp; x2 += disp;
                }
                if (y2 - y1 < lHeight) {
                    int disp = (lHeight - (y2 - y1)) / 2 + 1;
                    y1 -= disp; y2 += disp;
                }
            }

            Rectangle r = new Rectangle(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
            setBounds(r);
            width = r.width;
            height = r.height;

            // set weight
            if (getComponentCount() > 0) {
                getComponent(0).setBounds(Math.abs((x2-x1)/2)-10,
                                          Math.abs((y2-y1)/2)-10,
                                          20,20);
            }
        }
    }
    /*
    public Dimension getSize() {
        return new Dimension(width, height);
    }
    */

    /** Returns the source component of this reference. */
    public VisualComponent getSource() { return from; }

    /** Sets the source component of this reference. */
    public void setSource(VisualComponent t) { from = t; }

    /** Returns the component this reference is pointing to. */
    public VisualComponent getTarget() { 
        return (movingTarget == null) ? to : movingTarget; 
    }

    /** Sets the component this reference is pointing to. */
    public void setTarget(VisualComponent t) { to = t; }

    /*
    private Point getTargetLocation() {
        return (movingTarget == null) ?
            toPoint :
            movingTarget.getLocation();
            //new Point(pick_x, pick_y);
    }
    */
    private Point getTargetLocation() {
        return toPoint;
    }

    public void assign() {
        getApplication().showMessage("Sorry", "Cannot assign "+ read +
                               " into Reference.\n"+
                               "Try to write (dblClick) the new label.");
    }

    public void write() {
        if ((from != null) && 
            (from.getStructure() instanceof LabelDecorator)) {
            LabelDecorator ld = (LabelDecorator)from.getStructure();
            if (ld == null)
                Note.err(this, "Illegal (null) LabelDecorator");
            if (ld.isReferenceLabelEnabled()) {
                ld.setReferenceLabel(from.getReferenceNumber(to), 
                                     read.getStructure().toString());
                ((VisualLabel)weight).setKey(read.getStructure().toString());
            }
            //else Note.out(this, "Not referenceLabelEnabled "+ld);
        } else Note.warning(this, "from was null while trying to write()");
    }

    public void rename() {
	if (from.getStructure() instanceof LabelDecorator) {
	    LabelDecorator ld = (LabelDecorator)from.getStructure();
	    String newName = getApplication().getInput("Rename: ", ld.getReferenceLabel(from.getReferenceNumber(to)));
	    if (newName == ld.getReferenceLabel(from.getReferenceNumber(to)))
		return;
	    getAnimator().startOperation();
	    setName(newName);
	    ld.setReferenceLabel(from.getReferenceNumber(to), getName());
	    ((VisualLabel)weight).setKey(getName());
	    getAnimator().endOperation();
	    setInvalid();
	}
    }

    public Dimension getPreferredSize() { 
        return new Dimension(10,10);
    }

    public FDT getStructure() {
        return null;
    }

    public VisualType getMovingPart() {
        return movingTarget = new VisualArrow(this);
    }

    public void selectEntity(boolean isShiftDown) {
	if (getSource().getStructure() instanceof Source) {
	    getAnimator().startOperation();
	    ((Source)(getSource().getStructure())).referenceSelectedTo(getTarget().getStructure());
	    getAnimator().endOperation();
	}	
	setInvalid();
    }

    public void drag(int x, int y) {
        pick_x = x;
        pick_y = y;
        movingTarget.setBounds(x+getBounds().x,y+getBounds().y,1,1);
        doLayout();
        repaint();
        //super.drag(x,y);
    }

    public void released() {
	//Note.out(this, "Reference released to " + entered);
        picked = false;
	if (!dragEnabled())
	    return;
        from.moveReference(this, to, entered);         
        movingTarget = null;
	setInvalid();
    }


    private int min(int a, int b) {
        return a < b ? a : b;
    }

    private double distance(int x, int y) {
        int YA,XA,YB,XB,YC,XC;
        
        XC = x + min(fromPoint.x, toPoint.x);
        YC = y + min(fromPoint.y, toPoint.y);
        XA = fromPoint.x;
        YA = fromPoint.y;
        XB = toPoint.x;
        YB = toPoint.y;
        
        double LL = Math.sqrt((XA-XB)*(XA-XB) + (YA-YB)*(YA-YB));
        return ((YB-YA)*(XC-XA)-(XB-XA)*(YC-YA))/LL;
    }
    
    public boolean contains(int x, int y) {
        // XXX needs to use midPoints!
        if ((super.contains(x,y) && (Math.abs(distance(x,y)) < epsilon)))
            return true;
        return false;
    }

    /** Normalizes angle to [0..2pi). */
    private double normalizeAngle(double angle) {
        while (angle < 0)
            angle += 2 * Math.PI;
        while (angle >= 2 * Math.PI)
            angle -= 2 * Math.PI;

        return angle;
    }

    /** Returns the point on perpendicular projection of line (x1,y1)->(x2,y2)
     * and point (x,y) that lies the specified distance away from the 
     * reference line.
     */
    private Point moveAlongPerpendicularProjection(int x1, int y1,
                                                   int x2, int y2,
                                                   int x, int y, 
                                                   int d) {
        if (x1 == x2) return new Point(x+d,y);
        if (y1 == y2) return new Point(x,y-d);
        double k = (x2-x1)/(y1-y2);
        
        double delta_x = Math.sqrt((d*d)/(k*k+1));
        double delta_y = k*delta_x;

        return new Point(x+(int)delta_x,y+(int)delta_y);
    }
    /*
    private void paintWeight(Graphics g, int x1, int y1, int x2, int y2) {
        if (weighted) {
            String name = getName();
            if (name == null) name = "Unknown";
            int lWidth = getFontMetrics(getFont()).stringWidth(name);
            if (lWidth < Math.abs(x2-x1)) {
                lWidth = Math.abs(x2-x1) - lWidth;
            } else 
                lWidth = 0;
            FontMetrics fm = getFontMetrics(getFont());
            int lHeight = fm.getHeight() - fm.getMaxDescent();
            int x = (x1 < x2 ? x1 : x2) + lWidth/2;
            int y = y1 + (y2-y1)/2 + lHeight/2;
            //Point p = moveAlongPerpendicularProjection(x1,y1,x2,y2,x,y,10);
            //x = p.x; y = p.y;

            //g.drawString(name, x, y);
            //if (weight == null) weight = new VisualKey(new Key(name));
            //weight.setBounds(x,y,width,height);
            //weight.paint(g);
        }
    }
    */
    private void drawAngledLine(Graphics g, int x, int y,
                                double angle, int length) {
        g.drawLine(x,
                   y,
                   x + (int) (Math.cos(angle) * length),
                   y - (int) (Math.sin(angle) * length));
    }

    private void paintArrow(Graphics g, int x1, int y1, int x2, int y2) {
        if (isDirected()) {
            double toAngle = 2*Math.PI - Math.atan2(y1 - y2, x1 - x2);
            drawAngledLine(g, x2, y2, toAngle - arrowHalfAngle, arrowLength);
            drawAngledLine(g, x2, y2, toAngle + arrowHalfAngle, arrowLength);
        }
    }

    public boolean isHiding() {
        VisualComponent target = getTarget();
        if ((target == null) || (target.isHiding()))
            return true;
        return false;
    }
    
    public boolean isVisited() {
	if (getSource().getStructure() instanceof Source) 
	    return ((Source)getSource().getStructure()).isVisitedReferenceTo(getTarget().getStructure());
	else
	    return super.isVisited();
    }

    private StyleSheet defaultStyleSheet = null;
    public StyleSheet getDefaultStyleSheet() {
	if (defaultStyleSheet == null) {
            defaultStyleSheet = new StyleSheetAdapter();
	    //defaultStyleSheet.setFocusedBorderColor(Color.red);
	    //defaultStyleSheet.setVisitedFocusedBorderColor(Color.magenta);
	}
        return defaultStyleSheet;
    }

    void setBackgroundColor(Graphics g) {
	Color defaultColor;
        ColorDecorator cd;
	FDT fromFdt = from.getStructure();
	if (fromFdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fromFdt;
	} else if (fromFdt instanceof Styled) {
            cd = ((Styled)fromFdt).getStyleSheet();
	} else 
            cd = getDefaultStyleSheet();
        if (isVisited()) {
            if (hasFocus()) 
                defaultColor = cd.getVisitedFocusedBackgroundColor();
            else
                defaultColor = cd.getVisitedBackgroundColor();
        } else {
            if (hasFocus())
                defaultColor = cd.getFocusedBackgroundColor();
            else
                defaultColor = cd.getDefaultBackgroundColor();
        }
        if (g != null)
            g.setColor(defaultColor);
    }
    
    void setBorderColor(Graphics g) {
	Color defaultColor;
        ColorDecorator cd;
	FDT fromFdt = from.getStructure();
	if (fromFdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fromFdt;
	} else if (fromFdt instanceof Styled) {
            cd = ((Styled)fromFdt).getStyleSheet();
	} else 
            cd = getDefaultStyleSheet();
	if(hasFocus()) {
	  //Note.out(this, this + " hsa focus, using focused color");
	  defaultColor = cd.getFocusedReferenceColor(to.getStructure());
	} else if(isVisited()) {
	  //Note.out(this, this + " hsa focus, using visited color");
	  defaultColor = cd.getModifiedReferenceColor(to.getStructure());
	} else
	  defaultColor = cd.getReferenceColor(to.getStructure());
        if (g != null)
            g.setColor(defaultColor);
    }

    void setPenColor(Graphics g) {
	Color defaultColor;
        ColorDecorator cd;
	FDT fromFdt = from.getStructure();
	if (fromFdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fromFdt;
	} else if (fromFdt instanceof Styled) {
            cd = ((Styled)fromFdt).getStyleSheet();
	} else 
            cd = getDefaultStyleSheet();
        if (isVisited()) {
            if (hasFocus()) 
                defaultColor = cd.getVisitedFocusedPenColor();
            else
                defaultColor = cd.getVisitedPenColor();
        } else {
            if (hasFocus())
                defaultColor = cd.getFocusedPenColor();
            else
                defaultColor = cd.getDefaultPenColor();
        }
        if (g != null)
            g.setColor(defaultColor); 
    }


    public void paint(Graphics g) {
        if (isHiding()) return;
	setBorderColor(g);
        
        Point tp = getTargetLocation(); // because doLayout is not
                                        // called while dragging
        // reset translation to same coordinates as components
        g.translate(-getBounds().x, -getBounds().y);        
        Point last = fromPoint;
        Enumeration mpe = midPoints.elements();
        while (mpe.hasMoreElements()) {
            Point p = getCorrectedCentre(((VisualComponent)mpe.nextElement()));
	    if (!(g instanceof matrix.util.export.MatrixGraphics)) {
		// not painting when exporting, since the result looks bad
		// and references can't have focus when exporting
		if (!hasFocus())
		    g.setColor(getBackground());
		g.fillPolygon(new int[] { last.x + 2, p.x+2, p.x-2, last.x-2 },
			      new int[] { last.y, p.y, p.y, last.y }, 4);
		g.fillPolygon(new int[] { last.x, p.x, p.x, last.x },
			      new int[] { last.y + 2, p.y+2, p.y-2,last.y-2}, 4);
	    }
	    setBorderColor(g);
	    g.drawLine(last.x, last.y, p.x, p.y);
            last = p;
        }
	if (!(g instanceof matrix.util.export.MatrixGraphics)) {
	    // not painting when exporting, since the result looks bad
	    // and references can't have focus when exporting
	    if (!hasFocus()) 
		g.setColor(getBackground());
	    g.fillPolygon(new int[] { last.x + 2, tp.x+2, tp.x-2, last.x-2 },
			  new int[] { last.y, tp.y, tp.y, last.y }, 4);
	    g.fillPolygon(new int[] { last.x, tp.x, tp.x, last.x },
			  new int[] { last.y + 2, tp.y+2, tp.y-2,last.y-2}, 4);
	}
	setBorderColor(g);
	g.drawLine(last.x, last.y, tp.x, tp.y);
        paintArrow(g, last.x, last.y, tp.x, tp.y);
        //paintWeight(g, fromPoint.x, fromPoint.y, tp.x, tp.y);
        g.translate(getBounds().x, getBounds().y); // cancel above translation
        super.paint(g);
    }

    public String toString() {
        return "link " + from + "->" + to;
    }
}
